(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTGrammar`"];
Needs["FMTOptions`"];
Needs["FMTSymbols`"];


BeginPackage["FMTSystem`"];


fullyQualifiedName::usage="fullyQualifiedName[packageName,name] Returns the fully qualified symbol name.";


validateSystem::usage="validateSystem[system] Checks whether the argument is a correct softwareSystemT object.";


validatePackageIsKnown::usage="validatePackageIsKnown[sys,packageName] Checks whether the package name is already registered in the system.";


validateNoPackageDuplicates::usage="validateNoPackageDuplicates[sys,packageName] Checks whether the given package name is not yet registered in the system.";


validateTypeIsKnown::usage="validateTypeIsKnown[sys,packageName,typeName] Checks whether the type name is known globally or within the given package.";


validateObjectsAreKnown::usage="validateObjectsAreKnown[sys,objectNames] Checks whether fully-qualified object names from the given list are all already registered.";


validateOperationIsKnown::usage="validateOperationIsKnown[sys,operationName,bodyRequired] Checks whether the given operation is already registered in the system.\n"<>
"The bodyRequired option (False by default) selects the stronger version for body check in addition operation declaration.";


create::usage="See FMT`createSoftwareSystem.";


fmtGetCurrentOptions::usage="See FMT`getOptions.";


fmtSetCurrentOptions::usage="See FMT`setOptions.";


fmtDeclarePkg::usage="See FMT`declarePackage.";


allPackageNames::usage="allPackageNames[sys] Returns the list of known package names.";


fmtDefineTypes::usage="See FMT`defineDataTypes.";


packageTypeDefs::usage="packageTypeDefs[sys,packageName] Returns <|typeName->definition,...|> for the given package, type names are fully-qualified.";


typeHierarchy::usage="typeHierarchy[sys] Returns list {type->baseOrDef,...} for all types that are based on other types,\n"<>
"currently this involves only arithmetic types, based on integerT. Array and structure types are returned with their complete definitions.";


enumValues::usage="enumValues[sys,packageName,qTypeName] Returns list of enumeration numeric values {name->numVal,...} for the given enumeration type.";


allEnumValues::usage="allEnumValues[sys] Returns the association <|qTypeName->enumValues,...|> for all enum types in the system.";


simpleTypeQ::usage="simpleTypeQ[sys,qTypeName] Returns True if qTypeName is one of standard types (integer or boolean) or is derived from such a type.";


typeCategory::usage="typeCategory[sys,qTypeName] Returns base, array or structure as a category for the given type.";


typesForCategory::usage="typesForCategory[sys,cat] Returns the list of rules defining all types of the given category.";


typeInvariants::usage="typeInvariants[sys,qTypeName] Returns the (possibly empty) list of invariant statements for the given type.";


arrayTypeBounds::usage="arrayTypeBounds[sys,qTypeName] Returns the pair {min,max} as bounds of the given array type.";


fmtDefineData::usage="See FMT`defineDataObjects.";


packageDataDefs::usage="packageDataDefs[sys,packageName] Returns <|objName->type,...|> for the given package, type names are fully-qualified.";


fmtSetData::usage="See FMT`setDataObjectValues.";


fmtGetData::usage="See FMT`getDataObjectValues.";


fmtDeclareOps::usage="See FMT`declareOperations.";


fmtDefineOp::usage="See FMT`defineOperation.";


fmtSetGhostDomains::usage="See FMT`setGhostDomains.";


fmtGetGhostDomains::usage="fmtGetGhostDomains[sys] Returns the association of ghost domains.";


packageOperationDecls::usage="packageOperationDecls[sys,packageName] Returns <|opName->profile,...|> for the given package, operation names are fully-qualified.";


operationExists::usage="operationExists[sys,qOpName] Returns whether the given (fully-qualified) operation exists in the system.";


scanOperationsInPackages::usage="scanOperationsInPackages[sys,argPkgNames,argOpNames,process] Scans all operations that match the given criteria"<>
" and runs process on each selected operation.";


visibleOperationResolutions::usage="visibleOperationResolutions[sys,packageName] Returns the list of rules for operation name resolutions - "<>
"non-qualified from the given package and fully qualified from all visible packages.";


resolvedOperationBody::usage="resolvedOperationBody[sys,packageName,qOperationName,withNumberedWrappers,highlightedSet]"<>
" Returns the operation's body with local and package names resolved.";


operationPart::usage="operationPart[resolvedBody,index] Helper, returns the part of operation's body identified by the given index.\n"<>
"The resolvedBody argument is a value that should be obtained with prior call to resolvedOperationBody with numbered wrappers option selected.";


operationParams::usage="operationParams[sys,qOperationName] Returns the list of parameters' defining rules (name->type) for the given operation.";


operationParamModes::usage="operationParamModes[sys,qOperationName] Returns the list {name->mode,...} for operation's parameters.";


operationLocalVars::usage="operationLocalVars[sys,qOperationName] Returns the list of local variables defining rules (name->type) for the given operation.";


operationLocals::usage="operationLocals[sys,qOperationName] Returns the list of parameters and local variables defining rules (name->type) for the given operation.";


operationInitializedLocals::usage="operationInitializedLocals[sys,qOperationName] Returns the list of parameters that have mode in or inOut.";


operationUninitializedLocals::usage="operationUninitializedLocals[sys,qOperationName] Returns the list of parameters that have mode out plus all local variables.";


operationGlobals::usage="operationGlobals[sys,packageName,qOperationName,resolvedBody] Returns the list of fully qualified names with types\n"<>
"of global variables used by the operation (as given in its resolved body only).";


operationInitializedObjects::usage="operationInitializedObjects[sys,packageName,qOperationName,resolvedBody] Returns the list of fully qualified names with types\n"<>
"of those variables that are considered initialized (globals and in/inOut parameters),\n"<>
"as declared and given in the operation's resolved body plus those found in pre- and post-conditions.";


operationAllUsedObjects::usage="operationAllUsedObjects[sys,packageName,qOperationName,resolvedBody] Returns the list of fully qualified names with types\n"<>
"of all variables (globals and locals) used by the operation,\n"<>
"as declared and given in the operation's resolved body plus those found in pre- and post-conditions.";


operationAllMentionedObjects::usage="operationAllMentionedObjects[sys,packageName,qOperationName,resolvedBody] Returns the list of fully qualified names with types\n"<>
"of all variables (globals and locals) mentioned by the operation, including those that are not directly used\n"<>
"as declared and given in the operation's resolved body plus those found in pre- and post-conditions, params and local variables.";


operationAssumptions::usage="operationAssumptions[sys,packageName,qOperationName] Returns the list of proof statements representing the pre-conditions of the given operation.";


operationPromises::usage="operationPromises[sys,packageName,qOperationName] Returns the list of proof statements representing the post-conditions of the given operation.";


operationDependentPackages::usage="operationDependentPackages[sys,packageName,qOperationName] Returns the list of packages that the given operation refers to,\n"<>
"by means of type, enum value, data object references and operation calls.\n"<>
"The return value is {list-of-specification-dependencies,list-of-body-dependencies}";


dependentPackages::usage="dependentPackages[sys,packageName] Returns the list of packages that the given package refers to,\n"<>
"by means of data object references and operation definitions.\n"<>
"The return value is {list-of-specification-dependencies,list-of-body-dependencies}";


packageLevelNames::usage="packageLevelNames[sys,packageName] Returns the joint list of all package-level names (types, enums, operations, data).";


fmtGetComment::usage="fmtGetComment[sys,name] Returns the comment (String, Null or None) for the given entity name.";


Begin["`Private`"];


fullyQualifiedName[packageName_String,name_Symbol]:=
	Symbol[packageName<>SymbolName[name]]


(* Helper validators for function parameters.
	Note: calling these validators is the responsibility of main package FMT,
	all functions here assume that their parameters were already validated.
*)


validateSystem[system_]:=
	If[Head[system]=!=softwareSystemT,
		Message[validateSystem::notsys]
	]

validateSystem::notsys="First argument is not a software system.";


validatePackageIsKnown[sys_,packageName_String]:=
	If[Not[MemberQ[sys[packageNames],packageName]],
		Message[validatePackageIsKnown::unkn,packageName]
	]

validatePackageIsKnown::unkn="Package `1` is not known.";


validateNoPackageDuplicates[sys_,packageName_String]:=
	If[MemberQ[sys[packageNames],packageName],
		Message[validateNoPackageDuplicates::dup,packageName]
	]

validateNoPackageDuplicates::dup="Duplicate entry for package `1` is not allowed.";


validateTypeIsKnown[sys_,packageName_String,typeName_Symbol]:=With[
	{
		qAllTypes=Join[
			Keys[sys[typeDefinitions]],
			{FMTSymbols`booleanT,FMTSymbols`integerT} (* these types are always known *)
		],
		packageTypes=sys[packageTypes][packageName]
	},
	
	If[Not[MemberQ[qAllTypes,typeName]] && Not[MemberQ[packageTypes,typeName]],
		Message[validateTypeIsKnown::unkn,typeName,packageName]
	]
]

validateTypeIsKnown::unkn="Type `1` is not a fully-qualified type name and is not known in package `2`.";


validateObjectsAreKnown[sys_,objectNames_List]:=
	Scan[
		Function[objName,
			If[Not[MemberQ[Keys[sys[dataObjectValues]],objName]],
				Message[validateObjectsAreKnown::unkn,objName]
			]
		],
		objectNames
	]

validateObjectsAreKnown::unkn="Object `1` is not known.";


validateOperationIsKnown[sys_,operationName_,OptionsPattern[{bodyRequired->False}]]:=
	If[Not[KeyExistsQ[sys[operationDeclarations],operationName]],
		Message[validateOperationIsKnown::unkn,operationName],
		If[And[
			OptionValue[bodyRequired],
			Not[KeyExistsQ[sys[operationBodies],operationName]]],
			Message[validateOperationIsKnown::nobody,operationName]
		]
	]

validateOperationIsKnown::unkn="Operation `1` is not known.";
validateOperationIsKnown::nobody="Operation `1` has no body defined.";


create[options_Association]:=
	softwareSystemT[<|
		packageNames->{},

		(* keys are package names (strings) and values are lists of non-qualified names *)
		packageTypes-><||>,
		packageDataObjects-><||>,
		packageOperations-><||>,
		
		(* in all these associations, keys are fully qualified name symbols *)
		typeDefinitions-><||>,
		dataObjectDeclarations-><||>,
		dataObjectValues-><||>,
		operationDeclarations-><||>,
		operationPreconditions-><||>,
		operationPostconditions-><||>,
		operationLocalVariables-><||>,
		operationBodies-><||>,
		entityComments-><||>,
		ghostDomains-><||>,
		
		systemOptions->options
	|>]


fmtGetCurrentOptions[sys_]:=sys[systemOptions]


fmtSetCurrentOptions[sys_,options_Association]:=sys[systemOptions]=options

SetAttributes[fmtSetCurrentOptions,HoldFirst]


fmtDeclarePkg[sys_,packageName_String,comment_]:=(
	AppendTo[sys[packageNames],packageName];
	sys[packageTypes][packageName]={};
	sys[packageDataObjects][packageName]={};
	sys[packageOperations][packageName]={};
	fmtSetComment[sys,packageName,comment]
)

SetAttributes[fmtDeclarePkg,HoldFirst]


allPackageNames[sys_]:=sys[packageNames]


fmtDefineTypes[sys_,packageName_String,types_List,comment_]:=Module[
	{
		firstQName=Null
	},
	
	If[Not[KeyExistsQ[sys[packageTypes],packageName]],
		sys[packageTypes][packageName]={}
	];
	
	Scan[
		Function[type,
			With[
				{
					typeName=First[type],
					qTypeName=fullyQualifiedName[packageName,First[type]],
					typeDef=Last[type]
				},
				
				If[firstQName===Null,
					firstQName=qTypeName
				];
				
				If[KeyExistsQ[sys[typeDefinitions],qTypeName],
					Message[fmtDefineTypes::replace,qTypeName],
					AppendTo[sys[packageTypes][packageName],typeName]
				];
				
				Switch[Head[typeDef],
					FMTSymbols`number,
						(* simple numeric type, with some known base type *)
						sys[typeDefinitions][qTypeName]=Association[
							List@@typeDef/.typeName->qTypeName (* use fully-qualified type name
															to ease later variable name replacements *)
						],
					
					FMTSymbols`enum,
						(* enumeration type,
							Note: enum values are stored un-qualified *)
						sys[typeDefinitions][qTypeName]=typeDef,
					
					FMTSymbols`array,
						(* array type, typeDef is array[elementType,size] *)
						With[
							{
								elementType=typeDef[[1]]
							},
							
							If[MemberQ[FMTSymbols`standardTypes,elementType]||
								KeyExistsQ[sys[typeDefinitions],elementType],
								
								(* element type is already recognized as a standard type
									or is qualified and found within existing types *)
								sys[typeDefinitions][qTypeName]=typeDef,
								
								(* element type refers to the unqualified type within the same package *)
								With[
									{
										qElementType=fullyQualifiedName[packageName,elementType]
									},
									
									sys[typeDefinitions][qTypeName]=typeDef/.elementType->qElementType
								]
							]
						],
					
					FMTSymbols`structure,
						(* structure type, typeDef is structure[field\[Rule]type,...] *)
						Module[
							{
								fields=Association@@typeDef,
								fieldName,
								fieldType,
								qFieldType
							},
							
							(* check field types and coerce them to fully-qualified type names
								(if they are not standard types) *)
							Scan[
								Function[field,
									fieldName=First[field];
									fieldType=Last[field];
									
									If[Not[MemberQ[FMTSymbols`standardTypes,fieldType]||
										KeyExistsQ[sys[typeDefinitions],fieldType]],
										
										(* field type was not recognized as a standard type
											and was not found among existing types,
											assume it refers to the unqualified type within the same package *)
										qFieldType=fullyQualifiedName[packageName,fieldType];
										
										fields[fieldName]=qFieldType;
									]
								],
								typeDef
							];
							
							sys[typeDefinitions][qTypeName]=FMTSymbols`structure@@Normal[fields]
						]
				]
			]
		],
		types
	];
	
	fmtSetComment[sys,firstQName,comment]
]

SetAttributes[fmtDefineTypes,HoldFirst]

fmtDefineTypes::replace="Warning: Replacing existing definition for type `1`.";


packageTypeDefs[sys_,packageName_String]:=Association[
	Map[
		Function[typeName,
			With[
				{
					qTypeName=fullyQualifiedName[packageName,typeName]
				},
				
				qTypeName->sys[typeDefinitions][qTypeName]
			]
		],
		sys[packageTypes][packageName]
	]
]


typeHierarchy[sys_]:=With[
	{
		typeDefs=sys[typeDefinitions]
	},
	
	Map[
		Function[typeName,
			Switch[Head[typeDefs[typeName]],
				Association,
					(* this is the integer type definition, show its base type *)
					typeName->typeDefs[typeName][FMTSymbols`base],
				
				FMTSymbols`array|FMTSymbols`structure,
					(* this is array or structure, show its complete definition *)
					typeName->typeDefs[typeName],
				
				FMTSymbols`enum,
					(* there is no need for enum information in the type hierarchy *)
					Nothing
			]
		],
		Keys[typeDefs]
	]
]


enumValues[sys_,packageName_String,qTypeName_]:=With[
	{
		enumSymbols=List@@Map[
			Function[e,
				fullyQualifiedName[packageName,e]
			],
			sys[typeDefinitions][qTypeName]
		]
	},
	
	Thread[enumSymbols->Range[Length[enumSymbols]]]
]


allEnumValues[sys_]:=Association@Flatten[
	Map[
		Function[pkgName,
			Map[
				Function[typeName,
					With[
						{
							qTypeName=fullyQualifiedName[pkgName,typeName]
						},
						
						If[Head[sys[typeDefinitions][qTypeName]]===FMTSymbols`enum,
							qTypeName->enumValues[sys,pkgName,qTypeName],
							{}
						]
					]
				],
				sys[packageTypes][pkgName]
			]
		],
		sys[packageNames]
	]
]


simpleTypeQ[sys_,qTypeName_]:=With[
	{
		typeDefs=sys[typeDefinitions]
	},
	
	(* the type is considered simple if it belongs to the set of standard types
		or it is derived from such a type: *)
	MemberQ[FMTSymbols`standardTypes,qTypeName]||
		(KeyExistsQ[typeDefs,qTypeName]&&
			AssociationQ[typeDefs[qTypeName]]&&
				MemberQ[FMTSymbols`standardTypes,typeDefs[qTypeName][base]])
]


typeCategory[sys_,qTypeName_Symbol]:=
	If[qTypeName===FMTSymbols`booleanT||qTypeName===FMTSymbols`integerT,
		FMTSymbols`base,
		
		With[
			{
				typeDef=sys[typeDefinitions][qTypeName]
			},
			
			Head[typeDef]
		]
	]


typesForCategory[sys_,cat_]:=Select[
	Normal[sys[typeDefinitions]],
	Function[tr, (* type rule: typeName\[Rule]def *)
		With[
			{
				def=tr[[2]]
			},
			
			Switch[cat,
				FMTSymbols`base,
					AssociationQ[def]||Head[def]===FMTSymbols`booleanT,
				
				FMTSymbols`array,
					Head[def]===FMTSymbols`array,
				
				FMTSymbols`structure,
					Head[def]===FMTSymbols`structure
			]
		]
	]
]


(* Helper, returns the statement for expr being within base type invariants for typeName *)
baseTypeInvariants[sys_,typeName_,expr_]:=
	If[typeName===FMTSymbols`integerT,
		With[
			{
				intBounds=sys[systemOptions][FMTOptions`integerBounds]
			},
			
			intBounds[[1]]<=expr<=intBounds[[2]]
		],
		
		True
	]


typeInvariants[sys_,qTypeName_]:=Module[
	{
		typeDefs=sys[typeDefinitions],
		explicitInvariants=True,
		baseType=qTypeName,
		baseInvariants=True
	},
	
	If[KeyExistsQ[typeDefs,qTypeName],
		If[AssociationQ[typeDefs[qTypeName]],
			explicitInvariants=Lookup[typeDefs[qTypeName],FMTSymbols`invariant,True];
			baseType=typeDefs[qTypeName][base]
		]
	];
	
	If[MemberQ[FMTSymbols`standardTypes,baseType],
		baseInvariants=baseTypeInvariants[sys,baseType,qTypeName]
	];
	
	And[baseInvariants,explicitInvariants]
]


arrayTypeBounds[sys_,qTypeName_]:=Module[
	{
		typeDefs=sys[typeDefinitions],
		indexOffset=sys[systemOptions][FMTOptions`arrayIndexOffset]
	},
	
	{indexOffset,typeDefs[qTypeName][[2]]-1+indexOffset}
]


(* Helper, qualified type name if not already qualified *)
ensureTypeNameIsQualified[sys_,packageName_String,typeName_Symbol]:=
	If[MemberQ[Keys[sys[typeDefinitions]],typeName] ||
		MemberQ[FMTSymbols`standardTypes,typeName],
		
		(* type name is already qualified or is a standard type name *)
		typeName,
		
		(* type needs to be qualified *)
		fullyQualifiedName[packageName,typeName]
	]


fmtDefineData[sys_,packageName_String,objects_List,comment_]:=Module[
	{
		firstQName=Null
	},
	
	(* add objects as uninitialized entries *)

	If[Not[KeyExistsQ[sys[packageDataObjects],packageName]],
		sys[packageDataObjects][packageName]={}
	];
	
	Scan[
		Function[obj,
			With[
				{
					objName=First[obj],
					qObjName=fullyQualifiedName[packageName,First[obj]],
					objType=Last[obj]
				},
				
				If[firstQName===Null,
					firstQName=qObjName
				];
				
				If[KeyExistsQ[sys[dataObjectDeclarations],qObjName],
					Message[fmtDefineData::replace,qObjName],
					AppendTo[sys[packageDataObjects][packageName],objName]
				];
				
				sys[dataObjectDeclarations][qObjName]=
					ensureTypeNameIsQualified[sys,packageName,objType];
				
				sys[dataObjectValues][qObjName]=None
			]
		],
		objects
	];
	
	fmtSetComment[sys,firstQName,comment]
]

SetAttributes[fmtDefineData,HoldFirst]

fmtDefineData::replace="Warning: Replacing existing definition for data object `1`.";


packageDataDefs[sys_,packageName_String]:=Association[
	Map[
		Function[objName,
			With[
				{
					qObjName=fullyQualifiedName[packageName,objName]
				},
				
				qObjName->sys[dataObjectDeclarations][qObjName]
			]
		],
		sys[packageDataObjects][packageName]
	]
]


fmtSetData[sys_,settings_List]:=
	AppendTo[sys[dataObjectValues],settings]

SetAttributes[fmtSetData,HoldFirst]


fmtGetData[sys_,objNames_List]:=
	Thread[objNames->Lookup[sys[dataObjectValues],objNames]]

SetAttributes[fmtGetData,HoldFirst]


(* Helper, transforms parameter specification into mode[name\[Rule]type] *)
normalizeOperationSignature[sys_,packageName_String,signature_]:=
	Map[
		Function[param,
			If[Head[param]===Rule,
				(* user-provided specification has form name\[Rule]mode[type] *)
				
				With[
					{
						name=param[[1]],
						mode=param[[2,0]],
						type=ensureTypeNameIsQualified[sys,packageName,param[[2,1]]]
					},
					
					mode[name->type]
				],
				
				(* user-provided specification already has the form mode[name\[Rule]type],
					but there might be a need to qulify the type name: *)
				With[
					{
						name=param[[1,1]],
						mode=param[[0]],
						type=ensureTypeNameIsQualified[sys,packageName,param[[1,2]]]
					},
					
					mode[name->type]
				]
			]
		],
		signature
	]

(* Makes sure conditions are presented as lists *)
normalizePrePostConditions[cond_]:=
	If[Head[cond]===List,
		cond,
		List[cond]
	]

fmtDeclareOps[sys_,packageName_,operationProfiles_,comment_]:=Module[
	{
		firstQName=Null
	},
	
	If[Not[KeyExistsQ[sys[packageOperations],packageName]],
		sys[packageOperations][packageName]={}
	];
	
	Scan[
		Function[opProfile,
			Module[
				{
					opHead=Head[opProfile],
					opSignature,
					opName,
					qOpName,
					opOptions,
					preconditions={},
					postconditions={}
				},
				
				If[opHead===Rule,
					opSignature=First[opProfile];
					opName=Head[opSignature];
					qOpName=fullyQualifiedName[packageName,opName];
					opOptions=Association[Last[opProfile]];
					preconditions=normalizePrePostConditions[Lookup[opOptions,FMTSymbols`pre,{}]];
					postconditions=normalizePrePostConditions[Lookup[opOptions,FMTSymbols`post,{}]],
					
					opSignature=opProfile;
					opName=opHead;
					qOpName=fullyQualifiedName[packageName,opName]
				];
				
				If[firstQName===Null,
					firstQName=qOpName
				];
				
				If[MemberQ[Keys[sys[operationDeclarations]],qOpName],
					Message[fmtDeclareOps::replace,qOpName],
					AppendTo[sys[packageOperations][packageName],opName]
				];
				
				sys[operationDeclarations][qOpName]=
					normalizeOperationSignature[sys,packageName,opSignature];
				sys[operationPreconditions][qOpName]=preconditions;
				sys[operationPostconditions][qOpName]=postconditions
			]
		],
		operationProfiles
	];
	
	fmtSetComment[sys,firstQName,comment]
]

SetAttributes[fmtDeclareOps,HoldAll];

fmtDeclareOps::replace="Warning: Replacing existing declaration for operation `1`.";


fmtDefineOp[sys_,packageName_String,qOpName_,localObjects_List,operationBody_]:=With[
	{
		paramNames=List@@sys[operationDeclarations][qOpName][[All,1]]
	},
	
	If[KeyExistsQ[sys[operationBodies],qOpName],
		Message[fmtDefineOp::replace,qOpName]
	];
	
	(* check if local objects do not hide operation parameters
		or package-level object with the same name *)
	
	Scan[
		Function[objDef, (* name\[Rule]typeName *)
			With[
				{
					objName=objDef[[1]],
					
					(* package-qualified name corresponding to local var name: *)
					qObjName=fullyQualifiedName[packageName,objDef[[1]]]
				},
				
				If[MemberQ[paramNames,objName]||
					KeyExistsQ[sys[packageDataObjects],qObjName],
					Message[fmtDefineOp::varhiding,objName]
				]
			]
		],
		localObjects
	];
	
	sys[operationLocalVariables][qOpName]=Map[
		Function[objDef, (* name\[Rule]typeName *)
			With[
				{
					objName=objDef[[1]],
					objType=objDef[[2]]
				},
				
				objName->ensureTypeNameIsQualified[sys,packageName,objType]
			]
		],
		localObjects
	];
	
	sys[operationBodies][qOpName]=Hold[operationBody]
]

SetAttributes[fmtDefineOp,HoldAll];

fmtDefineOp::replace="Warning: Replacing existing definition for operation `1`.";
fmtDefineOp::varhiding="Warning: Local variable `1` hides an existing name.";


fmtSetGhostDomains[sys_,domains_]:=Scan[
	Function[domain, (* Element[symbol,domain] *)
		sys[ghostDomains][domain[[1]]]=domain[[2]]
	],
	domains
]

SetAttributes[fmtSetGhostDomains,HoldFirst]


fmtGetGhostDomains[sys_]:=sys[ghostDomains]


packageOperationDecls[sys_,packageName_String]:=Association[
	Map[
		Function[opName,
			With[
				{
					qOpName=fullyQualifiedName[packageName,opName]
				},
				
				qOpName->sys[operationDeclarations][qOpName]
			]
		],
		sys[packageOperations][packageName]
	]
]


operationExists[sys_,qOpName_]:=
	KeyExistsQ[sys[operationBodies],qOpName]


scanOperationsInPackages[sys_,argPkgNames_,argOpNames_,process_]:=Module[
	{
		pkgNames={},
		opNames={},
		results=<||>
	},
	
	If[argPkgNames===All,
		pkgNames=sys[packageNames],
		If[Head[argPkgNames]===String,
			pkgNames={argPkgNames},
			If[Head[argPkgNames]===List,
				pkgNames=argPkgNames
			]
		]
	];
	
	If[argOpNames===All,
		opNames=All,
		If[Head[argOpNames]===Symbol,
			opNames={argOpNames},
			If[Head[argOpNames]===List,
				opNames=argOpNames
			]
		]
	];
	
	Scan[
		Function[pkgName,
			If[Not[MemberQ[sys[packageNames],pkgName]],
				Message[scanOperationsInPackages::unkpkg,pkgName]
			];
			
			If[Not[KeyExistsQ[results,pkgName]],
				results[pkgName]=<||>
			];
			
			Scan[
				Function[opName,
					If[MemberQ[sys[packageOperations][pkgName],opName],
						results[pkgName][opName]=process[pkgName,opName],
						Message[scanOperationsInPackages::nooppkg,opName,pkgName]
					]
				],
				If[opNames===All,
					sys[packageOperations][pkgName],
					opNames
				]
			]
		],
		pkgNames
	];
	
	results
]

scanOperationsInPackages::unkpkg="Unknown package `1`.";
scanOperationsInPackages::nooppkg="Operation `1` was not found in package `2`.";


visibleOperationResolutions[sys_,packageName_String]:=Module[
	{
		qOpName,
		namesInPackage,
		allQualifiedNames
	},
	
	namesInPackage=Map[
		Function[opName,
			qOpName=FMTSystem`fullyQualifiedName[packageName,opName];
			opName->qOpName@@sys[operationDeclarations][qOpName]
		],
		sys[packageOperations][packageName]
	];
	
	allQualifiedNames=Map[
		Function[qOpName,
			qOpName->qOpName@@sys[operationDeclarations][qOpName]
		],
		Keys[sys[operationDeclarations]]
	];
	
	(* Note: the order of these names is important, as they are used in sanitizing operation calls;
		this order prevents unqualified names to be resolved twice (first unqual\[Rule]qual, and then qual\[Rule]qual).
	*)
	Join[
		allQualifiedNames,
		namesInPackage
	]
]


(* Helper, returns the list of object resolutions rules for the given package,
	if unqualified lookup isrequested, there are additional rules for unqialified
	object names from that package *)
objNameResolutionsForPackage[sys_,packageName_String,withUnqualifiedLookup_]:=Module[
	{
		result={}
	},
	
	Scan[
		Function[objName,
			Module[
				{
					qObjName=fullyQualifiedName[packageName,objName],
					qObjType
				},
				
				qObjType=sys[dataObjectDeclarations][qObjName];
				
				AppendTo[result,qObjName->{qObjName,qObjType}]; (* for qualified lookup *)
				If[withUnqualifiedLookup,
					AppendTo[result,objName->{qObjName,qObjType}] (* for unqualified lookup *)
				]
			]
		],
		sys[packageDataObjects][packageName]
	];
	
	result
]

(* Helper, returns the list of object resolution rules
	(parameters + local variables + package-level variables + variables from other packages)
	for the given operation *)
objNameResolutions[sys_,packageName_String,qOperationName_Symbol,withLocalVars_]:=Module[
	{
		result={}
	},
	
	(* 1. collect local objects for this operation *)
	Scan[
		Function[obj,
			Module[
				{
					objName=First[obj],
					objType=Last[obj],
					qObjType
				},
				
				qObjType=ensureTypeNameIsQualified[sys,packageName,objType];
				
				(* local objects resolve to themselves and do not have other qualified name *)
				AppendTo[result,objName->{objName,qObjType}]
			]
		],
		
		If[withLocalVars,
			operationLocals[sys,qOperationName], (* all params and local variables *)
			operationParams[sys,qOperationName]  (* operation parameters only *)
		]
	];
	
	(* 2. collect package data objects from the same package (including unqualified lookup)
		and from all the other packages (with fully-qualified lookup only) *)
	result=Join[
		result,
		Flatten@Map[
			Function[pkgName,
				objNameResolutionsForPackage[sys,pkgName,pkgName==packageName]
			],
			sys[packageNames]
		]
	];
	
	result
]


(* Helper, returns the list of resolutions for enumeration literals *)
enumResolutionsForPackage[sys_,packageName_String,withUnqualifiedLookup_]:=Module[
	{
		result={}
	},
	
	(* Gather resolutions from the given (current) package: *)
	Scan[
		Function[typeName,
			Module[
				{
					qTypeName=fullyQualifiedName[packageName,typeName],
					typeDef
				},
				
				typeDef=sys[typeDefinitions][qTypeName];
				If[Head[typeDef]===FMTSymbols`enum,
					Scan[
						Function[e, (* enum value *)
							With[
								{
									qE=fullyQualifiedName[packageName,e]
								},
								
								(* Note: rule for qualified lookup is created first,
									to avoid double-replacement of unqualified symbols
									when this list is processed in sequence *)
								AppendTo[result,qE->{qE,qTypeName}]; (* for qualified lookup *)
								If[withUnqualifiedLookup,
									AppendTo[result,e->{qE,qTypeName}] (* for unqualified lookup *)
								]
							]
						],
						typeDef
					]
				]
			]
		],
		sys[packageTypes][packageName]
	];
	
	result
]

enumResolutions[sys_,packageName_String]:=Flatten@Map[
	Function[pkgName,
		(* get all lookups (including unqualified) for the given package,
			and only qualified lookups from all other packages *)
		enumResolutionsForPackage[sys,pkgName,pkgName==packageName]
	],
	sys[packageNames]
]


resolvedOperationBody[sys_,packageName_String,qOperationName_Symbol,withNumberedWrappers_,highlightedSet_]:=
	If[KeyExistsQ[sys[operationBodies],qOperationName],
		FMTGrammar`resolveSyntax[
			sys[operationBodies][qOperationName],
			sys[systemOptions][FMTOptions`grammarPreprocessor],
			typeHierarchy[sys],
			allEnumValues[sys],
			objNameResolutions[sys,packageName,qOperationName,True], (* resolve with local vars *)
			enumResolutions[sys,packageName],
			visibleOperationResolutions[sys,packageName],
			numbered->withNumberedWrappers,
			highlighted->highlightedSet
		],
		Null
	]

SetAttributes[resolvedOperationBody,HoldFirst]


operationPart[resolvedBody_,index_]:=Cases[
	resolvedBody,
	FMTSymbols`numberedT[index,part_]->part,
	All
][[1]]


(* Helper, returns the rules for operation parameters only *)
operationParams[sys_,qOperationName_Symbol]:=
	(* parameter names with types, if there are any: *)
	List@@sys[operationDeclarations][qOperationName][[All,1]]


operationParamModes[sys_,qOperationName_Symbol]:=Map[
	Function[p, (* p is mode[name\[Rule]type] *)
		With[
			{
				mode=Head[p],
				name=p[[1,1]]
			},
			
			{name,mode}
		]
	],
	List@@sys[operationDeclarations][qOperationName]
]


operationLocalVars[sys_,qOperationName_Symbol]:=
	Lookup[sys[operationLocalVariables],qOperationName,Null]


operationLocals[sys_,qOperationName_Symbol]:=
	(* operation locals include both operation parameters (from its signature)
		and its local variables: *)
	Join[
		(* parameter names with types, if there are any: *)
		operationParams[sys,qOperationName],
		
		(* local variables: *)
		If[KeyExistsQ[sys[operationLocalVariables],qOperationName],
			sys[operationLocalVariables][qOperationName],
			
			(* operation is declared, but not yet defined *)
			{}
		]
	]


(* initialized locals are those parameters that are of mode in or inOut,
	as they are already initialized by the caller *)
operationInitializedLocals[sys_,qOperationName_Symbol]:=
	Cases[sys[operationDeclarations][qOperationName],p_FMTSymbols`in|p_FMTSymbols`inOut:>p[[1]],{1}]


(* uninitialized locals are parameters of mode out
	and all local variables *)
operationUninitializedLocals[sys_,qOperationName_Symbol]:=
	Join[
		(* out parameter names with types, if there are any: *)
		Cases[sys[operationDeclarations][qOperationName],p_FMTSymbols`out:>p[[1]],{1}],
		
		(* local variables: *)
		sys[operationLocalVariables][qOperationName]
	]


(* Helper, identifies objects from pre- and post-conditions
	(those that are found among globals or params) *)
operationPrePostConditionObjects[sys_,packageName_String,qOperationName_Symbol]:=Module[
	{
		resolvedPrePost
	},

	resolvedPrePost={
		operationAssumptions[sys,packageName,qOperationName],
		operationPromises[sys,packageName,qOperationName]
	};
	
	Cases[resolvedPrePost,FMTSymbols`objectT[n_,t_]->Rule[n,t],All]
]


operationInitializedObjects[sys_,packageName_String,qOperationName_Symbol,resolvedBody_]:=
	Union[
		operationGlobals[sys,packageName,qOperationName,resolvedBody], (* all objects (globals+locals) *)
		operationInitializedLocals[sys,qOperationName] (* in and inOut parameters *)
	]


operationGlobals[sys_,packageName_String,qOperationName_Symbol,resolvedBody_]:=Module[
	{
		(* all object names (globals+locals) with their types *)
		allUsedObjects=operationAllUsedObjects[sys,packageName,qOperationName,resolvedBody],
		topName, (* name of the top-level object, in case of partT or dotT constructs *)
		
		(* all object names with types, without those that depend on locals as their parents
			for example, element of local array is excluded from the list of globals) *)
		cleanedGlobals={},
		
		(* all parameters and local variable names with types *)
		locals=operationLocals[sys,qOperationName],
		localNames
	},
	
	localNames=locals[[All,1]];
	
	Scan[
		Function[obj, (* name\[Rule]type *)
			topName=FMTGrammar`accessTopObject[obj[[1]]];
			If[MatchQ[topName,_FMTSymbols`objectParentT],
				topName=topName[[1]]
			];
			
			If[Not[MemberQ[localNames,topName]],
				AppendTo[cleanedGlobals,obj]
			]
		],
		allUsedObjects
	];
	
	cleanedGlobals
]


operationAllUsedObjects[sys_,packageName_String,qOperationName_Symbol,resolvedBody_]:=Module[
	{
		allBodyObjects, (* anything mentioned in the body *)
		bodyFullObjects (* objects mentioned in the body,
							but without subobjects (array element or structure field accesses) *)
	},
	
	allBodyObjects=Cases[resolvedBody,
		FMTSymbols`objectT[v_,t_]|FMTSymbols`objectLhsT[v_,t_],
		All
	];
	
	(* TODO: verify, most likely there is no point in pruning this list
		and the bodyFullObjects variable is not needed at all here *)
	(*
	bodyFullObjects=DeleteCases[allBodyObjects,
		FMTSymbols`objectT[_FMTSymbols`partT,_]|FMTSymbols`objectLhsT[_FMTSymbols`partT,_]|
		FMTSymbols`objectT[_FMTSymbols`dotT,_]|FMTSymbols`objectLhsT[_FMTSymbols`dotT,_]
	];
	*)
	
	Union[
		operationPrePostConditionObjects[sys,packageName,qOperationName],
		(*Rule@@@bodyFullObjects*)
		Rule@@@allBodyObjects
	]
]


operationAllMentionedObjects[sys_,packageName_String,qOperationName_Symbol,resolvedBody_]:=
	Union[
		operationAllUsedObjects[sys,packageName,qOperationName,resolvedBody],
		operationLocals[sys,qOperationName]
	]


operationAssumptions[sys_,packageName_String,qOperationName_Symbol]:=With[
	{
		preConds=sys[operationPreconditions][qOperationName]
	},
	
	If[preConds=!={},
		FMTGrammar`resolveSyntax[
			preConds,
			sys[systemOptions][FMTOptions`grammarPreprocessor],
			typeHierarchy[sys],
			allEnumValues[sys],
			objNameResolutions[sys,packageName,qOperationName,False], (* resolve without local vars *)
			enumResolutions[sys,packageName],
			visibleOperationResolutions[sys,packageName]
		],
		
		{}
	]
]


operationPromises[sys_,packageName_String,qOperationName_Symbol]:=With[
	{
		postConds=sys[operationPostconditions][qOperationName]
	},
	
	If[postConds=!={},
		FMTGrammar`resolveSyntax[
			sys[operationPostconditions][qOperationName],
			sys[systemOptions][FMTOptions`grammarPreprocessor],
			typeHierarchy[sys],
			allEnumValues[sys],
			objNameResolutions[sys,packageName,qOperationName,False], (* resolve without local vars *)
			enumResolutions[sys,packageName],
			visibleOperationResolutions[sys,packageName]
		],
		
		{}
	]
]


(* Helper, extracts (as a string) the package-name part of the symbol name *)
extractPackageName[qItemName_]:=Module[
	{
		str,
		pos
	},
	
	Switch[Head[qItemName],
		Symbol,
			(* simple name *)
			str=ToString[qItemName];
			pos=StringPosition[str,"`"];
			
			If[pos!={},
				StringTake[str,pos[[-1,1]]],
				Null
			],
		
		FMTSymbols`partT|FMTSymbols`dotT,
			(* array access or struct field, descend with parent name only *)
			With[
				{
					(* qItemName is partT[objectParentT[arrayName],idx] or
						 dotT[objectParentT[structName],field] *)
					parent=qItemName[[1,1]]
				},
				extractPackageName[parent]
			]
	]
]

operationDependentPackages[sys_,packageName_String,qOperationName_Symbol]:=Module[
	{
		preConds, (* grammar-resolved *)
		postConds, (* grammar-resolved *)
		params,
		body, (* grammar-resolved *)
		localVars, (* {var\[Rule]type,...} *)
		
		bodyTypes,
		specEnums,bodyEnums,
		specObjects,bodyObjects,
		operationCalls,
		
		specDependencies,bodyDependencies
	},
	
	preConds=operationAssumptions[sys,packageName,qOperationName];
	postConds=operationPromises[sys,packageName,qOperationName];
	params=operationParams[sys,qOperationName];
	
	specEnums=Union[
		Cases[{preConds,postConds},FMTSymbols`literalT[v_Symbol,_]->v,All]
	];
	
	specObjects=Union[
		params[[All,1]], (* params is {name\[Rule]type,...} *)
		Cases[{preConds,postConds},
			FMTSymbols`objectT[v_,_]|
			FMTSymbols`objectParentT[v_,_]->v,
			All
		]
	];
	
	specDependencies=Union[
		extractPackageName/@Join[specEnums,specObjects]
	]/.Null->Nothing;
	
	(* body and localVars are meaningful only when the operation is fully defined: *)
	If[KeyExistsQ[sys[operationBodies],qOperationName],
		body=resolvedOperationBody[sys,packageName,qOperationName,True,{}];
		localVars=operationLocals[sys,qOperationName];
		
		bodyTypes=Union[
			Last/@localVars
		];
		
		bodyEnums=Union[
			Cases[body,FMTSymbols`literalT[v_Symbol,_]->v,All]
		];
		
		bodyObjects=Union[
			Cases[body,
				FMTSymbols`objectT[v_,_]|
				FMTSymbols`objectLhsT[v_,_]|
				FMTSymbols`objectParentT[v_,_]->v,
				All
			]
		];
		
		operationCalls=Union[
			Cases[body,
				FMTSymbols`callT[op_,fs_FMTSymbols`formalParamsT,as_FMTSymbols`actualParamsT]->op,
				All
			]
		];
		
		bodyDependencies=Union[
			extractPackageName/@Join[bodyTypes,bodyEnums,bodyObjects,operationCalls]
		]/.Null->Nothing,
		
		(* this operation is declared only and has no body defined: *)
		bodyDependencies={}
	];
	
	{specDependencies,bodyDependencies}
]


dependentPackages[sys_,packageName_String]:=Module[
	{
		packageVarTypes,
		varDependencies,
		operationDependencies,
		opSpecDependencies,opBodyDependencies,
		pkgSpecDependencies,pkgBodyDependencies
	},
	
	(* 1. gather dependencies from package data objects *)
	
	packageVarTypes=Map[
		Function[varName,
			With[
				{
					qVarName=fullyQualifiedName[packageName,varName]
				},
				
				sys[dataObjectDeclarations][qVarName]
			]
		],
		sys[packageDataObjects][packageName]
	];
	
	varDependencies=Union[extractPackageName/@packageVarTypes]/.Null->Nothing;
	varDependencies=DeleteCases[varDependencies,packageName];
	
	(* 2. gather dependencies from package operations *)
	
	operationDependencies=Map[
		Function[opName,
			With[
				{
					qOpName=fullyQualifiedName[packageName,opName]
				},
				
				operationDependentPackages[sys,packageName,qOpName]
			]
		],
		sys[packageOperations][packageName]
	];
	opSpecDependencies=Union@@operationDependencies[[All,1]];
	opBodyDependencies=Union@@operationDependencies[[All,2]];
	
	(* combine all dependencies *)
	
	pkgSpecDependencies=Union[varDependencies,opSpecDependencies];
	pkgBodyDependencies=opBodyDependencies;
	
	{pkgSpecDependencies,pkgBodyDependencies}
]


packageLevelNames[sys_,packageName_String]:=Module[
	{
		typeNames,
		enumElementNames,
		dataObjectNames,
		operationNames
	},
	
	typeNames=sys[packageTypes][packageName];
	
	enumElementNames=Flatten@Map[
		Function[typeName,
			With[
				{
					qTypeName=fullyQualifiedName[packageName,typeName]
				},
				
				If[Head[sys[typeDefinitions][qTypeName]]===FMTSymbols`enum,
					List@@sys[typeDefinitions][qTypeName],
					Nothing
				]
			]
		],
		typeNames
	];
	
	dataObjectNames=sys[packageDataObjects][packageName];
	operationNames=sys[packageOperations][packageName];
	
	Join[typeNames,enumElementNames,dataObjectNames,operationNames]
]


(* Helper, set comment for the named entity and replaces multi-line comments with commentT *)
fmtSetComment[sys_,name_,comment_]:=With[
	{
		sanitizedComment=Switch[Head[comment],
			String,
				FMTSymbols`commentT@@StringSplit[comment,"\n"],
			FMTSymbols`Comment,
				FMTSymbols`commentT@@comment,
			_,
				comment
		]
	},
	
	sys[entityComments][name]=sanitizedComment
]

SetAttributes[fmtSetComment,HoldFirst]


(* String is returned for those entities that are first in their groups and have comments.
	Null is returned for those entities that are first in their groups if no comment
	was provided during declaration/definition.
	None is returned for those entities that are not first in their groups.
*)
fmtGetComment[sys_,name_]:=Lookup[sys[entityComments],name,None]


End[];


EndPackage[];
