(* assignment to local variable with post-condition *)

(* this test is analogous to arith-package-6.wl,
    but operation post-conditions are intentionally resolved without taking local variables into account,
    which means that post-conditions never refer to them from the solver point of view,
    this test is therefore impossible to define *)

(* Note: it cannot be a grammar error, because the unresolved symbol in post-condition might be a meta-variable,
    with meaning supplied by the external context; this is seen in the proofs below *)

sys = newSystem[];
declarePackage[sys, "P`"];
declareOperations[sys, "P`",
    {
        foo[] -> {post -> x < 100} (* x here is not a local variable from foo *)
    }
];
defineOperation[sys, "P`", foo,
    {
        x -> integer
    },
    x = 0
];

proofSt = reportProofStatements[sys, "P`", foo];
Assert[proofSt ===
    {
        {
            {
                "start",
                {!initializedT[x]},
                {}
            },
            {
                1,
                {!initializedT[x], initializedT[Subscript[x, 1]], Subscript[x, 1] == 0},
                {
                    {
                        "Value should fit the target type.",
                        alwaysT[-2147483648 <= Subscript[x, 1] <= 2147483647]
                    }
                }
            },
            {
                "finish",
                {!initializedT[x], initializedT[Subscript[x, 1]], Subscript[x, 1] == 0},
                {
                    {
                        "Variable should be initialized on some path.",
                        possibleT[initializedT[Subscript[x, 1]]]
                    },
                    {
                        "Operation post-conditions.",
                        alwaysT[x < 100]
                    }
                }
            }
        },
        <|x -> Integers, Subscript[x, 1] -> Integers|>
    }
]

proofs = reportProofs[sys];
Assert[proofs ===
    <|"P`" ->
        <|foo ->
            {
                {
                    1,
                    {
                        "Value should fit the target type.", 
                        alwaysT[-2147483648 <= Subscript[x, 1] <= 2147483647]
                    },
                    provenT[]
                },
                {
                    "finish",
                    {
                        "Variable should be initialized on some path.",
                        possibleT[initializedT[Subscript[x, 1]]]
                    },
                    provenT[]
                },
                {
                    "finish",
                    {
                        "Operation post-conditions.",
                        alwaysT[x < 100]
                    },
                    failedT[x -> 100, Subscript[x, 1] -> 0, initializedT[x] -> False, initializedT[Subscript[x, 1]] -> True]
                }
            }
        |>
    |>
]
