(* array access with uninitialized value used as index *)

sys = newSystem[];
declarePackage[sys, "P`"];
defineDataTypes[sys, "P`",
    {
        myArrayType -> array[integer, 10]
    }
];
defineDataObjects[sys, "P`",
    {
        a -> myArrayType
    }
];
declareOperations[sys, "P`",
    {
        foo[]
    }
];
defineOperation[sys, "P`", foo,
    {
        x -> integer
    },
    a[[x]] = 42
];

proofSt = reportProofStatements[sys, "P`", foo];
Assert[proofSt ===
    {
        {
            {
                "start",
                {-2147483648 <= Inactive[Part][P`a, x] <= 2147483647, !initializedT[x]},
                {}
            },
            {
                1,
                {-2147483648 <= Inactive[Part][P`a, x] <= 2147483647, !initializedT[x]},
                {
                    {
                        "Variable should be initialized before use.",
                        alwaysT[initializedT[x]]
                    },
                    {
                        "Array index should be within bounds.",
                        alwaysT[0 <= x <= 9]
                    }
                }
            },
            {
                1,
                {-2147483648 <= Inactive[Part][P`a, x] <= 2147483647, !initializedT[x],
                    Inactive[Part][Subscript[P`a, 1], x] == 42},
                {
                    {
                        "Value should fit the target type.",
                        alwaysT[-2147483648 <= Inactive[Part][Subscript[P`a, 1], x] <= 2147483647]
                    }
                }
            },
            {
                "finish",
                {-2147483648 <= Inactive[Part][P`a, x] <= 2147483647, !initializedT[x],
                    Inactive[Part][Subscript[P`a, 1], x] == 42},
                {
                    {
                        "Variable should be initialized on some path.",
                        possibleT[initializedT[x]]
                    }
                }
            }
        },
        <|Inactive[Part][Subscript[P`a, 1], x] -> Integers, x -> Integers, Inactive[Part][P`a, x] -> Integers|>
    }
]

proofs = reportProofs[sys];
Assert[proofs ===
    <|"P`" ->
        <|foo ->
            {
                {
                    1,
                    {
                        "Variable should be initialized before use.",
                        alwaysT[initializedT[x]]
                    },
                    failedT[x -> -5, Inactive[Part][P`a, x] -> -2147483648, Inactive[Part][Subscript[P`a, 1], x] -> -18, initializedT[x] -> False]
                },
                {
                    1,
                    {
                        "Array index should be within bounds.",
                        alwaysT[0 <= x <= 9]
                    },
                    failedT[x -> -2, Inactive[Part][P`a, x] -> 0, Inactive[Part][Subscript[P`a, 1], x] -> 0, initializedT[x] -> False]
                },
                {
                    1,
                    {
                        "Value should fit the target type.",
                        alwaysT[-2147483648 <= Inactive[Part][Subscript[P`a, 1], x] <= 2147483647]
                    },
                    provenT[]
                },
                {
                    "finish",
                    {
                        "Variable should be initialized on some path.",
                        possibleT[initializedT[x]]
                    },
                    impossibleT[]
                }
            }
        |>
    |>
]
