(* general scheme for resolving loop invariants into primitive statements *)

sys = newSystem[];
declarePackage[sys, "P`"];
declareOperations[sys, "P`",
    {
        foo[]
    }
];
setGhostDomains[sys,
    {
        Element[cond, Booleans],
        Element[inv, Booleans]
    }
];
defineOperation[sys, "P`", foo,
    {
    },
    While[cond && LoopInvariant[inv],
        Null
    ]
];

proofSt = reportProofStatements[sys, "P`", foo];
Assert[proofSt ===
    {
        {
            {
                1,
                {},
                {
                    {
                        "Implicit pre-loop assertion from loop invariants.",
                        alwaysT[inv]
                    }
                }
            },
            {
                2,
                {},
                {
                    {
                        "Branch is reachable.",
                        possibleT[cond]
                    }
                }
            },
            {
                3,
                {inv, cond},
                {}
            },
            {
                4,
                {inv, cond},
                {
                    {
                        "Implicit in-loop assertion from loop invariants.",
                        alwaysT[inv]
                    }
                }
            },
            {
                5,
                {!cond, inv},
                {}
            }
        },
        <||>
    }
]

proofs = reportProofs[sys];
Assert[proofs ===
    <|"P`" ->
        <|foo ->
            {
                {
                    1,
                    {
                        "Implicit pre-loop assertion from loop invariants.",
                        alwaysT[inv]
                    },
                    failedT[cond -> False, inv -> False]
                },
                {
                    2,
                    {
                        "Branch is reachable.",
                        possibleT[cond]
                    },
                    provenT[]
                },
                {
                    4,
                    {
                        "Implicit in-loop assertion from loop invariants.",
                        alwaysT[inv]
                    },
                    provenT[]
                }
            }
        |>
    |>
]
