(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.3' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[    105759,       2503]
NotebookOptionsPosition[    102169,       2441]
NotebookOutlinePosition[    102521,       2457]
CellTagsIndexPosition[    102478,       2454]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Example: average value", "Subsection",
 CellChangeTimes->{{3.7191203670416203`*^9, 3.71912037087922*^9}, {
  3.729317197243264*^9, 
  3.7293172107850647`*^9}},ExpressionUUID->"2ad6339c-9066-4e84-baf7-\
b2bb178f514d"],

Cell["\<\
This notebook presents an example of using pre- and post-conditions with an \
operation for computing average values of two non-negative integers - with a \
bug and its fix.\
\>", "Text",
 CellChangeTimes->{{3.7293172201896644`*^9, 3.7293173407618647`*^9}, {
  3.7293175789972644`*^9, 
  3.7293175806664643`*^9}},ExpressionUUID->"d2c1866e-168c-4d3c-aaa5-\
c71777500a67"],

Cell["Load the FMT library and create a new model:", "Text",
 CellChangeTimes->{{3.7293173460190644`*^9, 
  3.7293173548058643`*^9}},ExpressionUUID->"1f86001b-89fc-4085-9e18-\
937b3f3eee96"],

Cell[BoxData[
 RowBox[{"<<", "FMT`"}]], "Input",
 CellChangeTimes->{{3.7191203731100197`*^9, 3.7191203746388197`*^9}},
 CellLabel->"In[1]:=",ExpressionUUID->"186125f4-f20d-453d-866f-e619333a3d14"],

Cell[BoxData[
 RowBox[{
  RowBox[{"sys", "=", 
   RowBox[{"newSystem", "[", "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}},
 CellLabel->"In[2]:=",ExpressionUUID->"59b4cee9-48d3-4917-a2a5-46f9cd058522"],

Cell["\<\
Declare new package (\[OpenCurlyDoubleQuote]Tools\[CloseCurlyDoubleQuote] is \
a good name here) with a single average operation that accepts two input \
parameters and returns the result via a third output parameter:\
\>", "Text",
 CellChangeTimes->{{3.7293173781560645`*^9, 3.729317448001664*^9}, {
  3.729317494090064*^9, 
  3.7293174949802647`*^9}},ExpressionUUID->"65372988-6111-4abb-b6ed-\
f3af50ebc278"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<Tools`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.7293174539452643`*^9}},
 CellLabel->"In[3]:=",ExpressionUUID->"0672f760-33ed-41cb-9ab9-a45602db82c1"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{"sys", ",", "\"\<Tools`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"average", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "integer", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"in", "[", "integer", "]"}]}], ",", 
       RowBox[{"result", "\[Rule]", 
        RowBox[{"out", "[", "integer", "]"}]}]}], "]"}], "\[Rule]", 
     "\[IndentingNewLine]", 
     RowBox[{"{", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"pre", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
          RowBox[{"y", "\[GreaterEqual]", "0"}]}], "}"}]}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{"post", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"result", "==", 
          FractionBox[
           RowBox[{"x", "+", "y"}], "2"]}], "}"}]}]}], "\[IndentingNewLine]", 
      "}"}]}], "\[IndentingNewLine]", "}"}]}], "\[IndentingNewLine]", 
  "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.729317486602064*^9}, {3.7293175469508643`*^9, 3.729317595083864*^9}, {
   3.7293176275026646`*^9, 3.729317645630864*^9}},
 CellLabel->"In[4]:=",ExpressionUUID->"b2760563-3a51-4235-8948-f2043a2f3aef"],

Cell["The declaration above has two important parts:", "Text",
 CellChangeTimes->{{3.7293177035520644`*^9, 
  3.7293177148006644`*^9}},ExpressionUUID->"19285a06-3fda-4be5-b1cf-\
fe7498d84b6c"],

Cell[CellGroupData[{

Cell[TextData[{
 "Pre-condition stating the ",
 StyleBox["expectation",
  FontSlant->"Italic"],
 " that both input parameters (x and y) are non-negative."
}], "Item",
 CellChangeTimes->{{3.729317717000264*^9, 
  3.729317775104825*^9}},ExpressionUUID->"df5c661d-1118-44b1-9853-\
372751383b17"],

Cell[TextData[{
 "Post-condition stating the ",
 StyleBox["promise",
  FontSlant->"Italic"],
 " that the result will be equal to the average value of x and y."
}], "Item",
 CellChangeTimes->{{3.7293177512191534`*^9, 3.729317780285258*^9}, {
  3.72931782486426*^9, 
  3.729317825332281*^9}},ExpressionUUID->"80993619-026a-44b1-af96-\
77d6e695b596"]
}, Open  ]],

Cell["\<\
In the first approach, an engineer implemented this operation using the easy \
formula copied from the post-condition:\
\>", "Text",
 CellChangeTimes->{{3.7293177953905373`*^9, 3.7293178042079334`*^9}, {
  3.729317868801832*^9, 3.729317900772667*^9}, {3.729402998964053*^9, 
  3.7294030415248537`*^9}},ExpressionUUID->"d341f9ee-c145-43e2-adc5-\
40e2bf70ed0e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Tools`\>\"", ",", "average", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"(*", " ", 
     RowBox[{
      RowBox[{
      "no", " ", "local", " ", "variables", " ", "needed", " ", "here"}], " ",
       "-", " ", 
      RowBox[{"it", " ", "is", " ", "an", " ", "empty", " ", "list"}]}], " ", 
     "*)"}], "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{"result", "=", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"x", "+", "y"}], ")"}], "/", "2"}]}]}], "\[IndentingNewLine]", 
  "]"}]], "Input",
 CellChangeTimes->{{3.719120761628022*^9, 3.7191207935300226`*^9}, {
   3.7191234295074396`*^9, 3.71912345393704*^9}, {3.7191236921238413`*^9, 
   3.7191236925762415`*^9}, {3.7191254587414527`*^9, 
   3.7191254691778526`*^9}, {3.719125627096654*^9, 3.7191256328998537`*^9}, {
   3.7191266533046603`*^9, 3.71912665427186*^9}, {3.7191271307426634`*^9, 
   3.7191271335194635`*^9}, 3.720351102816934*^9, {3.7293179166561794`*^9, 
   3.729317928420108*^9}, {3.729317962886054*^9, 3.7293179782869453`*^9}, {
   3.729318014110153*^9, 3.7293180196035995`*^9}},
 CellLabel->"In[5]:=",ExpressionUUID->"617e63c4-092d-4867-9083-7f58cce70e30"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[5]=",ExpressionUUID->"c55f4af3-9e25-4628-aae9-3dccbc70dffa"]
}, Open  ]],

Cell["\<\
However, running the checks reveals that there is a problem with this code:\
\>", "Text",
 CellChangeTimes->{{3.729318028310747*^9, 3.729318039653093*^9}, {
  3.7294044767035227`*^9, 
  3.729404482182122*^9}},ExpressionUUID->"c3be693d-edf4-4377-a8d5-\
7589b5ef1070"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"runAllChecks", "[", "sys", "]"}]], "Input",
 CellChangeTimes->{{3.7293180416343184`*^9, 3.729318043833946*^9}},
 CellLabel->"In[6]:=",ExpressionUUID->"1ed0f7ae-ef96-4e68-bd58-dc60e8e48b13"],

Cell[BoxData[
 TagBox[GridBox[{
    {
     StyleBox["\<\"Grammar OK.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"No name conflicts.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"Flows OK.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"No call cycles.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"Data aliasing OK.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     TemplateBox[{StyleBox["\"Proofs failed. \"", 
        RGBColor[1, 0, 0], StripOnInput -> False],ButtonBox[
       "\"  Show details  \"", ButtonFunction :> Function[{}, 
          CellPrint[
           ExpressionCell[
            FMTFormatters`formatProofChecks[
             FMTSystem`Private`softwareSystemT[
              Association[
              FMTSystem`Private`packageNames -> {"Tools`"}, 
               FMTSystem`Private`packageTypes -> Association["Tools`" -> {}], 
               FMTSystem`Private`packageDataObjects -> 
               Association["Tools`" -> {}], 
               FMTSystem`Private`packageOperations -> 
               Association["Tools`" -> {$CellContext`average}], 
               FMTSystem`Private`typeDefinitions -> Association[], 
               FMTSystem`Private`dataObjectDeclarations -> Association[], 
               FMTSystem`Private`dataObjectValues -> Association[], 
               FMTSystem`Private`operationDeclarations -> 
               Association[Tools`average -> $CellContext`average[
                   FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                   FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                   
                   FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
               FMTSystem`Private`operationPreconditions -> 
               Association[Tools`average -> {
                   And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
               FMTSystem`Private`operationPostconditions -> 
               Association[
                Tools`average -> {$CellContext`result == 
                   Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
               FMTSystem`Private`operationLocalVariables -> 
               Association[Tools`average -> {}], 
               FMTSystem`Private`operationBodies -> 
               Association[
                Tools`average -> 
                 Hold[$CellContext`result = ($CellContext`x + $CellContext`y)/
                    2]], FMTSystem`Private`entityComments -> 
               Association["Tools`" -> Null, Tools`average -> Null], 
               FMTSystem`Private`systemOptions -> 
               Association[
                FMTOptions`showInteractiveButtons -> True, 
                 FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                 FMTOptions`integerRoundingFunction -> IntegerPart, 
                 FMTOptions`arrayIndexOffset -> 0, 
                 FMTOptions`grammarPreprocessor -> 
                 FMTOptions`defaultPreprocessGrammar, 
                 FMTOptions`formulaFormatter -> StandardForm]]], 
             FMT`Private`result$2386, 
             FMT`Private`currentOptions$2385[FMTOptions`formulaFormatter]], 
            "Output"]]], Appearance -> Automatic, Evaluator -> Automatic, 
        Method -> "Preemptive"]},
      "RowDefault"]}
   },
   DefaultBaseStyle->"Column",
   GridBoxAlignment->{"Columns" -> {{Left}}},
   GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
  "Column"]], "Output",
 CellLabel->"Out[6]=",ExpressionUUID->"1d694789-7b3f-4b5c-9e0f-2d5d08fba03b"]
}, Open  ]],

Cell["\<\
The proofs failed for reasons that can be seen after pressing the \
\[OpenCurlyDoubleQuote]Show details\[CloseCurlyDoubleQuote] button and then \
following with \[OpenCurlyDoubleQuote]show proofs\[CloseCurlyDoubleQuote] or \
by executing the command below:\
\>", "Text",
 CellChangeTimes->{{3.7293180690166693`*^9, 3.7293181093637877`*^9}, {
  3.729318371421295*^9, 
  3.729318418741399*^9}},ExpressionUUID->"2bb5f645-8d33-49ce-8462-\
37f137419e53"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"showProofResults", "[", 
  RowBox[{"sys", ",", "\"\<Tools`\>\"", ",", "average"}], "]"}]], "Input",
 CellChangeTimes->{{3.7293181220633507`*^9, 3.7293181555297794`*^9}, {
  3.7293182668514194`*^9, 3.7293182726556945`*^9}, {3.729318426293847*^9, 
  3.729318464440092*^9}, {3.729318498967114*^9, 3.7293185314505224`*^9}},
 CellLabel->"In[7]:=",ExpressionUUID->"a69aecad-2545-4c11-b56e-1dcd1660c4a1"],

Cell[BoxData[
 TagBox[GridBox[{
    {"\<\"Index\"\>", "\<\"Verification\"\>", "\<\"Result\"\>"},
    {
     TagBox[GridBox[{
        {"\<\"start\"\>"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[
                   Tools`average -> 
                    Hold[$CellContext`result = ($CellContext`x + \
$CellContext`y)/2]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, 
                Private`highlighted -> {"start"}]], "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Operation pre-conditions are reachable.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"possible: \"", 
            RGBColor[0, 0, 1], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              
              GridBox[{{"True"}}, GridBoxAlignment -> {"Columns" -> {{Left}}},
                DefaultBaseStyle -> "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     StyleBox["\<\"OK\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     TagBox[GridBox[{
        {"1"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[
                   Tools`average -> 
                    Hold[$CellContext`result = ($CellContext`x + \
$CellContext`y)/2]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {1}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"x", "+", "y"}], "\[LessEqual]", "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Failed\"\>",
          StripOnInput->False,
          LineColor->RGBColor[1, 0, 0],
          FrontFaceColor->RGBColor[1, 0, 0],
          BackFaceColor->RGBColor[1, 0, 0],
          GraphicsColor->RGBColor[1, 0, 0],
          FontColor->RGBColor[1, 0, 0]]},
        {
         TagBox[
          TooltipBox[
           TemplateBox[{"\"see counter-examples\"",FrameStyle -> None},
            "Highlighted"],
           TagBox[
            GridBox[{{
               RowBox[{"result", "\[Rule]", "0"}]}, {
               RowBox[{"x", "\[Rule]", "2147483646"}]}, {
               RowBox[{"y", "\[Rule]", "2"}]}, {
               RowBox[{
                 
                 TemplateBox[{"result", "\"init\""}, "Superscript", 
                  SyntaxForm -> SuperscriptBox], "\[Rule]", "False"}]}}, 
             GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
             "Column", 
             GridBoxItemSize -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
             GridBoxSpacings -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{1}}}], "Column"]],
          Annotation[#, 
           
           Column[{$CellContext`result -> 0, $CellContext`x -> 
             2147483646, $CellContext`y -> 2, 
             Superscript[$CellContext`result, "init"] -> False}, Spacings -> 
            1], "Tooltip"]& ]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]},
    {
     TagBox[GridBox[{
        {"1"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[
                   Tools`average -> 
                    Hold[$CellContext`result = ($CellContext`x + \
$CellContext`y)/2]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {1}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"IntegerPart", "[", 
                    FractionBox[
                    RowBox[{"x", "+", "y"}], "2"], "]"}], "\[LessEqual]", 
                   "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     StyleBox["\<\"OK\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     TagBox[GridBox[{
        {"1"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[
                   Tools`average -> 
                    Hold[$CellContext`result = ($CellContext`x + \
$CellContext`y)/2]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {1}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Value should fit the target type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   SubscriptBox["result", "1"], "\[LessEqual]", 
                   "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     StyleBox["\<\"OK\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     TagBox[GridBox[{
        {"\<\"finish\"\>"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[
                   Tools`average -> 
                    Hold[$CellContext`result = ($CellContext`x + \
$CellContext`y)/2]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, 
                Private`highlighted -> {"finish"}]], "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Variable should be initialized on some path.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"possible: \"", 
            RGBColor[0, 0, 1], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 TemplateBox[{
                   SubscriptBox["result", "1"], "\"init\""}, "Superscript", 
                  SyntaxForm -> SuperscriptBox]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     StyleBox["\<\"OK\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     TagBox[GridBox[{
        {"\<\"finish\"\>"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[
                   Tools`average -> 
                    Hold[$CellContext`result = ($CellContext`x + \
$CellContext`y)/2]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, 
                Private`highlighted -> {"finish"}]], "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Operation post-conditions.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   SubscriptBox["result", "1"], "\[Equal]", 
                   RowBox[{"IntegerPart", "[", 
                    FractionBox[
                    RowBox[{"x", "+", "y"}], "2"], "]"}]}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     StyleBox["\<\"OK\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]}
   },
   AutoDelete->False,
   GridBoxAlignment->{"Columns" -> {Center, Left, Left}, "Rows" -> {{Top}}},
   GridBoxBackground->{
    "Columns" -> {{None}}, "RowsIndexed" -> {1 -> RGBColor[0.87, 0.94, 1]}},
   GridBoxFrame->{"Columns" -> {{True}}, "Rows" -> {{True}}},
   GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
   GridBoxSpacings->{"Columns" -> {{1}}, "Rows" -> {{1}}}],
  "Grid"]], "Output",
 CellLabel->"Out[7]=",ExpressionUUID->"f32f819a-0da2-4012-aac9-13691ec91aac"]
}, Open  ]],

Cell["\<\
What can be deduced from this list is that one of the verification conditions \
that were generated for this operation has failed. From the counter-examples \
that are seen in the yellow-painted tooltip it appears that there exists a \
possible set of inputs for which the condition is not met:\
\>", "Text",
 CellChangeTimes->{{3.729321011059849*^9, 3.7293211246418486`*^9}, {
  3.7293211678568487`*^9, 3.729321186047449*^9}, {3.729321219218049*^9, 
  3.7293212203110485`*^9}, {3.7293216993498487`*^9, 
  3.729321713921249*^9}},ExpressionUUID->"c9beaced-b6a1-4a8e-8006-\
62d93cc47cf9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[Rule]", "2147483646"}], ",", "\n", 
  RowBox[{"y", "\[Rule]", "2"}]}]], "Input",
 CellChangeTimes->{{3.729321011059849*^9, 3.7293211246418486`*^9}, {
  3.7293211678568487`*^9, 3.729321187669849*^9}, {3.7324420327480783`*^9, 
  3.7324420327530785`*^9}},ExpressionUUID->"c077f595-447e-4243-8ad8-\
b480a451fae3"],

Cell["\<\
In other words, the operation implementation contains an integer overflow in \
the intermediate result of adding two values, before the sum is divided by 2. \
It was not a problem at the level of post-condition, which is considered to \
be a pure mathematical formula, but it is a problem in the implementation, \
which will ultimately execute on a CPU with some real constraints. Existing \
software products were known to have exactly this kind of bug.\
\>", "Text",
 CellChangeTimes->{{3.729321231028249*^9, 3.729321282517249*^9}, {
  3.7294017248890843`*^9, 3.729401746390746*^9}, {3.7294017777651453`*^9, 
  3.7294019515276537`*^9}},ExpressionUUID->"dcccc5de-0541-4b54-9479-\
6fb089c3c805"],

Cell["\<\
The following is a possible fix - the idea is to add half of the distance to \
the lower value, which never creates out-of-range intermediate results (note: \
the fix replaces the already defined operation, hence the warning message):\
\>", "Text",ExpressionUUID->"e0261c5c-e947-4374-b97c-160d7dd673f2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Tools`\>\"", ",", "average", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{"If", "[", 
    RowBox[{
     RowBox[{"x", ">", "y"}], ",", "\[IndentingNewLine]", 
     RowBox[{"result", "=", 
      RowBox[{"y", "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "y"}], ")"}], "/", "2"}]}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"result", "=", 
      RowBox[{"x", "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"y", "-", "x"}], ")"}], "/", "2"}]}]}]}], 
    "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.719120761628022*^9, 3.7191207935300226`*^9}, {
   3.7191234295074396`*^9, 3.71912345393704*^9}, {3.7191236921238413`*^9, 
   3.7191236925762415`*^9}, {3.7191254587414527`*^9, 
   3.7191254691778526`*^9}, {3.719125627096654*^9, 3.7191256328998537`*^9}, {
   3.7191266533046603`*^9, 3.71912665427186*^9}, {3.7191271307426634`*^9, 
   3.7191271335194635`*^9}, 3.720351102816934*^9},
 CellLabel->"In[8]:=",ExpressionUUID->"335ffedc-85c5-4fde-828d-fde18c966add"],

Cell[BoxData[
 TemplateBox[{
  "fmtDefineOp","replace",
   "\"Warning: Replacing existing definition for operation \
\\!\\(\\*RowBox[{\\\"Tools`average\\\"}]\\).\"",2,8,1,17670441982730004284,
   "Local","FMTSystem`fmtDefineOp"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->
  "During evaluation of \
In[8]:=",ExpressionUUID->"7b3468a8-057e-4ccd-9986-0bee0f276542"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[8]=",ExpressionUUID->"56b83d56-0c86-4445-b72d-0d59e49ee86e"]
}, Open  ]],

Cell["The fix can be confirmed by running all checks again:", "Text",
 CellChangeTimes->{{3.729321362898449*^9, 
  3.7293213716354485`*^9}},ExpressionUUID->"5a8353a0-0f09-4e2c-9690-\
409a4d5228aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"runAllChecks", "[", "sys", "]"}]], "Input",
 CellChangeTimes->{{3.7293213733826485`*^9, 3.729321374958249*^9}},
 CellLabel->"In[9]:=",ExpressionUUID->"98117695-9656-4695-9392-0326ada35f24"],

Cell[BoxData[
 TagBox[GridBox[{
    {
     StyleBox["\<\"Grammar OK.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"No name conflicts.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"Flows OK.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"No call cycles.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"Data aliasing OK.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]},
    {
     StyleBox["\<\"Proofs OK.\"\>",
      StripOnInput->False,
      LineColor->RGBColor[0, 1, 0],
      FrontFaceColor->RGBColor[0, 1, 0],
      BackFaceColor->RGBColor[0, 1, 0],
      GraphicsColor->RGBColor[0, 1, 0],
      FontColor->RGBColor[0, 1, 0]]}
   },
   DefaultBaseStyle->"Column",
   GridBoxAlignment->{"Columns" -> {{Left}}},
   GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
  "Column"]], "Output",
 CellLabel->"Out[9]=",ExpressionUUID->"c8b3fe83-d18b-4926-8a0f-ca959eaece20"]
}, Open  ]],

Cell["\<\
It can be interesting (and reassuring) to inspect the list of all proof \
conditions that were generated and then verified for this simple operation. \
The middle column contains \[OpenCurlyDoubleQuote]knowns\
\[CloseCurlyDoubleQuote], which are facts accumulated so far and the right \
column shows the verification conditions that are checked by the prover.\
\>", "Text",
 CellChangeTimes->{{3.7294019897380533`*^9, 3.7294020589778533`*^9}, {
  3.7294020965164537`*^9, 
  3.7294021610470533`*^9}},ExpressionUUID->"74cd9685-8b18-4293-8cb6-\
d1ddd4f92efb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"showProofStatements", "[", 
  RowBox[{"sys", ",", "\"\<Tools`\>\"", ",", "average"}], "]"}]], "Input",
 CellChangeTimes->{{3.7294020628632536`*^9, 3.729402078340453*^9}},
 CellLabel->"In[10]:=",ExpressionUUID->"9b63e3b3-ef2b-4db1-bd36-fb788247ff8a"],

Cell[BoxData[
 TagBox[GridBox[{
    {"\<\"Index\"\>", "\<\"Knowns\"\>", "\<\"Verification\"\>"},
    {
     TagBox[GridBox[{
        {"\<\"start\"\>"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[Tools`average -> Hold[
                    
                    If[$CellContext`x > $CellContext`y, $CellContext`result = \
$CellContext`y + ($CellContext`x - $CellContext`y)/
                    2, $CellContext`result = $CellContext`x + ($CellContext`y - \
$CellContext`x)/2]]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, 
                Private`highlighted -> {"start"}]], "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "x", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "y", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
            RowBox[{"y", "\[GreaterEqual]", "0"}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"!", 
            TemplateBox[{"result","\"init\""},
             "Superscript"]}],
           StandardForm],
          StandardForm,
          Editable->True]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Operation pre-conditions are reachable.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"possible: \"", 
            RGBColor[0, 0, 1], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              
              GridBox[{{"True"}}, GridBoxAlignment -> {"Columns" -> {{Left}}},
                DefaultBaseStyle -> "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]},
    {
     TagBox[GridBox[{
        {"1"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[Tools`average -> Hold[
                    
                    If[$CellContext`x > $CellContext`y, $CellContext`result = \
$CellContext`y + ($CellContext`x - $CellContext`y)/
                    2, $CellContext`result = $CellContext`x + ($CellContext`y - \
$CellContext`x)/2]]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {1}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "x", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "y", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
            RowBox[{"y", "\[GreaterEqual]", "0"}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"!", 
            TemplateBox[{"result","\"init\""},
             "Superscript"]}],
           StandardForm],
          StandardForm,
          Editable->True]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Branch is reachable.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"possible: \"", 
            RGBColor[0, 0, 1], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{"x", ">", "y"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Alternative is reachable.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"possible: \"", 
            RGBColor[0, 0, 1], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{"x", "\[LessEqual]", "y"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]},
    {
     TagBox[GridBox[{
        {"2"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[Tools`average -> Hold[
                    
                    If[$CellContext`x > $CellContext`y, $CellContext`result = \
$CellContext`y + ($CellContext`x - $CellContext`y)/
                    2, $CellContext`result = $CellContext`x + ($CellContext`y - \
$CellContext`x)/2]]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {2}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "x", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "y", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
            RowBox[{"y", "\[GreaterEqual]", "0"}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"!", 
            TemplateBox[{"result","\"init\""},
             "Superscript"]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"x", ">", "y"}],
           StandardForm],
          StandardForm,
          Editable->True]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"-", "y"}], "\[LessEqual]", "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"x", "-", "y"}], "\[LessEqual]", "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"IntegerPart", "[", 
                    FractionBox[
                    RowBox[{"x", "-", "y"}], "2"], "]"}], "\[LessEqual]", 
                   "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"y", "+", 
                    RowBox[{"IntegerPart", "[", 
                    FractionBox[
                    RowBox[{"x", "-", "y"}], "2"], "]"}]}], "\[LessEqual]", 
                   "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]},
    {
     TagBox[GridBox[{
        {"2"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[Tools`average -> Hold[
                    
                    If[$CellContext`x > $CellContext`y, $CellContext`result = \
$CellContext`y + ($CellContext`x - $CellContext`y)/
                    2, $CellContext`result = $CellContext`x + ($CellContext`y - \
$CellContext`x)/2]]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {2}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "x", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "y", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
            RowBox[{"y", "\[GreaterEqual]", "0"}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"!", 
            TemplateBox[{"result","\"init\""},
             "Superscript"]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           TemplateBox[{SubscriptBox["result", "1"],"\"init\""},
            "Superscript"],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            SubscriptBox["result", "1"], "\[Equal]", 
            RowBox[{"y", "+", 
             RowBox[{"IntegerPart", "[", 
              FractionBox[
               RowBox[{"x", "-", "y"}], "2"], "]"}]}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"x", ">", "y"}],
           StandardForm],
          StandardForm,
          Editable->True]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Value should fit the target type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   SubscriptBox["result", "1"], "\[LessEqual]", 
                   "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]},
    {
     TagBox[GridBox[{
        {"3"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[Tools`average -> Hold[
                    
                    If[$CellContext`x > $CellContext`y, $CellContext`result = \
$CellContext`y + ($CellContext`x - $CellContext`y)/
                    2, $CellContext`result = $CellContext`x + ($CellContext`y - \
$CellContext`x)/2]]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {3}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "x", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "y", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
            RowBox[{"y", "\[GreaterEqual]", "0"}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"!", 
            TemplateBox[{"result","\"init\""},
             "Superscript"]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"x", "\[LessEqual]", "y"}],
           StandardForm],
          StandardForm,
          Editable->True]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"-", "x"}], "\[LessEqual]", "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{
                    RowBox[{"-", "x"}], "+", "y"}], "\[LessEqual]", 
                   "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"IntegerPart", "[", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "x"}], "+", "y"}], ")"}]}], "]"}], 
                   "\[LessEqual]", "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Subexpression should fit the base type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   RowBox[{"x", "+", 
                    RowBox[{"IntegerPart", "[", 
                    RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "x"}], "+", "y"}], ")"}]}], "]"}]}], 
                   "\[LessEqual]", "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]},
    {
     TagBox[GridBox[{
        {"3"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[Tools`average -> Hold[
                    
                    If[$CellContext`x > $CellContext`y, $CellContext`result = \
$CellContext`y + ($CellContext`x - $CellContext`y)/
                    2, $CellContext`result = $CellContext`x + ($CellContext`y - \
$CellContext`x)/2]]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, Private`highlighted -> {3}]], 
              "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "x", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "y", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
            RowBox[{"y", "\[GreaterEqual]", "0"}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"!", 
            TemplateBox[{"result","\"init\""},
             "Superscript"]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           TemplateBox[{SubscriptBox["result", "2"],"\"init\""},
            "Superscript"],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            SubscriptBox["result", "2"], "\[Equal]", 
            RowBox[{"x", "+", 
             RowBox[{"IntegerPart", "[", 
              RowBox[{
               FractionBox["1", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "x"}], "+", "y"}], ")"}]}], "]"}]}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"x", "\[LessEqual]", "y"}],
           StandardForm],
          StandardForm,
          Editable->True]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Value should fit the target type.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   RowBox[{"-", "2147483648"}], "\[LessEqual]", 
                   SubscriptBox["result", "2"], "\[LessEqual]", 
                   "2147483647"}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]},
    {
     TagBox[GridBox[{
        {"\<\"finish\"\>"},
        {
         ButtonBox["\<\"show\"\>",
          Appearance->Automatic,
          ButtonFunction:>CreateDocument[{
             TextCell[
              StringJoin["Control flow diagram for ", "Tools`", 
               ToString[$CellContext`average]], "Subsection"], 
             ExpressionCell[
              FMTFormatters`formatControlFlow[
               FMT`reportControlFlow[
                FMTSystem`Private`softwareSystemT[
                 Association[
                 FMTSystem`Private`packageNames -> {"Tools`"}, 
                  FMTSystem`Private`packageTypes -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageDataObjects -> 
                  Association["Tools`" -> {}], 
                  FMTSystem`Private`packageOperations -> 
                  Association["Tools`" -> {$CellContext`average}], 
                  FMTSystem`Private`typeDefinitions -> Association[], 
                  FMTSystem`Private`dataObjectDeclarations -> Association[], 
                  FMTSystem`Private`dataObjectValues -> Association[], 
                  FMTSystem`Private`operationDeclarations -> 
                  Association[Tools`average -> $CellContext`average[
                    FMTSymbols`in[$CellContext`x -> FMTSymbols`integerT], 
                    FMTSymbols`in[$CellContext`y -> FMTSymbols`integerT], 
                    
                    FMTSymbols`out[$CellContext`result -> 
                    FMTSymbols`integerT]]], 
                  FMTSystem`Private`operationPreconditions -> 
                  Association[Tools`average -> {
                    And[$CellContext`x >= 0, $CellContext`y >= 0]}], 
                  FMTSystem`Private`operationPostconditions -> 
                  Association[
                   Tools`average -> {$CellContext`result == 
                    Rational[1, 2] ($CellContext`x + $CellContext`y)}], 
                  FMTSystem`Private`operationLocalVariables -> 
                  Association[Tools`average -> {}], 
                  FMTSystem`Private`operationBodies -> 
                  Association[Tools`average -> Hold[
                    
                    If[$CellContext`x > $CellContext`y, $CellContext`result = \
$CellContext`y + ($CellContext`x - $CellContext`y)/
                    2, $CellContext`result = $CellContext`x + ($CellContext`y - \
$CellContext`x)/2]]], FMTSystem`Private`entityComments -> 
                  Association["Tools`" -> Null, Tools`average -> Null], 
                  FMTSystem`Private`systemOptions -> 
                  Association[
                   FMTOptions`showInteractiveButtons -> True, 
                    FMTOptions`integerBounds -> {-2147483648, 2147483647}, 
                    FMTOptions`integerRoundingFunction -> IntegerPart, 
                    FMTOptions`arrayIndexOffset -> 0, 
                    FMTOptions`grammarPreprocessor -> 
                    FMTOptions`defaultPreprocessGrammar, 
                    FMTOptions`formulaFormatter -> StandardForm]]], 
                "Tools`", $CellContext`average, 
                Private`highlighted -> {"finish"}]], "Output"]}],
          Evaluator->Automatic,
          Method->"Preemptive"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Center}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "x", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"-", "2147483648"}], "\[LessEqual]", "y", "\[LessEqual]", 
            "2147483647"}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[GreaterEqual]", "0"}], "&&", 
            RowBox[{"y", "\[GreaterEqual]", "0"}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{"!", 
            TemplateBox[{"result","\"init\""},
             "Superscript"]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            SubscriptBox["result", "1"], "\[Equal]", 
            RowBox[{"y", "+", 
             RowBox[{"IntegerPart", "[", 
              FractionBox[
               RowBox[{"x", "-", "y"}], "2"], "]"}]}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           TemplateBox[{SubscriptBox["result", "1"],"\"init\""},
            "Superscript"],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            SubscriptBox["result", "2"], "\[Equal]", 
            RowBox[{"x", "+", 
             RowBox[{"IntegerPart", "[", 
              RowBox[{
               FractionBox["1", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "x"}], "+", "y"}], ")"}]}], "]"}]}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           TemplateBox[{SubscriptBox["result", "2"],"\"init\""},
            "Superscript"],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", ">", "y"}], "\[Implies]", 
            RowBox[{
             SubscriptBox["result", "3"], "\[Equal]", 
             SubscriptBox["result", "1"]}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", ">", "y"}], "\[Implies]", 
            RowBox[{
             TemplateBox[{SubscriptBox["result", "3"],"\"init\""},
              "Superscript"], "\[Equal]", 
             TemplateBox[{SubscriptBox["result", "1"],"\"init\""},
              "Superscript"]}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[LessEqual]", "y"}], "\[Implies]", 
            RowBox[{
             SubscriptBox["result", "3"], "\[Equal]", 
             SubscriptBox["result", "2"]}]}],
           StandardForm],
          StandardForm,
          Editable->True]},
        {
         TagBox[
          FormBox[
           RowBox[{
            RowBox[{"x", "\[LessEqual]", "y"}], "\[Implies]", 
            RowBox[{
             TemplateBox[{SubscriptBox["result", "3"],"\"init\""},
              "Superscript"], "\[Equal]", 
             TemplateBox[{SubscriptBox["result", "2"],"\"init\""},
              "Superscript"]}]}],
           StandardForm],
          StandardForm,
          Editable->True]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"], 
     TagBox[GridBox[{
        {
         StyleBox["\<\"Variable should be initialized on some path.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"possible: \"", 
            RGBColor[0, 0, 1], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 TemplateBox[{
                   SubscriptBox["result", "3"], "\"init\""}, "Superscript", 
                  SyntaxForm -> SuperscriptBox]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]},
        {"\<\"\\n\"\>"},
        {
         StyleBox["\<\"Operation post-conditions.\"\>",
          StripOnInput->False,
          FontSlant->Italic]},
        {
         TemplateBox[{StyleBox["\"always: \"", 
            RGBColor[0, 1, 0], Italic, StripOnInput -> False],TagBox[
            FormBox[
             TagBox[
              GridBox[{{
                 RowBox[{
                   SubscriptBox["result", "3"], "\[Equal]", 
                   RowBox[{"IntegerPart", "[", 
                    FractionBox[
                    RowBox[{"x", "+", "y"}], "2"], "]"}]}]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}}, DefaultBaseStyle -> 
               "Column", 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "Column"], StandardForm], StandardForm, Editable -> True]},
          "RowDefault"]}
       },
       DefaultBaseStyle->"Column",
       GridBoxAlignment->{"Columns" -> {{Left}}},
       GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
      "Column"]}
   },
   AutoDelete->False,
   GridBoxAlignment->{"Columns" -> {Center, Left, Left}, "Rows" -> {{Top}}},
   GridBoxBackground->{
    "Columns" -> {{None}}, "RowsIndexed" -> {1 -> RGBColor[0.87, 0.94, 1]}},
   GridBoxFrame->{"Columns" -> {{True}}, "Rows" -> {{True}}},
   GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
   GridBoxSpacings->{"Columns" -> {{1}}, "Rows" -> {{1}}}],
  "Grid"]], "Output",
 CellLabel->"Out[10]=",ExpressionUUID->"71635101-13b4-4f9a-9d8f-41fc21f33539"]
}, Open  ]]
}, Open  ]]
},
WindowSize->{808, 911},
WindowMargins->{{4, Automatic}, {Automatic, 4}},
FrontEndVersion->"11.3 for Mac OS X x86 (32-bit, 64-bit Kernel) (March 5, \
2018)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 222, 4, 54, "Subsection",ExpressionUUID->"2ad6339c-9066-4e84-baf7-b2bb178f514d"],
Cell[805, 28, 380, 8, 58, "Text",ExpressionUUID->"d2c1866e-168c-4d3c-aaa5-c71777500a67"],
Cell[1188, 38, 190, 3, 35, "Text",ExpressionUUID->"1f86001b-89fc-4085-9e18-937b3f3eee96"],
Cell[1381, 43, 196, 3, 30, "Input",ExpressionUUID->"186125f4-f20d-453d-866f-e619333a3d14"],
Cell[1580, 48, 376, 7, 30, "Input",ExpressionUUID->"59b4cee9-48d3-4917-a2a5-46f9cd058522"],
Cell[1959, 57, 420, 8, 58, "Text",ExpressionUUID->"65372988-6111-4abb-b6ed-f3af50ebc278"],
Cell[2382, 67, 432, 7, 30, "Input",ExpressionUUID->"0672f760-33ed-41cb-9ab9-a45602db82c1"],
Cell[2817, 76, 1513, 34, 255, "Input",ExpressionUUID->"b2760563-3a51-4235-8948-f2043a2f3aef"],
Cell[4333, 112, 192, 3, 35, "Text",ExpressionUUID->"19285a06-3fda-4be5-b1cf-fe7498d84b6c"],
Cell[CellGroupData[{
Cell[4550, 119, 292, 8, 32, "Item",ExpressionUUID->"df5c661d-1118-44b1-9853-372751383b17"],
Cell[4845, 129, 347, 9, 32, "Item",ExpressionUUID->"80993619-026a-44b1-af96-77d6e695b596"]
}, Open  ]],
Cell[5207, 141, 369, 7, 58, "Text",ExpressionUUID->"d341f9ee-c145-43e2-adc5-40e2bf70ed0e"],
Cell[CellGroupData[{
Cell[5601, 152, 1291, 26, 157, "Input",ExpressionUUID->"617e63c4-092d-4867-9083-7f58cce70e30"],
Cell[6895, 180, 349, 8, 34, "Output",ExpressionUUID->"c55f4af3-9e25-4628-aae9-3dccbc70dffa"]
}, Open  ]],
Cell[7259, 191, 276, 6, 35, "Text",ExpressionUUID->"c3be693d-edf4-4377-a8d5-7589b5ef1070"],
Cell[CellGroupData[{
Cell[7560, 201, 213, 3, 30, "Input",ExpressionUUID->"1ed0f7ae-ef96-4e68-bd58-dc60e8e48b13"],
Cell[7776, 206, 4533, 101, 135, "Output",ExpressionUUID->"1d694789-7b3f-4b5c-9e0f-2d5d08fba03b"]
}, Open  ]],
Cell[12324, 310, 459, 9, 58, "Text",ExpressionUUID->"2bb5f645-8d33-49ce-8462-37f137419e53"],
Cell[CellGroupData[{
Cell[12808, 323, 421, 6, 30, "Input",ExpressionUUID->"a69aecad-2545-4c11-b56e-1dcd1660c4a1"],
Cell[13232, 331, 29465, 630, 399, "Output",ExpressionUUID->"f32f819a-0da2-4012-aac9-13691ec91aac"]
}, Open  ]],
Cell[42712, 964, 596, 10, 81, "Text",ExpressionUUID->"c9beaced-b6a1-4a8e-8006-62d93cc47cf9"],
Cell[43311, 976, 351, 7, 52, "Input",ExpressionUUID->"c077f595-447e-4243-8ad8-b480a451fae3"],
Cell[43665, 985, 705, 11, 127, "Text",ExpressionUUID->"dcccc5de-0541-4b54-9479-6fb089c3c805"],
Cell[44373, 998, 313, 4, 58, "Text",ExpressionUUID->"e0261c5c-e947-4374-b97c-160d7dd673f2"],
Cell[CellGroupData[{
Cell[44711, 1006, 1224, 27, 199, "Input",ExpressionUUID->"335ffedc-85c5-4fde-828d-fde18c966add"],
Cell[45938, 1035, 375, 9, 24, "Message",ExpressionUUID->"7b3468a8-057e-4ccd-9986-0bee0f276542"],
Cell[46316, 1046, 349, 8, 34, "Output",ExpressionUUID->"56b83d56-0c86-4445-b72d-0d59e49ee86e"]
}, Open  ]],
Cell[46680, 1057, 197, 3, 35, "Text",ExpressionUUID->"5a8353a0-0f09-4e2c-9690-409a4d5228aa"],
Cell[CellGroupData[{
Cell[46902, 1064, 213, 3, 30, "Input",ExpressionUUID->"98117695-9656-4695-9392-0326ada35f24"],
Cell[47118, 1069, 1895, 55, 121, "Output",ExpressionUUID->"c8b3fe83-d18b-4926-8a0f-ca959eaece20"]
}, Open  ]],
Cell[49028, 1127, 566, 10, 81, "Text",ExpressionUUID->"74cd9685-8b18-4293-8cb6-d1ddd4f92efb"],
Cell[CellGroupData[{
Cell[49619, 1141, 273, 4, 30, "Input",ExpressionUUID->"9b63e3b3-ef2b-4db1-bd36-fb788247ff8a"],
Cell[49895, 1147, 52246, 1290, 1603, "Output",ExpressionUUID->"71635101-13b4-4f9a-9d8f-41fc21f33539"]
}, Open  ]]
}, Open  ]]
}
]
*)

