(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.3' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[     58184,       1364]
NotebookOptionsPosition[     48986,       1225]
NotebookOutlinePosition[     49340,       1241]
CellTagsIndexPosition[     49297,       1238]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Example: Code Generator", "Subsection",
 CellChangeTimes->{{3.72700820255727*^9, 
  3.72700822301727*^9}},ExpressionUUID->"05bd2374-6763-4b48-9d00-\
b6ad3c43375c"],

Cell["\<\
This notebook contains example definitions to exercise code generation \
features. As such, it does not represent any practical or even reasonable \
software design and exists only to demonstrate how different model \
definitions translate into final program code.\
\>", "Text",
 CellChangeTimes->{{3.7270082243582697`*^9, 3.72700833246267*^9}, {
  3.7270672493990498`*^9, 
  3.7270672517566495`*^9}},ExpressionUUID->"192f22d7-ba0c-4e63-bfa3-\
ef6cd1a12ba5"],

Cell["\<\
In some sense, this notebook can be treated as an interactive test procedure \
for the code generator.\
\>", "Text",
 CellChangeTimes->{{3.7270083383238697`*^9, 
  3.7270083727220697`*^9}},ExpressionUUID->"1a47bcf0-e8d0-4bb4-8c90-\
f4a021590068"],

Cell["\<\
It is not necessary to execute all definitions in this notebook - as long as \
dependencies are fulfilled, only a subset of these definitions can be \
executed if needed, with fewer code files generated at the end.\
\>", "Text",
 CellChangeTimes->{{3.7271673361400003`*^9, 
  3.727167426567*^9}},ExpressionUUID->"b9a993fc-6224-42a0-8e55-ef522a531633"],

Cell[CellGroupData[{

Cell["Initial setup", "Subsubsection",
 CellChangeTimes->{{3.72700838571367*^9, 
  3.7270084021334696`*^9}},ExpressionUUID->"a464d44a-24aa-423c-833f-\
9f57d7eb82d4"],

Cell["Load the toolkit and create an empty model:", "Text",
 CellChangeTimes->{{3.72700844999007*^9, 
  3.72700845865407*^9}},ExpressionUUID->"bc86f5ed-4c0d-4882-8849-\
0b91eb832fe7"],

Cell[BoxData[
 RowBox[{"<<", "FMT`"}]], "Input",
 CellChangeTimes->{{3.7270084609944696`*^9, 3.72700846200947*^9}},
 CellLabel->"In[1]:=",ExpressionUUID->"a4cc503b-7891-4ccb-801e-891d461c559a"],

Cell[BoxData[
 RowBox[{
  RowBox[{"sys", "=", 
   RowBox[{"newSystem", "[", "]"}]}], ";"}]], "Input",
 CellLabel->"In[2]:=",ExpressionUUID->"a3d15deb-5c28-4de2-ab83-bec7ee79585d"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Smallest possible package - empty package", "Subsubsection",
 CellChangeTimes->{{3.72700855049827*^9, 
  3.72700856373347*^9}},ExpressionUUID->"6c12f0e2-191a-4504-8dca-\
6f7efbfd9598"],

Cell["An empty package does not contain any definitions:", "Text",
 CellChangeTimes->{{3.7270085676704698`*^9, 
  3.7270085924172697`*^9}},ExpressionUUID->"19dad156-9720-4acc-a45d-\
c9865a6d9cb6"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<Empty`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.7270086028614697`*^9, 3.72700861037207*^9}, 
   3.727170136333*^9},
 CellLabel->"In[3]:=",ExpressionUUID->"cb3cbfb0-0543-4be2-b390-7cf4630360bf"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simple packages with no dependencies", "Subsubsection",
 CellChangeTimes->{{3.72700864064807*^9, 3.72700866548427*^9}, {
  3.7270092065456696`*^9, 
  3.7270092144986696`*^9}},ExpressionUUID->"a91bf0c9-5347-4b79-be9f-\
5ff6692df981"],

Cell["\<\
Packages can be very simple and can define as little as just types or data \
objects.\
\>", "Text",
 CellChangeTimes->{{3.7270086678708696`*^9, 
  3.72700872065687*^9}},ExpressionUUID->"40a33470-9c71-4f69-a36c-\
6a0e3f72f1f6"],

Cell["Package with just one type defined:", "Text",
 CellChangeTimes->{{3.72700872685727*^9, 3.72700873771127*^9}, {
  3.72700880234787*^9, 
  3.7270088061242695`*^9}},ExpressionUUID->"c0316d1d-8fa7-4d28-87ad-\
e4e0fd5f4a2c"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<OneTypePkg`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.7270087397416697`*^9, 3.72700875956287*^9}, 
   3.727170133645*^9},
 CellLabel->"In[4]:=",ExpressionUUID->"8a7187e9-489d-4078-be64-2afcd65ddcac"],

Cell[BoxData[
 RowBox[{"defineDataTypes", "[", 
  RowBox[{"sys", ",", "\"\<OneTypePkg`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"myIntType", "\[Rule]", 
     RowBox[{"number", "[", 
      RowBox[{"base", "\[Rule]", "integer"}], "]"}]}], "\[IndentingNewLine]", 
    "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.72700876912307*^9, 3.72700879640407*^9}, 
   3.7301976753048115`*^9},
 CellLabel->"In[5]:=",ExpressionUUID->"578f4083-1411-4fe1-a1da-99b07a788987"],

Cell["Package with just one object:", "Text",
 CellChangeTimes->{{3.7270088326446695`*^9, 
  3.72700883848707*^9}},ExpressionUUID->"c88a996c-233e-42f0-a94a-\
a020ed01e86f"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<OneObjectPkg`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.7270088411292696`*^9, 3.7270088535260696`*^9}, 
   3.7271701320769997`*^9},
 CellLabel->"In[6]:=",ExpressionUUID->"21a39a78-4e26-42c0-9089-cb0b90bbf1a8"],

Cell[BoxData[
 RowBox[{"defineDataObjects", "[", 
  RowBox[{"sys", ",", "\"\<OneObjectPkg`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"myVariable", "\[Rule]", "integer"}], "\[IndentingNewLine]", 
    "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270088576804695`*^9, 3.72700891045967*^9}, {
   3.72706656732685*^9, 3.7270665687026496`*^9}, 3.7271701312530003`*^9},
 CellLabel->"In[7]:=",ExpressionUUID->"1b150e7a-0ba6-4c7a-a883-68783b4e8a0a"],

Cell["\<\
Package can also define operations, but they need separate profile \
declaration and body definitions:\
\>", "Text",
 CellChangeTimes->{{3.72700892935807*^9, 3.72700895402567*^9}, {
  3.7270673291282496`*^9, 
  3.7270673405946493`*^9}},ExpressionUUID->"6a6b50f6-a6f8-45c0-90f3-\
a0e8d14c3bec"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<OneOperationPkg`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.72700895534307*^9, 3.7270089646270695`*^9}, 
   3.727170129269*^9},
 CellLabel->"In[8]:=",ExpressionUUID->"ad7ac871-e9e3-4f50-b3b2-23eaf98aa7fd"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{
  "sys", ",", "\"\<OneOperationPkg`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"myEmptyOperation", "[", "]"}], "\[IndentingNewLine]", "}"}]}], 
  "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270090352148695`*^9, 3.7270090661770697`*^9}, 
   3.727170128309*^9},
 CellLabel->"In[9]:=",ExpressionUUID->"ffbd2ea2-0610-48e0-8172-e2fd557ff384"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<OneOperationPkg`\>\"", ",", "myEmptyOperation", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"(*", " ", 
     RowBox[{
      RowBox[{
      "there", " ", "are", " ", "no", " ", "local", " ", "variables", " ", 
       "in", " ", "this", " ", "operation"}], ",", "\[IndentingNewLine]", 
      "\t", 
      RowBox[{"but", " ", "the", " ", "list"}], ",", " ", 
      RowBox[{"even", " ", "if", " ", "empty"}], ",", " ", 
      RowBox[{"is", " ", "still", " ", "needed"}]}], " ", "*)"}], 
    "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{"the", " ", "code", " ", "comments"}], ",", " ", 
     RowBox[{"in", " ", "general"}], ",", " ", 
     RowBox[{
      RowBox[{
      "are", " ", "not", " ", "visible", " ", "to", " ", "FMT", 
       "\[IndentingNewLine]", "and", " ", "for", " ", "that", " ", "reason", 
       " ", "they", " ", "are", " ", "not", " ", "part", " ", "of", " ", 
       "the", " ", "model"}], "\[IndentingNewLine]", "-", " ", 
      "consequenty"}], ",", " ", 
     RowBox[{
     "they", " ", "are", " ", "not", " ", "propagated", " ", "to", " ", "the",
       " ", "final", " ", "generated", " ", "code"}]}], " ", "*)"}], 
   "\[IndentingNewLine]", "Null"}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270090742176695`*^9, 3.72700916983247*^9}, {
  3.7270656739440002`*^9, 3.7270657708821383`*^9}, {3.7270660129252524`*^9, 
  3.727066071440235*^9}, {3.7270673609158497`*^9, 3.7270673810356493`*^9}},
 CellLabel->"In[10]:=",ExpressionUUID->"91689c66-9ef7-4ea8-97c6-9b84e33dca42"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[10]=",ExpressionUUID->"9a655037-0f27-4edf-bf42-e3d652765798"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simple packages with dependencies", "Subsubsection",
 CellChangeTimes->{{3.72700864064807*^9, 3.72700866548427*^9}, {
  3.7270092065456696`*^9, 3.7270092144986696`*^9}, {3.727065799168229*^9, 
  3.7270657994490395`*^9}},ExpressionUUID->"d89ef6f8-4031-499e-b568-\
a95cb12623af"],

Cell["\<\
Package with just one data object of a type defined in a separate package:\
\>", "Text",
 CellChangeTimes->{{3.727065803105581*^9, 
  3.7270658284821577`*^9}},ExpressionUUID->"d5e27bc1-a8a4-455c-b4a4-\
b1700457fff4"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<OneDepObjectPkg`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.7270088411292696`*^9, 3.7270088535260696`*^9}, {
   3.7270658491075516`*^9, 3.7270658494195633`*^9}, 3.727170126053*^9},
 CellLabel->"In[11]:=",ExpressionUUID->"1821f8a2-4f5b-4155-b199-53689c39f1b6"],

Cell[BoxData[
 RowBox[{"defineDataObjects", "[", 
  RowBox[{
  "sys", ",", "\"\<OneDepObjectPkg`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"myVariable", "\[Rule]", "OneTypePkg`myIntType"}], 
    "\[IndentingNewLine]", "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270088576804695`*^9, 3.72700891045967*^9}, {
   3.7270658565354376`*^9, 3.7270658706991835`*^9}, {3.72706657513085*^9, 
   3.7270665762238493`*^9}, 3.727170124829*^9},
 CellLabel->"In[12]:=",ExpressionUUID->"2f873a83-f256-4bab-bebd-3df07b7aa396"],

Cell["\<\
Package with a simple operation referring to a data object from another \
package:\
\>", "Text",
 CellChangeTimes->{{3.7270658853299465`*^9, 3.727065907863572*^9}, 
   3.7270661660436506`*^9},ExpressionUUID->"76218c4d-5da7-4a04-b794-\
8fb128bcc82f"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<OneDataDepOpPkg`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.72700895534307*^9, 3.7270089646270695`*^9}, {
   3.727065916353027*^9, 3.727065916696229*^9}, {3.727066143210703*^9, 
   3.7270661803629436`*^9}, 3.727170123789*^9},
 CellLabel->"In[13]:=",ExpressionUUID->"8158fecd-a9aa-484d-b089-447cc4afe92d"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{
  "sys", ",", "\"\<OneDataDepOpPkg`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"updateDataSomewhere", "[", "]"}], "\[IndentingNewLine]", "}"}]}],
   "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270090352148695`*^9, 3.7270090661770697`*^9}, {
   3.727065918866643*^9, 3.727065961577921*^9}, 3.727066183696766*^9, 
   3.727170122885*^9},
 CellLabel->"In[14]:=",ExpressionUUID->"4bd724b1-c8cf-41e2-beb8-bb615cd68fc0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<OneDataDepOpPkg`\>\"", ",", "updateDataSomewhere", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   RowBox[{"OneObjectPkg`myVariable", "=", "42"}]}], " ", 
  RowBox[{"(*", " ", 
   RowBox[{
   "this", " ", "variable", " ", "is", " ", "in", " ", "a", " ", "different", 
    " ", "package"}], " ", "*)"}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270090742176695`*^9, 3.72700916983247*^9}, {
   3.7270656739440002`*^9, 3.7270657708821383`*^9}, {3.7270659374813657`*^9, 
   3.727065982770858*^9}, {3.7270660843389187`*^9, 3.727066093376377*^9}, 
   3.727066184663972*^9, 3.72706716442725*^9},
 CellLabel->"In[15]:=",ExpressionUUID->"3a97fbc1-4dc3-4b40-9613-9b1cdfa394b1"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[15]=",ExpressionUUID->"a4e922f1-7f3d-4c84-9454-83d7bdd67c2e"]
}, Open  ]],

Cell["\<\
Finally, package with a simple operation calling an operation from another \
package :\
\>", "Text",
 CellChangeTimes->{{3.727066097277402*^9, 
  3.727066124442381*^9}},ExpressionUUID->"fecce6e4-823b-48d7-9428-\
0497e5ac8fc8"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<OneOpDepOpPkg`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.72700895534307*^9, 3.7270089646270695`*^9}, {
   3.727065916353027*^9, 3.727065916696229*^9}, {3.727066143210703*^9, 
   3.7270661957740498`*^9}, 3.727170120677*^9},
 CellLabel->"In[16]:=",ExpressionUUID->"5268662d-8858-4bc6-8467-b4dc9be6ca65"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{"sys", ",", "\"\<OneOpDepOpPkg`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"callOperationSomewhere", "[", "]"}], "\[IndentingNewLine]", 
    "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270090352148695`*^9, 3.7270090661770697`*^9}, {
   3.727065918866643*^9, 3.727065961577921*^9}, {3.727066183696766*^9, 
   3.7270662113206496`*^9}, 3.7271701190290003`*^9},
 CellLabel->"In[17]:=",ExpressionUUID->"e870360e-81b5-49c3-8b07-436d251bedff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<OneOpDepOpPkg`\>\"", ",", "callOperationSomewhere", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   RowBox[{"OneOperationPkg`myEmptyOperation", "[", "]"}]}], " ", 
  RowBox[{"(*", " ", 
   RowBox[{
   "this", " ", "operation", " ", "is", " ", "in", " ", "a", " ", "different",
     " ", "package"}], " ", "*)"}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7270090742176695`*^9, 3.72700916983247*^9}, {
   3.7270656739440002`*^9, 3.7270657708821383`*^9}, {3.7270659374813657`*^9, 
   3.727065982770858*^9}, {3.7270660843389187`*^9, 3.727066093376377*^9}, {
   3.727066184663972*^9, 3.72706624038785*^9}, {3.7277757729697533`*^9, 
   3.7277757788051157`*^9}, {3.728118168104323*^9, 3.728118168665923*^9}, {
   3.7281191641537685`*^9, 3.7281191724104333`*^9}, 3.7281192186351156`*^9, {
   3.7281195733318615`*^9, 3.7281195785422945`*^9}, {3.7281241601133757`*^9, 
   3.728124164933776*^9}, {3.728124607744176*^9, 3.7281246083681755`*^9}},
 CellLabel->"In[18]:=",ExpressionUUID->"759efa02-f7c2-4322-aea2-bd07f998a5dd"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[18]=",ExpressionUUID->"8e470fb1-a804-43ae-9dce-488ed4f94511"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Data types and their operations", "Subsubsection",
 CellChangeTimes->{{3.72700864064807*^9, 3.72700866548427*^9}, {
  3.7270092065456696`*^9, 3.7270092144986696`*^9}, {3.727065799168229*^9, 
  3.7270657994490395`*^9}, {3.727167023367*^9, 
  3.727167031023*^9}},ExpressionUUID->"44645ae2-0e1c-4d6f-a9db-e29254098753"],

Cell["\<\
A package with multiple different types with data objects and operations \
exercising all parameter passing modes :\
\>", "Text",
 CellChangeTimes->{{3.7271670404849997`*^9, 3.727167082337*^9}, {
  3.727167526743*^9, 
  3.7271675417349997`*^9}},ExpressionUUID->"a0a9f5aa-22a2-4571-a815-\
2b48f2ccd57c"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<TypesAndOps`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.727167086426*^9, 3.727167105943*^9}, 3.727170115869*^9},
 
 CellLabel->"In[19]:=",ExpressionUUID->"2314ce70-b0ba-4709-8b64-0b430ae89b55"],

Cell[BoxData[
 RowBox[{"defineDataTypes", "[", 
  RowBox[{"sys", ",", "\"\<TypesAndOps`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"intType", "\[Rule]", 
      RowBox[{"number", "[", 
       RowBox[{"base", "\[Rule]", "integer"}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"enumType", "\[Rule]", 
      RowBox[{"enum", "[", 
       RowBox[{"black", ",", "red", ",", "green", ",", "blue", ",", "white"}],
        "]"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"intArray", "\[Rule]", 
      RowBox[{"array", "[", 
       RowBox[{"integer", ",", "10"}], "]"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"structType", "\[Rule]", 
      RowBox[{"structure", "[", 
       RowBox[{
        RowBox[{"i", "\[Rule]", "integer"}], ",", 
        RowBox[{"b", "\[Rule]", "boolean"}]}], "]"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"structArray", "\[Rule]", 
      RowBox[{"array", "[", 
       RowBox[{"structType", ",", "20"}], "]"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"arrayArray", "\[Rule]", 
      RowBox[{"array", "[", 
       RowBox[{"intArray", ",", "30"}], "]"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"nestedStruct", "\[Rule]", 
      RowBox[{"structure", "[", 
       RowBox[{
        RowBox[{"s", "\[Rule]", "structType"}], ",", 
        RowBox[{"a", "\[Rule]", "intArray"}]}], "]"}]}]}], 
    "\[IndentingNewLine]", "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.727167109579*^9, 3.727167309968*^9}, {3.72716761408*^9, 
   3.727167661487*^9}, 3.727170113533*^9},
 CellLabel->"In[20]:=",ExpressionUUID->"7189c0cc-96ed-4d21-9e97-0f20ee9e67e9"],

Cell[BoxData[
 RowBox[{"defineDataObjects", "[", 
  RowBox[{"sys", ",", "\"\<TypesAndOps`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"b", "\[Rule]", "boolean"}], ",", "\[IndentingNewLine]", 
     RowBox[{"i1", "\[Rule]", "integer"}], ",", "\[IndentingNewLine]", 
     RowBox[{"i2", "\[Rule]", "intType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"e", "\[Rule]", "enumType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"ia", "\[Rule]", "intArray"}], ",", "\[IndentingNewLine]", 
     RowBox[{"s", "\[Rule]", "structType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"sa", "\[Rule]", "structArray"}], ",", "\[IndentingNewLine]", 
     RowBox[{"aa", "\[Rule]", "arrayArray"}], ",", "\[IndentingNewLine]", 
     RowBox[{"ns", "\[Rule]", "nestedStruct"}]}], "\[IndentingNewLine]", 
    "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.727167470354*^9, 3.7271675058310003`*^9}, {
   3.727167548639*^9, 3.727167603951*^9}, {3.727167690727*^9, 
   3.727167715263*^9}, 3.727170110653*^9},
 CellLabel->"In[21]:=",ExpressionUUID->"3586f360-9377-461a-93cb-fe122b6e587b"],

Cell["\<\
Operations use the above types (but not package data objects) as parameters, \
in all available modes: in, out and inOut.\
\>", "Text",
 CellChangeTimes->{{3.727168890138*^9, 
  3.7271689272469997`*^9}},ExpressionUUID->"e6ab7ad2-88a1-41d2-9b43-\
2116a35ff6ec"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{"sys", ",", "\"\<TypesAndOps`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"(*", " ", 
     RowBox[{
     "parameter", " ", "specification", " ", "can", " ", "have", " ", 
      RowBox[{"form", ":", " ", 
       RowBox[{"mode", "[", 
        RowBox[{"name", "\[Rule]", "type"}], "]"}]}]}], " ", "*)"}], 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"opB", "[", 
      RowBox[{
       RowBox[{"in", "[", 
        RowBox[{"x", "->", "boolean"}], "]"}], ",", 
       RowBox[{"out", "[", 
        RowBox[{"y", "->", "boolean"}], "]"}], ",", 
       RowBox[{"inOut", "[", 
        RowBox[{"z", "->", "boolean"}], "]"}]}], "]"}], ",", 
     "\[IndentingNewLine]", "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{
       RowBox[{
       "parameters", " ", "can", " ", "be", " ", "also", " ", "specified", 
        " ", 
        RowBox[{"as", ":", " ", "name"}]}], "\[Rule]", 
       RowBox[{"mode", "[", "type", "]"}]}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{"opI", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "integer", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"out", "[", "integer", "]"}]}], ",", 
       RowBox[{"z", "\[Rule]", 
        RowBox[{"inOut", "[", "integer", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opE", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "enumType", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"out", "[", "enumType", "]"}]}], ",", 
       RowBox[{"z", "\[Rule]", 
        RowBox[{"inOut", "[", "enumType", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opIa", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "intArray", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"out", "[", "intArray", "]"}]}], ",", 
       RowBox[{"z", "\[Rule]", 
        RowBox[{"inOut", "[", "intArray", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opS", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "structType", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"out", "[", "structType", "]"}]}], ",", 
       RowBox[{"z", "\[Rule]", 
        RowBox[{"inOut", "[", "structType", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opSa", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "structArray", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"out", "[", "structArray", "]"}]}], ",", 
       RowBox[{"z", "\[Rule]", 
        RowBox[{"inOut", "[", "structArray", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opAa", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "arrayArray", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"out", "[", "arrayArray", "]"}]}], ",", 
       RowBox[{"z", "\[Rule]", 
        RowBox[{"inOut", "[", "arrayArray", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opNs", "[", 
      RowBox[{
       RowBox[{"x", "\[Rule]", 
        RowBox[{"in", "[", "nestedStruct", "]"}]}], ",", 
       RowBox[{"y", "\[Rule]", 
        RowBox[{"out", "[", "nestedStruct", "]"}]}], ",", 
       RowBox[{"z", "\[Rule]", 
        RowBox[{"inOut", "[", "nestedStruct", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", "\[IndentingNewLine]", 
     RowBox[{"logicOps", "[", "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"enumOps", "[", "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"arithmeticOps", "[", "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"compositeObjects", "[", "]"}]}], "\[IndentingNewLine]", "}"}]}],
   "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.727168830823*^9, 3.727168867086*^9}, {
   3.7271689312320004`*^9, 3.72716896851*^9}, {3.727169042462*^9, 
   3.727169068242*^9}, {3.727169924462*^9, 3.727170107142*^9}, 
   3.727170158984*^9, {3.727172786909499*^9, 3.727172797997608*^9}, {
   3.727412505155748*^9, 3.7274125179635944`*^9}, {3.727412651637806*^9, 
   3.7274126546818647`*^9}},
 CellLabel->"In[22]:=",ExpressionUUID->"d42d5514-6bb8-413d-bce3-821c17c53df8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<TypesAndOps`\>\"", ",", "opB", ",", "\[IndentingNewLine]", 
   
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"y", "=", "z"}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", "x"}]}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.727168971075*^9, 3.7271690242530003`*^9}, 
   3.727170209709*^9, {3.727170396041*^9, 3.727170398741*^9}},
 CellLabel->"In[23]:=",ExpressionUUID->"1795441e-0a5e-4ebb-b293-d280e6608e93"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[23]=",ExpressionUUID->"865b7a72-6714-4f38-a664-41869ecafaa0"]
}, Open  ]],

Cell["\<\
All the other test operations would have the same form as the one above for \
the boolean type, but it is easier to generate them automatically \
(metaprogramming):\
\>", "Text",
 CellChangeTimes->{{3.7271702128*^9, 3.727170249525*^9}, {3.727170368973*^9, 
  3.727170369108*^9}, {3.7271727001708264`*^9, 
  3.727172705170326*^9}},ExpressionUUID->"d40b654b-556d-4730-9bd6-\
70ec9633814d"],

Cell[BoxData[
 RowBox[{"Scan", "[", "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"Function", "[", 
    RowBox[{"opName", ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"defineOperation", "[", 
       RowBox[{"sys", ",", "\"\<TypesAndOps`\>\"", ",", "opName", ",", 
        RowBox[{"{", "}"}], ",", 
        RowBox[{
         RowBox[{"y", "=", "z"}], ";", 
         RowBox[{"z", "=", "x"}]}]}], "]"}], ";"}]}], "\[IndentingNewLine]", 
    "]"}], ",", "\[IndentingNewLine]", 
   RowBox[{"{", 
    RowBox[{
    "opI", ",", "opE", ",", "opIa", ",", "opS", ",", "opSa", ",", "opAa", ",",
      "opNs"}], "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7271702527349997`*^9, 3.727170373902*^9}, {
  3.7271704050179996`*^9, 3.7271704306210003`*^9}},
 CellLabel->"In[24]:=",ExpressionUUID->"087efacc-0c08-4257-9465-c09c69067193"],

Cell["Logic operations with the boolean type:", "Text",
 CellChangeTimes->{{3.7271728438181896`*^9, 3.7271728499047985`*^9}, 
   3.727412664603656*^9},ExpressionUUID->"8a59b64f-b723-4f0d-9d98-\
2d9e776a35fc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<TypesAndOps`\>\"", ",", "logicOps", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"x", "\[Rule]", "boolean"}], ",", "\[IndentingNewLine]", 
     RowBox[{"y", "\[Rule]", "boolean"}], ",", "\[IndentingNewLine]", 
     RowBox[{"z", "\[Rule]", "boolean"}]}], "\[IndentingNewLine]", "}"}], ",",
    "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"x", "=", "True"}], ";", "\[IndentingNewLine]", 
    RowBox[{"y", "=", "False"}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "&&", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "||", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"!", "x"}]}], ";", "\[IndentingNewLine]", "\[IndentingNewLine]", 
    
    RowBox[{"z", "=", 
     RowBox[{"x", "\[Equal]", "True"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "\[Equal]", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "\[NotEqual]", "y"}]}]}]}], "\[IndentingNewLine]", 
  "]"}]], "Input",
 CellChangeTimes->{{3.727172851760984*^9, 3.7271729255690002`*^9}, 
   3.7274120250912223`*^9, {3.727412360648506*^9, 3.7274124004335327`*^9}, 
   3.727412492843111*^9, 3.7274127100531006`*^9, {3.7274128676582336`*^9, 
   3.7274128805141397`*^9}},
 CellLabel->"In[25]:=",ExpressionUUID->"748f1be7-c446-415b-ae2e-38b8d1bf928b"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[25]=",ExpressionUUID->"85c598bb-7a17-44e6-984e-5882a70dd989"]
}, Open  ]],

Cell["Enum operations:", "Text",
 CellChangeTimes->{{3.727412666600494*^9, 
  3.727412669595752*^9}},ExpressionUUID->"70d583f9-8b4c-48b9-acbc-\
270b7097efdc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<TypesAndOps`\>\"", ",", "enumOps", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"x", "\[Rule]", "enumType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"y", "\[Rule]", "enumType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"z", "\[Rule]", "boolean"}]}], "\[IndentingNewLine]", "}"}], ",",
    "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"x", "=", "black"}], ";", "\[IndentingNewLine]", 
    RowBox[{"y", "=", "white"}], ";", "\[IndentingNewLine]", 
    "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "\[Equal]", "black"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "\[Equal]", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "!=", "y"}]}]}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.727172851760984*^9, 3.7271729255690002`*^9}, 
   3.7274120250912223`*^9, {3.727412360648506*^9, 3.7274124004335327`*^9}, 
   3.727412492843111*^9, {3.727412678271493*^9, 3.727412706634657*^9}, {
   3.727412849873918*^9, 3.727412853587575*^9}},
 CellLabel->"In[26]:=",ExpressionUUID->"0b287b16-a044-4f6f-a928-0548eca90fd0"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[26]=",ExpressionUUID->"0fd7289e-7226-4277-b7b2-7cf64496f7fa"]
}, Open  ]],

Cell["Arithmetic and relational operations with integer type:", "Text",
 CellChangeTimes->{{3.727412526309755*^9, 
  3.727412538917797*^9}},ExpressionUUID->"627e99c2-9279-458a-8a71-\
5891c3d27a59"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<TypesAndOps`\>\"", ",", "arithmeticOps", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"x", "\[Rule]", "intType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"y", "\[Rule]", "intType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"z", "\[Rule]", "intType"}], ",", "\[IndentingNewLine]", 
     RowBox[{"bl", "\[Rule]", "boolean"}]}], "\[IndentingNewLine]", "}"}], 
   ",", "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"x", "=", "42"}], ";", "\[IndentingNewLine]", 
    RowBox[{"y", "=", 
     RowBox[{"-", "42"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "+", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "-", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "*", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"x", "/", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"z", "=", 
     RowBox[{"-", "x"}]}], ";", "\[IndentingNewLine]", "\[IndentingNewLine]", 
    
    RowBox[{"bl", "=", 
     RowBox[{"x", "\[Equal]", "7"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"bl", "=", 
     RowBox[{"x", "\[Equal]", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"bl", "=", 
     RowBox[{"x", "\[NotEqual]", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"bl", "=", 
     RowBox[{"x", ">", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"bl", "=", 
     RowBox[{"x", "\[GreaterEqual]", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"bl", "=", 
     RowBox[{"x", "<", "y"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{"bl", "=", 
     RowBox[{"x", "\[LessEqual]", "y"}]}], ";"}]}], "\[IndentingNewLine]", 
  "]"}]], "Input",
 CellLabel->"In[28]:=",ExpressionUUID->"3a00c166-ce62-4de5-960e-5564148a5ce0"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[28]=",ExpressionUUID->"22163997-25ae-4176-93e3-a71a4942e305"]
}, Open  ]],

Cell["Structure and array access operations:", "Text",
 CellChangeTimes->{{3.7274125441750984`*^9, 
  3.7274125518192453`*^9}},ExpressionUUID->"01400c2e-f767-4346-8091-\
bd34903458b5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<TypesAndOps`\>\"", ",", "compositeObjects", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"(*", " ", 
     RowBox[{
      RowBox[{"no", " ", "local", " ", "vars", " ", "here"}], ",", " ", 
      RowBox[{
      "operations", " ", "refer", " ", "to", " ", "package", " ", 
       "variables"}]}], " ", "*)"}], "\[IndentingNewLine]", "}"}], ",", 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"ia", "[", 
      RowBox[{"[", "0", "]"}], "]"}], "=", "42"}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"s", ".", "i"}], "=", "42"}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"sa", "[", 
       RowBox[{"[", "0", "]"}], "]"}], ".", "i"}], "=", "42"}], ";", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"aa", "[", 
       RowBox[{"[", "0", "]"}], "]"}], "[", 
      RowBox[{"[", "0", "]"}], "]"}], "=", "42"}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"ns", ".", "s", ".", "i"}], "=", "42"}], ";", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"ns", ".", 
      RowBox[{"a", "[", 
       RowBox[{"[", "0", "]"}], "]"}]}], "=", "42"}]}]}], 
  "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.727172851760984*^9, 3.7271729255690002`*^9}, {
   3.7274120438406253`*^9, 3.727412088107281*^9}, {3.72741228837512*^9, 
   3.727412288687132*^9}, {3.7274123592892537`*^9, 3.7274123979842863`*^9}, {
   3.727412428626275*^9, 3.727412449485877*^9}, 3.727412495341159*^9, 
   3.7274126080631676`*^9, {3.727412906552538*^9, 3.72741294862712*^9}, {
   3.727413833360767*^9, 3.7274138564849668`*^9}, {3.727414121707967*^9, 
   3.727414133240367*^9}, {3.727414261353367*^9, 3.727414265799367*^9}, {
   3.727414357294767*^9, 3.7274144341825666`*^9}},
 CellLabel->"In[29]:=",ExpressionUUID->"064e84a6-eed8-4afc-bee5-fbb565d0e73e"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[29]=",ExpressionUUID->"1297d8a7-5a66-493d-a65f-58527bef8164"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Control flow statements", "Subsubsection",
 CellChangeTimes->{{3.72700864064807*^9, 3.72700866548427*^9}, {
  3.7270092065456696`*^9, 3.7270092144986696`*^9}, {3.727065799168229*^9, 
  3.7270657994490395`*^9}, {3.727167023367*^9, 3.727167031023*^9}, {
  3.728029434093382*^9, 
  3.7280294469597816`*^9}},ExpressionUUID->"abfdc63a-5c9d-4867-890a-\
52727b198d9d"],

Cell["\<\
A single package with operations exercising all available control flow \
statements.\
\>", "Text",
 CellChangeTimes->{{3.728029453527382*^9, 
  3.7280294814869823`*^9}},ExpressionUUID->"9262cbfa-ad94-4207-9e54-\
35d9f388c547"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<Control`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.7280295120173817`*^9, 3.728029521788982*^9}},
 CellLabel->"In[30]:=",ExpressionUUID->"7c774c02-460e-4f41-bc33-536f7f0e0eae"],

Cell[BoxData[
 RowBox[{"defineDataTypes", "[", 
  RowBox[{"sys", ",", "\"\<Control`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"enumType", "\[Rule]", 
     RowBox[{"enum", "[", 
      RowBox[{"black", ",", "red", ",", "green", ",", "blue", ",", "white"}], 
      "]"}]}], "\[IndentingNewLine]", "}"}]}], "\[IndentingNewLine]", 
  "]"}]], "Input",
 CellChangeTimes->{{3.7280296777797823`*^9, 3.728029689063582*^9}},
 CellLabel->"In[31]:=",ExpressionUUID->"f41aec5d-deb4-4a0c-8a3e-505ac330a0f4"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{"sys", ",", "\"\<Control`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"opIf", "[", 
      RowBox[{
       RowBox[{"b", "\[Rule]", 
        RowBox[{"in", "[", "boolean", "]"}]}], ",", 
       RowBox[{"x", "\[Rule]", 
        RowBox[{"out", "[", "integer", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opWhile", "[", 
      RowBox[{
       RowBox[{"c", "\[Rule]", 
        RowBox[{"in", "[", "integer", "]"}]}], ",", 
       RowBox[{"x", "\[Rule]", 
        RowBox[{"inOut", "[", "integer", "]"}]}]}], "]"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"opSwitch", "[", 
      RowBox[{
       RowBox[{"e", "\[Rule]", 
        RowBox[{"in", "[", "enumType", "]"}]}], ",", 
       RowBox[{"x", "\[Rule]", 
        RowBox[{"out", "[", "integer", "]"}]}]}], "]"}]}], 
    "\[IndentingNewLine]", "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.728029526427182*^9, 3.7280296383541822`*^9}, {
  3.728029696745782*^9, 3.7280297063895817`*^9}},
 CellLabel->"In[32]:=",ExpressionUUID->"77c66bb9-374c-4400-81ab-7ebfdf62c6b3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Control`\>\"", ",", "opIf", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"no", " ", "else", " ", 
     RowBox[{"branch", ":"}]}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"If", "[", 
     RowBox[{"b", ",", "\[IndentingNewLine]", 
      RowBox[{"x", "=", "7"}]}], "\[IndentingNewLine]", "]"}], ";", 
    "\[IndentingNewLine]", "\[IndentingNewLine]", 
    RowBox[{"(*", " ", 
     RowBox[{"else", " ", "branch", " ", 
      RowBox[{"present", ":"}]}], " ", "*)"}], "\[IndentingNewLine]", 
    RowBox[{"If", "[", 
     RowBox[{"b", ",", "\[IndentingNewLine]", 
      RowBox[{"x", "=", "8"}], ",", "\[IndentingNewLine]", 
      RowBox[{"x", "=", "42"}]}], "\[IndentingNewLine]", "]"}]}]}], 
  "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.728029709291182*^9, 3.7280297779359818`*^9}, 
   3.728029808682982*^9},
 CellLabel->"In[33]:=",ExpressionUUID->"b3a4f768-9b2c-43b1-950c-ba5e2f0fbaab"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[33]=",ExpressionUUID->"60b8cf9d-a83d-49e5-a715-d5c11e7b292e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Control`\>\"", ",", "opWhile", ",", "\[IndentingNewLine]", 
   
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"temp", "\[Rule]", "integer"}], "\[IndentingNewLine]", "}"}], ",",
    "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"temp", "=", "c"}], ";", "\[IndentingNewLine]", 
    RowBox[{"While", "[", 
     RowBox[{
      RowBox[{"temp", ">", "0"}], ",", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"x", "=", 
        RowBox[{"x", "+", "1"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{"temp", "=", 
        RowBox[{"temp", "-", "1"}]}]}]}], "\[IndentingNewLine]", "]"}]}]}], 
  "\[IndentingNewLine]", "]"}]], "Input",
 CellLabel->"In[35]:=",ExpressionUUID->"664cd531-b2ba-4f20-8cd9-e5197a36186e"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[35]=",ExpressionUUID->"ee090ac4-f081-4be8-8cd5-599dfe5ab8a6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Control`\>\"", ",", "opSwitch", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{"Switch", "[", 
    RowBox[{"e", ",", "\[IndentingNewLine]", "black", ",", " ", 
     RowBox[{"(*", " ", 
      RowBox[{"single", " ", "label"}], " ", "*)"}], "\[IndentingNewLine]", 
     "\t", 
     RowBox[{"x", "=", "7"}], ",", "\[IndentingNewLine]", 
     RowBox[{"red", "|", "green"}], ",", " ", 
     RowBox[{"(*", " ", "alternatives", " ", "*)"}], "\[IndentingNewLine]", 
     "\t", 
     RowBox[{"x", "=", "42"}], ",", "\[IndentingNewLine]", "_", ",", " ", 
     RowBox[{"(*", " ", 
      RowBox[{"default", " ", "branch"}], " ", "*)"}], "\[IndentingNewLine]", 
     "\t", 
     RowBox[{"x", "=", "123"}]}], "\[IndentingNewLine]", "]"}]}], 
  "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.7280298633713818`*^9, 3.728029966218382*^9}},
 CellLabel->"In[36]:=",ExpressionUUID->"43788f0b-2014-4bcf-a97b-6b16da462bcf"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[36]=",ExpressionUUID->"b4eaceea-d719-4fd9-a7ce-a65e5f4e7adc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Generating code", "Subsubsection",
 CellChangeTimes->{{3.72700864064807*^9, 3.72700866548427*^9}, {
  3.7270092065456696`*^9, 3.7270092144986696`*^9}, {3.727065799168229*^9, 
  3.7270657994490395`*^9}, {3.7270666142578497`*^9, 
  3.7270666173430495`*^9}},ExpressionUUID->"4fd4a635-5a77-4890-bf79-\
fff326cb0abd"],

Cell["\<\
Code can be generated and immediately saved in \
files.\[LineSeparator]Un-comment below example commands to actually have them \
executed.\
\>", "Text",
 CellChangeTimes->{{3.7270666201354494`*^9, 3.72706664548625*^9}, {
  3.7270667206230497`*^9, 
  3.7270667748298492`*^9}},ExpressionUUID->"b5f6aef0-3cf9-4120-a3f1-\
11d0439c3752"],

Cell["\<\
Ada code for all packages, saved in the current directory (note: the current \
directory is not necessarily the one where the notebook is located):\
\>", "Text",ExpressionUUID->"4a47f72c-4d7b-4d85-916a-302572df4ada"],

Cell[BoxData[
 RowBox[{"(*", "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"saveCode", "[", 
    RowBox[{"sys", ",", "\"\<Ada\>\""}], "]"}], ";"}], "\[IndentingNewLine]", 
  "*)"}]], "Input",
 CellChangeTimes->{{3.7270666473622494`*^9, 3.7270666868184495`*^9}, 
   3.7270667392322493`*^9, {3.7270669599038496`*^9, 
   3.72706696449325*^9}},ExpressionUUID->"f732a42e-2ec3-485d-99ab-\
e30614d77f30"],

Cell["\<\
C code for a single selected package, saved in the current directory :\
\>", "Text",
 CellChangeTimes->{{3.7270667859218493`*^9, 3.7270668192398496`*^9}, {
  3.7270669906714497`*^9, 
  3.7270669910770493`*^9}},ExpressionUUID->"c2d41c7e-da50-45be-9600-\
759439d01389"],

Cell[BoxData[
 RowBox[{"(*", "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"saveCode", "[", 
    RowBox[{"sys", ",", "\"\<C\>\"", ",", 
     RowBox[{"packageNames", "->", "\"\<OneOperationPkg`\>\""}]}], "]"}], 
   ";"}], "\[IndentingNewLine]", "*)"}]], "Input",
 CellChangeTimes->{{3.7270666618032494`*^9, 3.7270667178452497`*^9}, {
  3.7270667840498495`*^9, 3.7270667843306494`*^9}, {3.72706696717645*^9, 
  3.72706696973485*^9}},ExpressionUUID->"f42db428-cb36-49e9-ba3f-\
243128822043"],

Cell["\<\
C++ code for two selected packages, saved in explicitly named directory :\
\>", "Text",
 CellChangeTimes->{{3.7270667859218493`*^9, 3.7270668390512495`*^9}, {
  3.7270669966346493`*^9, 
  3.7270669967906494`*^9}},ExpressionUUID->"1a323323-b933-4540-ba13-\
282deb436069"],

Cell[BoxData[
 RowBox[{"(*", "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"saveCode", "[", 
    RowBox[{"sys", ",", "\"\<C++\>\"", ",", "\[IndentingNewLine]", 
     RowBox[{"packageNames", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<OneOperationPkg`\>\"", ",", "\"\<OneOpDepOpPkg`\>\""}], 
       "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
     "destinationDirectory", "\[Rule]", "\"\<~/my-project-files\>\""}]}], 
    "]"}], ";"}], "\[IndentingNewLine]", "*)"}]], "Input",
 CellChangeTimes->{{3.7270668488666496`*^9, 
  3.7270669751032495`*^9}},ExpressionUUID->"34482232-1160-449d-a9da-\
424531f2461e"],

Cell["\<\
Code can be also generated without saving to files and used as string values \
for further processing. Here the showCode function presents the code as a \
framed grid with columns for file names and their contents and with file \
names that do not have any content omitted:\
\>", "Text",
 CellChangeTimes->{{3.7270670210168495`*^9, 3.7270670664388494`*^9}, {
  3.7270675154182496`*^9, 3.7270675581518497`*^9}, {3.727072031389544*^9, 
  3.7270720873205442`*^9}, {3.7270721649715443`*^9, 3.727072262485544*^9}, {
  3.732530990869012*^9, 
  3.732530993508999*^9}},ExpressionUUID->"a732cfb9-3a0e-4be1-9095-\
d45e008f0f80"],

Cell[BoxData[
 RowBox[{
  RowBox[{"showCode", "[", "code_Association", "]"}], ":=", 
  RowBox[{"Grid", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Join", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"\"\<File name\>\"", ",", "Bold"}], "]"}], ",", 
         RowBox[{"Style", "[", 
          RowBox[{"\"\<File content\>\"", ",", "Bold"}], "]"}]}], "}"}], 
       "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{"List", "@@@", 
       RowBox[{"DeleteCases", "[", 
        RowBox[{
         RowBox[{"Normal", "[", "code", "]"}], ",", 
         RowBox[{"fileName_", "\[Rule]", "Null"}]}], "]"}]}]}], 
     "\[IndentingNewLine]", "]"}], ",", "\[IndentingNewLine]", 
    RowBox[{"Alignment", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"Left", ",", "Baseline"}], "}"}]}], ",", "\[IndentingNewLine]", 
    
    RowBox[{"Frame", "\[Rule]", "All"}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 CellChangeTimes->{{3.727072083242544*^9, 3.727072113792544*^9}, {
  3.727072286060544*^9, 3.727072380853544*^9}, {3.727072508352544*^9, 
  3.727072508692544*^9}},
 CellLabel->"In[37]:=",ExpressionUUID->"f8e95ed6-5097-42db-8037-47cf1256ae9f"],

Cell["Ada code for all packages:", "Text",
 CellChangeTimes->{{3.7270721441625443`*^9, 
  3.727072148704544*^9}},ExpressionUUID->"839b4617-f6ac-42a7-9c51-\
d859a1c5251b"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"codeAda", "=", 
   RowBox[{"generateCode", "[", 
    RowBox[{"sys", ",", "\"\<Ada\>\""}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"showCode", "[", "codeAda", "]"}]}], "Input",
 CellChangeTimes->{{3.7270671024212494`*^9, 3.72706711906485*^9}, {
  3.727072123929544*^9, 3.727072127126544*^9}},
 CellLabel->"In[38]:=",ExpressionUUID->"e57bd22c-5844-495e-98ec-6d6f235ef0ac"],

Cell["C code for all packages:", "Text",
 CellChangeTimes->{{3.7270721441625443`*^9, 3.727072148704544*^9}, 
   3.7270724691045437`*^9},ExpressionUUID->"827b63ca-b9e4-4b86-ad55-\
54b70303d2c2"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"codeC", "=", 
   RowBox[{"generateCode", "[", 
    RowBox[{"sys", ",", "\"\<C\>\""}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"showCode", "[", "codeC", "]"}]}], "Input",
 CellChangeTimes->{{3.7270671024212494`*^9, 3.72706711906485*^9}, {
   3.727072123929544*^9, 3.727072127126544*^9}, {3.727072471261544*^9, 
   3.727072477474544*^9}, {3.7276960602171564`*^9, 3.7276960892312546`*^9}, 
   3.72769650404161*^9},
 CellLabel->"In[40]:=",ExpressionUUID->"654490fd-e8df-4a61-b5be-88dbac044493"],

Cell["C++ code for all packages:", "Text",
 CellChangeTimes->{{3.7270721441625443`*^9, 3.727072148704544*^9}, 
   3.7270724691045437`*^9, {3.727072544474544*^9, 
   3.727072544633544*^9}},ExpressionUUID->"e9b50439-fda8-4d5e-aed3-\
465635432ecf"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"codeCpp", "=", 
   RowBox[{"generateCode", "[", 
    RowBox[{"sys", ",", "\"\<C++\>\""}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"showCode", "[", "codeCpp", "]"}]}], "Input",
 CellChangeTimes->{{3.7270671024212494`*^9, 3.72706711906485*^9}, {
  3.727072123929544*^9, 3.727072127126544*^9}, {3.727072471261544*^9, 
  3.727072477474544*^9}, {3.727072547018544*^9, 3.727072550783544*^9}, {
  3.7276965068496094`*^9, 3.72769652356761*^9}},
 CellLabel->"In[42]:=",ExpressionUUID->"0ee98fa0-326f-44d7-aaf5-9a2e853a8383"]
}, Open  ]]
}, Open  ]]
},
WindowSize->{808, 911},
WindowMargins->{{12, Automatic}, {Automatic, 24}},
FrontEndVersion->"11.3 for Mac OS X x86 (32-bit, 64-bit Kernel) (March 5, \
2018)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 169, 3, 54, "Subsection",ExpressionUUID->"05bd2374-6763-4b48-9d00-b6ad3c43375c"],
Cell[752, 27, 468, 9, 81, "Text",ExpressionUUID->"192f22d7-ba0c-4e63-bfa3-ef6cd1a12ba5"],
Cell[1223, 38, 256, 6, 35, "Text",ExpressionUUID->"1a47bcf0-e8d0-4bb4-8c90-f4a021590068"],
Cell[1482, 46, 361, 6, 58, "Text",ExpressionUUID->"b9a993fc-6224-42a0-8e55-ef522a531633"],
Cell[CellGroupData[{
Cell[1868, 56, 165, 3, 45, "Subsubsection",ExpressionUUID->"a464d44a-24aa-423c-833f-9f57d7eb82d4"],
Cell[2036, 61, 183, 3, 35, "Text",ExpressionUUID->"bc86f5ed-4c0d-4882-8849-0b91eb832fe7"],
Cell[2222, 66, 193, 3, 30, "Input",ExpressionUUID->"a4cc503b-7891-4ccb-801e-891d461c559a"],
Cell[2418, 71, 179, 4, 30, "Input",ExpressionUUID->"a3d15deb-5c28-4de2-ab83-bec7ee79585d"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2634, 80, 190, 3, 45, "Subsubsection",ExpressionUUID->"6c12f0e2-191a-4504-8dca-6f7efbfd9598"],
Cell[2827, 85, 196, 3, 35, "Text",ExpressionUUID->"19dad156-9720-4acc-a45d-c9865a6d9cb6"],
Cell[3026, 90, 273, 5, 30, "Input",ExpressionUUID->"cb3cbfb0-0543-4be2-b390-7cf4630360bf"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3336, 100, 238, 4, 45, "Subsubsection",ExpressionUUID->"a91bf0c9-5347-4b79-be9f-5ff6692df981"],
Cell[3577, 106, 236, 6, 35, "Text",ExpressionUUID->"40a33470-9c71-4f69-a36c-6a0e3f72f1f6"],
Cell[3816, 114, 225, 4, 35, "Text",ExpressionUUID->"c0316d1d-8fa7-4d28-87ad-e4e0fd5f4a2c"],
Cell[4044, 120, 278, 5, 30, "Input",ExpressionUUID->"8a7187e9-489d-4078-be64-2afcd65ddcac"],
Cell[4325, 127, 529, 10, 115, "Input",ExpressionUUID->"578f4083-1411-4fe1-a1da-99b07a788987"],
Cell[4857, 139, 172, 3, 35, "Text",ExpressionUUID->"c88a996c-233e-42f0-a94a-a020ed01e86f"],
Cell[5032, 144, 288, 5, 30, "Input",ExpressionUUID->"21a39a78-4e26-42c0-9089-cb0b90bbf1a8"],
Cell[5323, 151, 512, 8, 115, "Input",ExpressionUUID->"1b150e7a-0ba6-4c7a-a883-68783b4e8a0a"],
Cell[5838, 161, 303, 7, 35, "Text",ExpressionUUID->"6a6b50f6-a6f8-45c0-90f3-a0e8d14c3bec"],
Cell[6144, 170, 283, 5, 30, "Input",ExpressionUUID->"ad7ac871-e9e3-4f50-b3b2-23eaf98aa7fd"],
Cell[6430, 177, 461, 9, 115, "Input",ExpressionUUID->"ffbd2ea2-0610-48e0-8172-e2fd557ff384"],
Cell[CellGroupData[{
Cell[6916, 190, 1694, 34, 220, "Input",ExpressionUUID->"91689c66-9ef7-4ea8-97c6-9b84e33dca42"],
Cell[8613, 226, 350, 8, 34, "Output",ExpressionUUID->"9a655037-0f27-4edf-bf42-e3d652765798"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[9012, 240, 283, 4, 45, "Subsubsection",ExpressionUUID->"d89ef6f8-4031-499e-b568-a95cb12623af"],
Cell[9298, 246, 226, 5, 35, "Text",ExpressionUUID->"d5e27bc1-a8a4-455c-b4a4-b1700457fff4"],
Cell[9527, 253, 337, 5, 30, "Input",ExpressionUUID->"1821f8a2-4f5b-4155-b199-53689c39f1b6"],
Cell[9867, 260, 581, 10, 115, "Input",ExpressionUUID->"2f873a83-f256-4bab-bebd-3df07b7aa396"],
Cell[10451, 272, 259, 6, 35, "Text",ExpressionUUID->"76218c4d-5da7-4a04-b794-8fb128bcc82f"],
Cell[10713, 280, 382, 6, 30, "Input",ExpressionUUID->"8158fecd-a9aa-484d-b089-447cc4afe92d"],
Cell[11098, 288, 537, 10, 115, "Input",ExpressionUUID->"4bd724b1-c8cf-41e2-beb8-bb615cd68fc0"],
Cell[CellGroupData[{
Cell[11660, 302, 830, 15, 115, "Input",ExpressionUUID->"3a97fbc1-4dc3-4b40-9613-9b1cdfa394b1"],
Cell[12493, 319, 350, 8, 34, "Output",ExpressionUUID->"a4e922f1-7f3d-4c84-9454-83d7bdd67c2e"]
}, Open  ]],
Cell[12858, 330, 236, 6, 35, "Text",ExpressionUUID->"fecce6e4-823b-48d7-9428-0497e5ac8fc8"],
Cell[13097, 338, 380, 6, 30, "Input",ExpressionUUID->"5268662d-8858-4bc6-8467-b4dc9be6ca65"],
Cell[13480, 346, 568, 9, 115, "Input",ExpressionUUID->"e870360e-81b5-49c3-8b07-436d251bedff"],
Cell[CellGroupData[{
Cell[14073, 359, 1174, 19, 115, "Input",ExpressionUUID->"759efa02-f7c2-4322-aea2-bd07f998a5dd"],
Cell[15250, 380, 350, 8, 34, "Output",ExpressionUUID->"8e470fb1-a804-43ae-9dce-488ed4f94511"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[15649, 394, 322, 4, 45, "Subsubsection",ExpressionUUID->"44645ae2-0e1c-4d6f-a9db-e29254098753"],
Cell[15974, 400, 312, 7, 58, "Text",ExpressionUUID->"a0a9f5aa-22a2-4571-a815-2b48f2ccd57c"],
Cell[16289, 409, 271, 5, 30, "Input",ExpressionUUID->"2314ce70-b0ba-4709-8b64-0b430ae89b55"],
Cell[16563, 416, 1671, 36, 241, "Input",ExpressionUUID->"7189c0cc-96ed-4d21-9e97-0f20ee9e67e9"],
Cell[18237, 454, 1140, 18, 283, "Input",ExpressionUUID->"3586f360-9377-461a-93cb-fe122b6e587b"],
Cell[19380, 474, 270, 6, 58, "Text",ExpressionUUID->"e6ab7ad2-88a1-41d2-9b43-2116a35ff6ec"],
Cell[19653, 482, 4294, 103, 430, "Input",ExpressionUUID->"d42d5514-6bb8-413d-bce3-821c17c53df8"],
Cell[CellGroupData[{
Cell[23972, 589, 566, 11, 136, "Input",ExpressionUUID->"1795441e-0a5e-4ebb-b293-d280e6608e93"],
Cell[24541, 602, 350, 8, 34, "Output",ExpressionUUID->"865b7a72-6714-4f38-a664-41869ecafaa0"]
}, Open  ]],
Cell[24906, 613, 397, 8, 58, "Text",ExpressionUUID->"d40b654b-556d-4730-9bd6-70ec9633814d"],
Cell[25306, 623, 855, 19, 136, "Input",ExpressionUUID->"087efacc-0c08-4257-9465-c09c69067193"],
Cell[26164, 644, 208, 3, 35, "Text",ExpressionUUID->"8a59b64f-b723-4f0d-9d98-2d9e776a35fc"],
Cell[CellGroupData[{
Cell[26397, 651, 1507, 32, 367, "Input",ExpressionUUID->"748f1be7-c446-415b-ae2e-38b8d1bf928b"],
Cell[27907, 685, 350, 8, 34, "Output",ExpressionUUID->"85c598bb-7a17-44e6-984e-5882a70dd989"]
}, Open  ]],
Cell[28272, 696, 158, 3, 35, "Text",ExpressionUUID->"70d583f9-8b4c-48b9-acbc-270b7097efdc"],
Cell[CellGroupData[{
Cell[28455, 703, 1267, 25, 304, "Input",ExpressionUUID->"0b287b16-a044-4f6f-a928-0548eca90fd0"],
Cell[29725, 730, 350, 8, 34, "Output",ExpressionUUID->"0fd7289e-7226-4277-b7b2-7cf64496f7fa"]
}, Open  ]],
Cell[30090, 741, 197, 3, 35, "Text",ExpressionUUID->"627e99c2-9279-458a-8a71-5891c3d27a59"],
Cell[CellGroupData[{
Cell[30312, 748, 1888, 42, 514, "Input",ExpressionUUID->"3a00c166-ce62-4de5-960e-5564148a5ce0"],
Cell[32203, 792, 350, 8, 34, "Output",ExpressionUUID->"22163997-25ae-4176-93e3-a71a4942e305"]
}, Open  ]],
Cell[32568, 803, 184, 3, 35, "Text",ExpressionUUID->"01400c2e-f767-4346-8091-bd34903458b5"],
Cell[CellGroupData[{
Cell[32777, 810, 1962, 45, 262, "Input",ExpressionUUID->"064e84a6-eed8-4afc-bee5-fbb565d0e73e"],
Cell[34742, 857, 350, 8, 34, "Output",ExpressionUUID->"1297d8a7-5a66-493d-a65f-58527bef8164"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[35141, 871, 367, 6, 45, "Subsubsection",ExpressionUUID->"abfdc63a-5c9d-4867-890a-52727b198d9d"],
Cell[35511, 879, 236, 6, 35, "Text",ExpressionUUID->"9262cbfa-ad94-4207-9e54-35d9f388c547"],
Cell[35750, 887, 254, 4, 30, "Input",ExpressionUUID->"7c774c02-460e-4f41-bc33-536f7f0e0eae"],
Cell[36007, 893, 537, 10, 115, "Input",ExpressionUUID->"f41aec5d-deb4-4a0c-8a3e-505ac330a0f4"],
Cell[36547, 905, 1178, 28, 157, "Input",ExpressionUUID->"77c66bb9-374c-4400-81ab-7ebfdf62c6b3"],
Cell[CellGroupData[{
Cell[37750, 937, 1120, 24, 325, "Input",ExpressionUUID->"b3a4f768-9b2c-43b1-950c-ba5e2f0fbaab"],
Cell[38873, 963, 350, 8, 34, "Output",ExpressionUUID->"60b8cf9d-a83d-49e5-a715-d5c11e7b292e"]
}, Open  ]],
Cell[CellGroupData[{
Cell[39260, 976, 820, 19, 241, "Input",ExpressionUUID->"664cd531-b2ba-4f20-8cd9-e5197a36186e"],
Cell[40083, 997, 350, 8, 34, "Output",ExpressionUUID->"ee090ac4-f081-4be8-8cd5-599dfe5ab8a6"]
}, Open  ]],
Cell[CellGroupData[{
Cell[40470, 1010, 1088, 22, 283, "Input",ExpressionUUID->"43788f0b-2014-4bcf-a97b-6b16da462bcf"],
Cell[41561, 1034, 350, 8, 34, "Output",ExpressionUUID->"b4eaceea-d719-4fd9-a7ce-a65e5f4e7adc"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[41960, 1048, 318, 5, 45, "Subsubsection",ExpressionUUID->"4fd4a635-5a77-4890-bf79-fff326cb0abd"],
Cell[42281, 1055, 342, 8, 58, "Text",ExpressionUUID->"b5f6aef0-3cf9-4120-a3f1-11d0439c3752"],
Cell[42626, 1065, 226, 3, 58, "Text",ExpressionUUID->"4a47f72c-4d7b-4d85-916a-302572df4ada"],
Cell[42855, 1070, 397, 9, 73, "Input",ExpressionUUID->"f732a42e-2ec3-485d-99ab-e30614d77f30"],
Cell[43255, 1081, 277, 6, 35, "Text",ExpressionUUID->"c2d41c7e-da50-45be-9600-759439d01389"],
Cell[43535, 1089, 488, 10, 73, "Input",ExpressionUUID->"f42db428-cb36-49e9-ba3f-243128822043"],
Cell[44026, 1101, 280, 6, 35, "Text",ExpressionUUID->"1a323323-b933-4540-ba13-282deb436069"],
Cell[44309, 1109, 621, 14, 115, "Input",ExpressionUUID->"34482232-1160-449d-a9da-424531f2461e"],
Cell[44933, 1125, 628, 11, 81, "Text",ExpressionUUID->"a732cfb9-3a0e-4be1-9095-d45e008f0f80"],
Cell[45564, 1138, 1249, 30, 178, "Input",ExpressionUUID->"f8e95ed6-5097-42db-8037-47cf1256ae9f"],
Cell[46816, 1170, 170, 3, 35, "Text",ExpressionUUID->"839b4617-f6ac-42a7-9c51-d859a1c5251b"],
Cell[46989, 1175, 418, 9, 52, "Input",ExpressionUUID->"e57bd22c-5844-495e-98ec-6d6f235ef0ac"],
Cell[47410, 1186, 193, 3, 35, "Text",ExpressionUUID->"827b63ca-b9e4-4b86-ad55-54b70303d2c2"],
Cell[47606, 1191, 538, 11, 52, "Input",ExpressionUUID->"654490fd-e8df-4a61-b5be-88dbac044493"],
Cell[48147, 1204, 245, 4, 35, "Text",ExpressionUUID->"e9b50439-fda8-4d5e-aed3-465635432ecf"],
Cell[48395, 1210, 563, 11, 52, "Input",ExpressionUUID->"0ee98fa0-326f-44d7-aaf5-9a2e853a8383"]
}, Open  ]]
}, Open  ]]
}
]
*)

