(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.3' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[     71692,       1771]
NotebookOptionsPosition[     65453,       1676]
NotebookOutlinePosition[     65808,       1692]
CellTagsIndexPosition[     65765,       1689]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Example: extending the language with preprocessor", "Subsection",
 CellChangeTimes->{{3.7191203670416203`*^9, 3.71912037087922*^9}, {
  3.729317197243264*^9, 3.7293172107850647`*^9}, {3.7294856657056656`*^9, 
  3.7294856682172976`*^9}, {3.7294947148287277`*^9, 
  3.7294947204924192`*^9}},ExpressionUUID->"16c46bba-dc55-4b32-bb63-\
389c38155a22"],

Cell["\<\
This notebook presents the use of grammar preprocessor to extend the core FMT \
language.\
\>", "Text",
 CellChangeTimes->{{3.7293172201896644`*^9, 3.7293173407618647`*^9}, {
  3.7293175789972644`*^9, 3.7293175806664643`*^9}, {3.7294856740361724`*^9, 
  3.729485779714141*^9}, {3.7294871739117503`*^9, 3.7294871771253505`*^9}, {
  3.7294886801272507`*^9, 3.729488683655887*^9}, {3.7294947318812003`*^9, 
  3.729494761814515*^9}},ExpressionUUID->"61f8085e-1a1e-44ce-8efd-\
0d2afe16fcb8"],

Cell[BoxData[
 RowBox[{"<<", "FMT`"}]], "Input",
 CellChangeTimes->{{3.7191203731100197`*^9, 3.7191203746388197`*^9}},
 CellLabel->"In[1]:=",ExpressionUUID->"1c63c929-62a5-414c-b53d-e9020ac47055"],

Cell["\<\
A simple motivating example is the Boolean Xor operation, which is not \
present in the core language, but can be easily implemented in terms of \
existing And and Or operations:\
\>", "Text",
 CellChangeTimes->{{3.729502778830348*^9, 
  3.729502878500948*^9}},ExpressionUUID->"49e5c6d6-0e23-400d-bf01-\
ab14ea17c3f7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BooleanConvert", "[", 
  RowBox[{"Xor", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.729502927256148*^9, 3.729502931873785*^9}},
 CellLabel->"In[2]:=",ExpressionUUID->"75744d54-9543-4c08-ac73-9eb97c0a2ad7"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{"a", "&&", 
    RowBox[{"!", "b"}]}], ")"}], "||", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"!", "a"}], "&&", "b"}], ")"}]}]], "Output",
 CellLabel->"Out[2]=",ExpressionUUID->"da7ba3af-e5b7-48a7-9ceb-6bbc0f7ecadc"]
}, Open  ]],

Cell["\<\
The following function is sufficient to replace two-argument Xors in any \
expression that is given by its argument:\
\>", "Text",
 CellChangeTimes->{{3.7295029501008196`*^9, 3.7295030496112976`*^9}, {
  3.7295036977596064`*^9, 
  3.7295037407722063`*^9}},ExpressionUUID->"f91e836d-0aa3-4344-8a51-\
8a08d45cb2be"],

Cell[BoxData[
 RowBox[{
  RowBox[{"replaceXors", "[", "expression_", "]"}], ":=", 
  RowBox[{"expression", "/.", 
   RowBox[{
    RowBox[{"Xor", "[", 
     RowBox[{"a_", ",", "b_"}], "]"}], "\[Rule]", 
    RowBox[{"BooleanConvert", "[", 
     RowBox[{"Xor", "[", 
      RowBox[{"a", ",", "b"}], "]"}], "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.7295030514053206`*^9, 3.7295030798524857`*^9}, {
  3.7295037453314066`*^9, 3.7295037628210063`*^9}, {3.7295053728354278`*^9, 
  3.72950537456705*^9}},
 CellLabel->"In[3]:=",ExpressionUUID->"9ad736ca-0960-4a0d-ac75-16e9fd435dd8"],

Cell["\<\
In practice, such preprocessing functions will be much more complex, but in \
this example a single line of code is exactly what is needed.\
\>", "Text",
 CellChangeTimes->{{3.7295037720904064`*^9, 
  3.7295038449742064`*^9}},ExpressionUUID->"03feeaff-15c7-438e-ad69-\
c89ae6b3c51d"],

Cell["\<\
It is possible to instruct FMT to use this (or any other) preprocessing \
function in one of the first steps of grammar analysis. A dedicated option \
can be set to user-provided processor any time, before the operation that \
uses the new construct is defined.\
\>", "Text",
 CellChangeTimes->{{3.729503090134018*^9, 3.7295031566012726`*^9}, {
  3.7295035629342065`*^9, 3.729503594081806*^9}, {3.7295038499598064`*^9, 
  3.7295039345272064`*^9}},ExpressionUUID->"ce5422d1-6b08-4dfd-9bb0-\
cff38edc3377"],

Cell[BoxData[
 RowBox[{
  RowBox[{"sys", "=", 
   RowBox[{"newSystem", "[", "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}},
 CellLabel->"In[4]:=",ExpressionUUID->"2c78ecb3-6467-456f-8333-dfb50ec25ff7"],

Cell["The new option value can be set with the following:", "Text",
 CellChangeTimes->{{3.7295035993702064`*^9, 
  3.7295036327822065`*^9}},ExpressionUUID->"d9a95ebc-4068-40f8-9a63-\
e999232d46b8"],

Cell[BoxData[
 RowBox[{"setOption", "[", 
  RowBox[{"sys", ",", "FMTOptions`grammarPreprocessor", ",", "replaceXors"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.7295036355922065`*^9, 3.729503663664606*^9}, {
  3.7295039409096065`*^9, 3.729503944966606*^9}},
 CellLabel->"In[5]:=",ExpressionUUID->"2011dc5e-d8d8-4ac8-ab36-86a34f7a11e5"],

Cell["\<\
The package and some example operation can be defined in the usual way:\
\>", "Text",
 CellChangeTimes->{{3.7295040248748064`*^9, 
  3.729504037528406*^9}},ExpressionUUID->"8f6bdbfa-2c78-4e69-93d9-\
69105985558e"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<Tools`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.7293174539452643`*^9}},
 CellLabel->"In[6]:=",ExpressionUUID->"e4067743-3291-4a93-8a1c-cbd6bfaff6f0"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{"sys", ",", "\"\<Tools`\>\"", ",", 
   RowBox[{"{", 
    RowBox[{"someOperation", "[", "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.729317486602064*^9}, {3.7293175469508643`*^9, 3.729317595083864*^9}, {
   3.7293176275026646`*^9, 3.729317645630864*^9}, {3.7294864992459507`*^9, 
   3.7294865426279507`*^9}, {3.7294870058049507`*^9, 
   3.7294870273857503`*^9}, {3.7294872700283504`*^9, 
   3.7294872887311506`*^9}, {3.7295035285756063`*^9, 3.7295035496446066`*^9}},
 
 CellLabel->"In[7]:=",ExpressionUUID->"ca4a08a3-e3bc-49d4-b163-70b06f8c010d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Tools`\>\"", ",", "someOperation", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"x", "\[Rule]", "boolean"}], ",", "\[IndentingNewLine]", 
     RowBox[{"y", "\[Rule]", "boolean"}], ",", "\[IndentingNewLine]", 
     RowBox[{"z", "\[Rule]", "boolean"}]}], "\[IndentingNewLine]", "}"}], ",",
    "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{"z", "=", 
    RowBox[{"Xor", "[", 
     RowBox[{"x", ",", "y"}], "]"}]}]}], "\[IndentingNewLine]", 
  "]"}]], "Input",
 CellChangeTimes->{{3.719120761628022*^9, 3.7191207935300226`*^9}, {
   3.7191234295074396`*^9, 3.71912345393704*^9}, {3.7191236921238413`*^9, 
   3.7191236925762415`*^9}, {3.7191254587414527`*^9, 
   3.7191254691778526`*^9}, {3.719125627096654*^9, 3.7191256328998537`*^9}, {
   3.7191266533046603`*^9, 3.71912665427186*^9}, {3.7191271307426634`*^9, 
   3.7191271335194635`*^9}, 3.720351102816934*^9, {3.7293179166561794`*^9, 
   3.729317928420108*^9}, {3.729317962886054*^9, 3.7293179782869453`*^9}, {
   3.729318014110153*^9, 3.7293180196035995`*^9}, {3.7294866616083508`*^9, 
   3.7294866861087503`*^9}, 3.72948693442455*^9, 3.7294870313793507`*^9, {
   3.72948708562195*^9, 3.72948708829255*^9}, {3.7295040469362063`*^9, 
   3.7295040797190065`*^9}},
 CellLabel->"In[8]:=",ExpressionUUID->"7244aafc-445e-4833-9236-4a17824f1daa"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[8]=",ExpressionUUID->"d6d687a5-afb7-48fa-95ed-184b601b07f4"]
}, Open  ]],

Cell["\<\
It is possible to see that Xor was indeed expanded with appropriate \
expression by showing the internal grammar representation of the whole \
operation:\
\>", "Text",
 CellChangeTimes->{{3.7295041301902065`*^9, 3.7295041842436066`*^9}, {
  3.7295044820674067`*^9, 
  3.7295045352358065`*^9}},ExpressionUUID->"f012fa81-7314-4574-a8b3-\
0899277d16b0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"showGrammar", "[", 
  RowBox[{"sys", ",", "\"\<Tools`\>\"", ",", "someOperation"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.7295040985728064`*^9, 3.7295041197216063`*^9}, {
  3.7295044671684065`*^9, 3.7295044742508063`*^9}},
 CellLabel->"In[9]:=",ExpressionUUID->"08e2ce84-f5aa-4b86-a724-80608a2d95e1"],

Cell[BoxData[
 GraphicsBox[
  TagBox[GraphicsComplexBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQPa31z06PuZ/tu3VU+DikhRwuOJmJ9mi8tn/x5VkO
ry6fw5GXsZney37C+TsntO7lUrlhvwtMczgwQAFEPxNU/oU9jL/OfK3gqYVs
DjD1EPkfcPnnYHPfQc37Yd8FFn9kvwdqzqJ7V1I75b/C+fWXvZOOz2BwgKmH
8WHyEPf/toe5C+I/Zrg7G8DqBRxg9t+ZzXfwzCoeuHkwPsy8o2D/c8L1f/bd
L+G3nh/OF4gUSNG4IwTXz8pgseN5pCBcv8ymF/yZcSJwPgAxLozQ
    "], {
     {RGBColor[0.55, 0.45, 0.45], 
      LineBox[{{1, 2}, {1, 3}, {3, 4}, {3, 7}, {4, 5}, {4, 6}, {7, 8}, {7, 
       16}, {8, 9}, {8, 12}, {9, 10}, {9, 11}, {12, 13}, {13, 14}, {13, 15}, {
       16, 17}, {16, 21}, {17, 18}, {18, 19}, {18, 20}, {21, 22}, {21, 
       23}}]}, {
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["numberedT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 1],
        TagBox[
         RowBox[{"numberedT", "[", 
           RowBox[{"1", ",", 
             RowBox[{"assignmentT", "[", 
               RowBox[{
                 RowBox[{"objectLhsT", "[", 
                   RowBox[{"z", ",", "booleanT"}], "]"}], ",", 
                 RowBox[{"orT", "[", 
                   RowBox[{
                    RowBox[{"andT", "[", 
                    RowBox[{
                    RowBox[{"objectT", "[", 
                    RowBox[{"x", ",", "booleanT"}], "]"}], ",", 
                    RowBox[{"notT", "[", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"y", ",", "booleanT"}], "]"}], "]"}]}], "]"}], 
                    ",", 
                    RowBox[{"andT", "[", 
                    RowBox[{
                    RowBox[{"notT", "[", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"x", ",", "booleanT"}], "]"}], "]"}], ",", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"y", ",", "booleanT"}], "]"}]}], "]"}]}], "]"}]}],
                "]"}]}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`numberedT[1, 
          FMTSymbols`assignmentT[
           FMTSymbols`objectLhsT[$CellContext`z, FMTSymbols`booleanT], 
           FMTSymbols`orT[
            FMTSymbols`andT[
             FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT], 
             FMTSymbols`notT[
              FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]], 
            FMTSymbols`andT[
             FMTSymbols`notT[
              FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT]], 
             FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["1",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 2],
        TagBox["1", HoldForm]],
       Annotation[#, 
        HoldForm[1], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["assignmentT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 3],
        TagBox[
         RowBox[{"assignmentT", "[", 
           RowBox[{
             RowBox[{"objectLhsT", "[", 
               RowBox[{"z", ",", "booleanT"}], "]"}], ",", 
             RowBox[{"orT", "[", 
               RowBox[{
                 RowBox[{"andT", "[", 
                   RowBox[{
                    RowBox[{"objectT", "[", 
                    RowBox[{"x", ",", "booleanT"}], "]"}], ",", 
                    RowBox[{"notT", "[", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"y", ",", "booleanT"}], "]"}], "]"}]}], "]"}], 
                 ",", 
                 RowBox[{"andT", "[", 
                   RowBox[{
                    RowBox[{"notT", "[", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"x", ",", "booleanT"}], "]"}], "]"}], ",", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"y", ",", "booleanT"}], "]"}]}], "]"}]}], "]"}]}],
            "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`assignmentT[
          FMTSymbols`objectLhsT[$CellContext`z, FMTSymbols`booleanT], 
          FMTSymbols`orT[
           FMTSymbols`andT[
            FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT], 
            FMTSymbols`notT[
             FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]], 
           FMTSymbols`andT[
            FMTSymbols`notT[
             FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT]], 
            FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["objectLhsT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 4],
        TagBox[
         RowBox[{"objectLhsT", "[", 
           RowBox[{"z", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`objectLhsT[$CellContext`z, FMTSymbols`booleanT]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["z",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 5],
        TagBox["z", HoldForm]],
       Annotation[#, 
        HoldForm[$CellContext`z], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 6],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["orT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 7],
        TagBox[
         RowBox[{"orT", "[", 
           RowBox[{
             RowBox[{"andT", "[", 
               RowBox[{
                 RowBox[{"objectT", "[", 
                   RowBox[{"x", ",", "booleanT"}], "]"}], ",", 
                 RowBox[{"notT", "[", 
                   RowBox[{"objectT", "[", 
                    RowBox[{"y", ",", "booleanT"}], "]"}], "]"}]}], "]"}], 
             ",", 
             RowBox[{"andT", "[", 
               RowBox[{
                 RowBox[{"notT", "[", 
                   RowBox[{"objectT", "[", 
                    RowBox[{"x", ",", "booleanT"}], "]"}], "]"}], ",", 
                 RowBox[{"objectT", "[", 
                   RowBox[{"y", ",", "booleanT"}], "]"}]}], "]"}]}], "]"}], 
         HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`orT[
          FMTSymbols`andT[
           FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT], 
           FMTSymbols`notT[
            FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]], 
          FMTSymbols`andT[
           FMTSymbols`notT[
            FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT]], 
           FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["andT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 8],
        TagBox[
         RowBox[{"andT", "[", 
           RowBox[{
             RowBox[{"objectT", "[", 
               RowBox[{"x", ",", "booleanT"}], "]"}], ",", 
             RowBox[{"notT", "[", 
               RowBox[{"objectT", "[", 
                 RowBox[{"y", ",", "booleanT"}], "]"}], "]"}]}], "]"}], 
         HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`andT[
          FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT], 
          FMTSymbols`notT[
           FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["objectT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 9],
        TagBox[
         RowBox[{"objectT", "[", 
           RowBox[{"x", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["x",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 10],
        TagBox["x", HoldForm]],
       Annotation[#, 
        HoldForm[$CellContext`x], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 11],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["notT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 12],
        TagBox[
         RowBox[{"notT", "[", 
           RowBox[{"objectT", "[", 
             RowBox[{"y", ",", "booleanT"}], "]"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`notT[
          FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["objectT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 13],
        TagBox[
         RowBox[{"objectT", "[", 
           RowBox[{"y", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["y",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 14],
        TagBox["y", HoldForm]],
       Annotation[#, 
        HoldForm[$CellContext`y], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 15],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["andT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 16],
        TagBox[
         RowBox[{"andT", "[", 
           RowBox[{
             RowBox[{"notT", "[", 
               RowBox[{"objectT", "[", 
                 RowBox[{"x", ",", "booleanT"}], "]"}], "]"}], ",", 
             RowBox[{"objectT", "[", 
               RowBox[{"y", ",", "booleanT"}], "]"}]}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`andT[
          FMTSymbols`notT[
           FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT]], 
          FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["notT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 17],
        TagBox[
         RowBox[{"notT", "[", 
           RowBox[{"objectT", "[", 
             RowBox[{"x", ",", "booleanT"}], "]"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`notT[
          FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["objectT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 18],
        TagBox[
         RowBox[{"objectT", "[", 
           RowBox[{"x", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`objectT[$CellContext`x, FMTSymbols`booleanT]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["x",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 19],
        TagBox["x", HoldForm]],
       Annotation[#, 
        HoldForm[$CellContext`x], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 20],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["objectT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 21],
        TagBox[
         RowBox[{"objectT", "[", 
           RowBox[{"y", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`objectT[$CellContext`y, FMTSymbols`booleanT]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["y",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 22],
        TagBox["y", HoldForm]],
       Annotation[#, 
        HoldForm[$CellContext`y], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.025],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 23],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ]}}],
   Annotation[#, VertexCoordinateRules -> CompressedData["
1:eJxTTMoPSmViYGAQB2IQPa31z06PuZ/tu3VU+DikhRwuOJmJ9mi8tn/x5VkO
ry6fw5GXsZney37C+TsntO7lUrlhvwtMczgwQAFEPxNU/oU9jL/OfK3gqYVs
DjD1EPkfcPnnYHPfQc37Yd8FFn9kvwdqzqJ7V1I75b/C+fWXvZOOz2BwgKmH
8WHyEPf/toe5C+I/Zrg7G8DqBRxg9t+ZzXfwzCoeuHkwPsy8o2D/c8L1f/bd
L+G3nh/OF4gUSNG4IwTXz8pgseN5pCBcv8ymF/yZcSJwPgAxLozQ
     "]]& ],
  AspectRatio->NCache[Rational[1, 4] 7^Rational[1, 2], 0.6614378277661477],
  FormatType:>StandardForm,
  FrameTicks->Automatic,
  PlotRange->All,
  PlotRangePadding->Scaled[0.1]]], "Output",
 CellLabel->
  "Out[9]//TreeForm=",ExpressionUUID->"69c39ad0-7278-4076-9d10-407cbff5b9fa"]
}, Open  ]],

Cell["And ultimately, the generated source code, here in Ada:", "Text",
 CellChangeTimes->{{3.72948686770195*^9, 3.7294868857249503`*^9}, {
  3.7295045614030066`*^9, 
  3.7295045753202066`*^9}},ExpressionUUID->"f4596866-d996-4e91-8e1a-\
7c62c67c37cc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"generateCode", "[", 
   RowBox[{"sys", ",", "\"\<Ada\>\""}], "]"}], "[", "\"\<tools.adb\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.72948689004615*^9, 3.7294868957421503`*^9}, {
  3.7294869486265507`*^9, 3.72948695201175*^9}, {3.7295045805774064`*^9, 
  3.729504584930806*^9}},
 CellLabel->"In[10]:=",ExpressionUUID->"eb4c6ac3-0c47-4003-831f-caac2c9ddc70"],

Cell[BoxData["\<\"package body Tools is\\n   \\n   -- Operation \
definitions.\\n   \\n   procedure Some_Operation is\\n      X : Boolean;\\n   \
   Y : Boolean;\\n      Z : Boolean;\\n   begin\\n      Z := (X and not Y) or \
(not X and Y);\\n   end Some_Operation;\\n   \\nend Tools;\\n\"\>"], "Output",\

 CellLabel->"Out[10]=",ExpressionUUID->"dab9b77c-b830-4e06-91c7-a87f1beb779a"]
}, Open  ]],

Cell["\<\
Another simple example is the implementation of infinite loop. The core \
language does not offer any special construct for this purpose, and \
traditionally C and C++ programmers use some variant of while or for control \
structure to implement infinite loops (Ada programmers are in a better \
position in this regard).\
\>", "Text",ExpressionUUID->"bd5c8098-42d9-486c-b633-09bc2b414254"],

Cell["\<\
FMT allows to extend the core language with appropriate replacement, for \
example:\
\>", "Text",
 CellChangeTimes->{{3.7295046202490063`*^9, 
  3.7295048234900064`*^9}},ExpressionUUID->"376636f8-8ea6-4160-8157-\
c8a1d3e36084"],

Cell[BoxData[
 RowBox[{
  RowBox[{"replaceInfLoops", "[", "expression_", "]"}], ":=", 
  RowBox[{"expression", "/.", 
   RowBox[{
    RowBox[{"InfiniteLoop", "[", "body_", "]"}], "\[Rule]", 
    RowBox[{"FMTSymbols`whileT", "[", 
     RowBox[{"True", ",", "body"}], "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.7295048452414064`*^9, 3.7295048865728064`*^9}, 
   3.729504947084606*^9, {3.7295053454032755`*^9, 3.7295053685921736`*^9}, {
   3.729505443813739*^9, 3.7295054454995604`*^9}},
 CellLabel->"In[11]:=",ExpressionUUID->"193d7606-5e5a-47cb-802f-470a64c27b15"],

Cell["\<\
Above, FMTSymbols`whileT was used instead of standard While for a reason - \
the standard construct has a predefined meaning and its use could trigger \
unintended evaluation at this point. The internal symbol FMTSymbols`whileT, \
which internally represents the While statement anyway, has no special \
meaning from the point of view of the Wolfram language and can be safely used \
in this way.\
\>", "Text",
 CellChangeTimes->{{3.7295049530750065`*^9, 
  3.7295051179160366`*^9}},ExpressionUUID->"fa0ab6e3-c0f8-4165-a6ce-\
62048aa10028"],

Cell["\<\
The remaining steps are as above - install the preprocessor as an option and \
define the operation with the use of newly recognized language construct:\
\>", "Text",
 CellChangeTimes->{{3.7295051208177853`*^9, 
  3.7295051576748757`*^9}},ExpressionUUID->"c8b69731-b240-4b35-b78c-\
4c5772ec647d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"sys", "=", 
   RowBox[{"newSystem", "[", "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}},
 CellLabel->"In[12]:=",ExpressionUUID->"741ff627-8a85-48e6-8221-4b3c53c3d52b"],

Cell["The new option value can be set with the following:", "Text",
 CellChangeTimes->{{3.7295035993702064`*^9, 
  3.7295036327822065`*^9}},ExpressionUUID->"92707186-880c-4e89-b58d-\
fc39d0b09865"],

Cell[BoxData[
 RowBox[{"setOption", "[", 
  RowBox[{
  "sys", ",", "FMTOptions`grammarPreprocessor", ",", "replaceInfLoops"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.7295036355922065`*^9, 3.729503663664606*^9}, {
   3.7295039409096065`*^9, 3.729503944966606*^9}, 3.7295051760586185`*^9},
 CellLabel->"In[13]:=",ExpressionUUID->"70812cec-44e2-4306-8c32-cb03f4ffc032"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<Blink`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.7293174539452643`*^9}, {3.729505232964138*^9, 3.7295052349454393`*^9}},
 CellLabel->"In[14]:=",ExpressionUUID->"36f5f3e3-f615-4a0f-8216-c0ac58ad92d5"],

Cell[BoxData[
 RowBox[{"defineDataObjects", "[", 
  RowBox[{"sys", ",", "\"\<Blink`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"light", "\[Rule]", "boolean"}], "\[IndentingNewLine]", "}"}]}], 
  "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.729505212870907*^9, 3.7295052617078123`*^9}},
 CellLabel->"In[15]:=",ExpressionUUID->"0566b070-a97a-4199-a8dc-2aa87b5caf4e"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{"sys", ",", "\"\<Blink`\>\"", ",", 
   RowBox[{"{", 
    RowBox[{"toggle", "[", "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.729317486602064*^9}, {3.7293175469508643`*^9, 3.729317595083864*^9}, {
   3.7293176275026646`*^9, 3.729317645630864*^9}, {3.7294864992459507`*^9, 
   3.7294865426279507`*^9}, {3.7294870058049507`*^9, 
   3.7294870273857503`*^9}, {3.7294872700283504`*^9, 
   3.7294872887311506`*^9}, {3.7295035285756063`*^9, 
   3.7295035496446066`*^9}, {3.72950518894688*^9, 3.7295052018401413`*^9}, {
   3.7295052408591433`*^9, 3.7295052722237515`*^9}},
 CellLabel->"In[16]:=",ExpressionUUID->"652011de-6bc4-45b8-bc3e-753464b3e1c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Blink`\>\"", ",", "toggle", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{"InfiniteLoop", "[", "\[IndentingNewLine]", 
    RowBox[{"light", "=", 
     RowBox[{"\[Not]", "light"}]}], "\[IndentingNewLine]", "]"}]}], 
  "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.719120761628022*^9, 3.7191207935300226`*^9}, {
   3.7191234295074396`*^9, 3.71912345393704*^9}, {3.7191236921238413`*^9, 
   3.7191236925762415`*^9}, {3.7191254587414527`*^9, 
   3.7191254691778526`*^9}, {3.719125627096654*^9, 3.7191256328998537`*^9}, {
   3.7191266533046603`*^9, 3.71912665427186*^9}, {3.7191271307426634`*^9, 
   3.7191271335194635`*^9}, 3.720351102816934*^9, {3.7293179166561794`*^9, 
   3.729317928420108*^9}, {3.729317962886054*^9, 3.7293179782869453`*^9}, {
   3.729318014110153*^9, 3.7293180196035995`*^9}, {3.7294866616083508`*^9, 
   3.7294866861087503`*^9}, 3.72948693442455*^9, 3.7294870313793507`*^9, {
   3.72948708562195*^9, 3.72948708829255*^9}, {3.7295040469362063`*^9, 
   3.7295040797190065`*^9}, {3.729505291638584*^9, 3.7295053252016153`*^9}, {
   3.729505889906249*^9, 3.7295058918874493`*^9}},
 CellLabel->"In[17]:=",ExpressionUUID->"04de1c2e-fa4b-47ca-a405-13bfbd3877b5"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[17]=",ExpressionUUID->"a4003d70-5b2f-45b8-8d96-81f9d320d144"]
}, Open  ]],

Cell["\<\
Both the expanded grammar and the resulting source code (here in C++) confirm \
that the language extension is properly implemented:\
\>", "Text",
 CellChangeTimes->{{3.7295054687820597`*^9, 3.7295055268308053`*^9}, {
  3.729505653362849*^9, 3.7295056578244495`*^9}, {3.7295060714592495`*^9, 
  3.729506071584049*^9}},ExpressionUUID->"a09bed30-6669-4d7e-86f9-\
70b32dcf07e4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"showGrammar", "[", 
  RowBox[{"sys", ",", "\"\<Blink`\>\"", ",", "toggle"}], "]"}]], "Input",
 CellChangeTimes->{{3.729505528047621*^9, 3.7295055354275155`*^9}},
 CellLabel->"In[18]:=",ExpressionUUID->"abfaa02e-2ef3-4583-bf57-37cbe6ed9df9"],

Cell[BoxData[
 GraphicsBox[
  TagBox[GraphicsComplexBox[{{0.8528028654224417, 5.116817192534651}, {
    0.42640143271122083`, 4.264014327112209}, {1.2792042981336624`, 
    4.264014327112209}, {0.42640143271122083`, 3.411211461689767}, {0., 
    2.5584085962673253`}, {0.8528028654224417, 2.5584085962673253`}, {
    2.132007163556104, 3.411211461689767}, {1.7056057308448835`, 
    2.5584085962673253`}, {2.5584085962673253`, 2.5584085962673253`}, {
    1.9188064472004938`, 1.705605730844884}, {1.492405014489273, 
    0.8528028654224418}, {2.3452078799117144`, 0.8528028654224418}, {
    3.1980107453341566`, 1.705605730844884}, {3.1980107453341566`, 
    0.8528028654224418}, {2.7716093126229358`, 0.}, {3.624412178045377, 0.}}, {
     {RGBColor[0.55, 0.45, 0.45], 
      LineBox[{{1, 2}, {1, 3}, {3, 4}, {3, 7}, {4, 5}, {4, 6}, {7, 8}, {7, 
       9}, {9, 10}, {9, 13}, {10, 11}, {10, 12}, {13, 14}, {14, 15}, {14, 
       16}}]}, {
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["numberedT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 1],
        TagBox[
         RowBox[{"numberedT", "[", 
           RowBox[{"1", ",", 
             RowBox[{"whileT", "[", 
               RowBox[{
                 RowBox[{"literalT", "[", 
                   RowBox[{"True", ",", "booleanT"}], "]"}], ",", 
                 RowBox[{"numberedT", "[", 
                   RowBox[{"2", ",", 
                    RowBox[{"assignmentT", "[", 
                    RowBox[{
                    RowBox[{"objectLhsT", "[", 
                    RowBox[{"Blink`light", ",", "booleanT"}], "]"}], ",", 
                    RowBox[{"notT", "[", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"Blink`light", ",", "booleanT"}], "]"}], "]"}]}], 
                    "]"}]}], "]"}]}], "]"}]}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`numberedT[1, 
          FMTSymbols`whileT[
           FMTSymbols`literalT[True, FMTSymbols`booleanT], 
           FMTSymbols`numberedT[2, 
            FMTSymbols`assignmentT[
             FMTSymbols`objectLhsT[Blink`light, FMTSymbols`booleanT], 
             FMTSymbols`notT[
              FMTSymbols`objectT[Blink`light, FMTSymbols`booleanT]]]]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["1",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 2],
        TagBox["1", HoldForm]],
       Annotation[#, 
        HoldForm[1], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["whileT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 3],
        TagBox[
         RowBox[{"whileT", "[", 
           RowBox[{
             RowBox[{"literalT", "[", 
               RowBox[{"True", ",", "booleanT"}], "]"}], ",", 
             RowBox[{"numberedT", "[", 
               RowBox[{"2", ",", 
                 RowBox[{"assignmentT", "[", 
                   RowBox[{
                    RowBox[{"objectLhsT", "[", 
                    RowBox[{"Blink`light", ",", "booleanT"}], "]"}], ",", 
                    RowBox[{"notT", "[", 
                    RowBox[{"objectT", "[", 
                    RowBox[{"Blink`light", ",", "booleanT"}], "]"}], "]"}]}], 
                   "]"}]}], "]"}]}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`whileT[
          FMTSymbols`literalT[True, FMTSymbols`booleanT], 
          FMTSymbols`numberedT[2, 
           FMTSymbols`assignmentT[
            FMTSymbols`objectLhsT[Blink`light, FMTSymbols`booleanT], 
            FMTSymbols`notT[
             FMTSymbols`objectT[Blink`light, FMTSymbols`booleanT]]]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["literalT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 4],
        TagBox[
         RowBox[{"literalT", "[", 
           RowBox[{"True", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`literalT[True, FMTSymbols`booleanT]], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["True",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 5],
        TagBox["True", HoldForm]],
       Annotation[#, 
        HoldForm[True], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 6],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["numberedT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 7],
        TagBox[
         RowBox[{"numberedT", "[", 
           RowBox[{"2", ",", 
             RowBox[{"assignmentT", "[", 
               RowBox[{
                 RowBox[{"objectLhsT", "[", 
                   RowBox[{"Blink`light", ",", "booleanT"}], "]"}], ",", 
                 RowBox[{"notT", "[", 
                   RowBox[{"objectT", "[", 
                    RowBox[{"Blink`light", ",", "booleanT"}], "]"}], "]"}]}], 
               "]"}]}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`numberedT[2, 
          FMTSymbols`assignmentT[
           FMTSymbols`objectLhsT[Blink`light, FMTSymbols`booleanT], 
           FMTSymbols`notT[
            FMTSymbols`objectT[Blink`light, FMTSymbols`booleanT]]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["2",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 8],
        TagBox["2", HoldForm]],
       Annotation[#, 
        HoldForm[2], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["assignmentT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 9],
        TagBox[
         RowBox[{"assignmentT", "[", 
           RowBox[{
             RowBox[{"objectLhsT", "[", 
               RowBox[{"Blink`light", ",", "booleanT"}], "]"}], ",", 
             RowBox[{"notT", "[", 
               RowBox[{"objectT", "[", 
                 RowBox[{"Blink`light", ",", "booleanT"}], "]"}], "]"}]}], 
           "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`assignmentT[
          FMTSymbols`objectLhsT[Blink`light, FMTSymbols`booleanT], 
          FMTSymbols`notT[
           FMTSymbols`objectT[Blink`light, FMTSymbols`booleanT]]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["objectLhsT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 10],
        TagBox[
         RowBox[{"objectLhsT", "[", 
           RowBox[{"Blink`light", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`objectLhsT[Blink`light, FMTSymbols`booleanT]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["Blink`light",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 11],
        TagBox["Blink`light", HoldForm]],
       Annotation[#, 
        HoldForm[Blink`light], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 12],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["notT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 13],
        TagBox[
         RowBox[{"notT", "[", 
           RowBox[{"objectT", "[", 
             RowBox[{"Blink`light", ",", "booleanT"}], "]"}], "]"}], 
         HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`notT[
          FMTSymbols`objectT[Blink`light, FMTSymbols`booleanT]]], 
        "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox["objectT", "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 14],
        TagBox[
         RowBox[{"objectT", "[", 
           RowBox[{"Blink`light", ",", "booleanT"}], "]"}], HoldForm]],
       Annotation[#, 
        HoldForm[
         FMTSymbols`objectT[Blink`light, FMTSymbols`booleanT]], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["Blink`light",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 15],
        TagBox["Blink`light", HoldForm]],
       Annotation[#, 
        HoldForm[Blink`light], "Tooltip"]& ], 
      TagBox[
       TooltipBox[InsetBox[
         FrameBox[
          StyleBox[
           TagBox["booleanT",
            HoldForm], "StandardForm", "Output",
           StripOnInput->False,
           LineColor->GrayLevel[0],
           FrontFaceColor->GrayLevel[0],
           BackFaceColor->GrayLevel[0],
           GraphicsColor->GrayLevel[0],
           FontSize->Scaled[0.03076923076923077],
           FontColor->GrayLevel[0]],
          Background->RGBColor[1., 1., 0.871],
          FrameStyle->GrayLevel[0.5],
          StripOnInput->False], 16],
        TagBox["booleanT", HoldForm]],
       Annotation[#, 
        HoldForm[FMTSymbols`booleanT], "Tooltip"]& ]}}],
   Annotation[#, 
    VertexCoordinateRules -> {{0.8528028654224417, 5.116817192534651}, {
     0.42640143271122083`, 4.264014327112209}, {1.2792042981336624`, 
     4.264014327112209}, {0.42640143271122083`, 3.411211461689767}, {0., 
     2.5584085962673253`}, {0.8528028654224417, 2.5584085962673253`}, {
     2.132007163556104, 3.411211461689767}, {1.7056057308448835`, 
     2.5584085962673253`}, {2.5584085962673253`, 2.5584085962673253`}, {
     1.9188064472004938`, 1.705605730844884}, {1.492405014489273, 
     0.8528028654224418}, {2.3452078799117144`, 0.8528028654224418}, {
     3.1980107453341566`, 1.705605730844884}, {3.1980107453341566`, 
     0.8528028654224418}, {2.7716093126229358`, 0.}, {3.624412178045377, 
     0.}}]& ],
  AspectRatio->NCache[Rational[7, 13]^Rational[1, 2], 0.7337993857053428],
  FormatType:>StandardForm,
  FrameTicks->Automatic,
  PlotRange->All,
  PlotRangePadding->Scaled[0.1]]], "Output",
 CellLabel->
  "Out[18]//TreeForm=",ExpressionUUID->"381fa0f7-e8c4-4b28-985a-ddaf63a0b6d1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"generateCode", "[", 
   RowBox[{"sys", ",", "\"\<C++\>\""}], "]"}], "[", "\"\<blink.cpp\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.72950554433523*^9, 3.7295055622834606`*^9}, {
  3.7295056027746496`*^9, 3.7295056195486493`*^9}, {3.729506062720249*^9, 
  3.7295060757970495`*^9}},
 CellLabel->"In[19]:=",ExpressionUUID->"67321467-88fe-4aef-b94b-379be442c5f1"],

Cell[BoxData["\<\"#include \\\"blink.h\\\"\\n\\nusing namespace \
blink;\\n\\n// Data definitions.\\n\\nbool light;\\n\\n// Operation \
definitions.\\n\\nvoid blink::toggle()\\n{\\n    while (true)\\n    {\\n      \
  light = !light;\\n    }\\n}\\n\"\>"], "Output",
 CellLabel->"Out[19]=",ExpressionUUID->"f5564976-404e-4f73-a533-1772c3b81dc4"]
}, Open  ]],

Cell["\<\
The last quick example implements the power operation (with constant \
exponent), which is not present in the core language. The reason to introduce \
dedicated support for powers with constant exponent is not just because it is \
so much more funny to write this:\
\>", "Text",ExpressionUUID->"6c98d15c-f2e5-4571-84f1-2aa28274caae"],

Cell[BoxData[
 SuperscriptBox["x", "3"]], "Input",
 CellChangeTimes->{{3.7295062775166492`*^9, 3.7295063267924495`*^9}, {
   3.7295664588174505`*^9, 3.729566534529462*^9}, 
   3.7295674715056615`*^9},ExpressionUUID->"bc714c7c-758f-4bf7-a995-\
47f8bb0d4d41"],

Cell["instead of this:", "Text",
 CellChangeTimes->{{3.7295665366866617`*^9, 
  3.7295665403994617`*^9}},ExpressionUUID->"e93e15f8-b675-4f30-bc94-\
8018c27051c5"],

Cell[BoxData[
 RowBox[{"x", "*", "x", "*", "x"}]], "Input",
 CellChangeTimes->{{3.7295665426780615`*^9, 
  3.7295665435672617`*^9}},ExpressionUUID->"cd04adc6-65aa-462f-8358-\
435f8baebcf3"],

Cell["\<\
Hand-written code can be accommodated to either style. The actual issue is \
with code that results from formula transformations, where repeated \
multiplications are always simplified to powers. An example transformation \
that produces powers is:\
\>", "Text",
 CellChangeTimes->{{3.7295062775166492`*^9, 3.7295063267924495`*^9}, {
  3.7295664588174505`*^9, 3.7295665078922615`*^9}, {3.7295665583296614`*^9, 
  3.7295666859722614`*^9}, {3.7295675095016613`*^9, 
  3.7295675095796614`*^9}},ExpressionUUID->"27267746-2a9e-4859-a4a5-\
bd6ff342adf8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ExpandAll", "[", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"a", "+", "y", "+", "z"}], ")"}], "3"], "]"}]], "Input",
 CellChangeTimes->{{3.7295666893750615`*^9, 3.729566698253462*^9}},
 CellLabel->"In[20]:=",ExpressionUUID->"8c0a32af-b274-42bd-aa33-f627c5fcb12a"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox["a", "3"], "+", 
  RowBox[{"3", " ", 
   SuperscriptBox["a", "2"], " ", "y"}], "+", 
  RowBox[{"3", " ", "a", " ", 
   SuperscriptBox["y", "2"]}], "+", 
  SuperscriptBox["y", "3"], "+", 
  RowBox[{"3", " ", 
   SuperscriptBox["a", "2"], " ", "z"}], "+", 
  RowBox[{"6", " ", "a", " ", "y", " ", "z"}], "+", 
  RowBox[{"3", " ", 
   SuperscriptBox["y", "2"], " ", "z"}], "+", 
  RowBox[{"3", " ", "a", " ", 
   SuperscriptBox["z", "2"]}], "+", 
  RowBox[{"3", " ", "y", " ", 
   SuperscriptBox["z", "2"]}], "+", 
  SuperscriptBox["z", "3"]}]], "Output",
 CellLabel->"Out[20]=",ExpressionUUID->"5f4cb438-bedc-440a-a3c8-c10001b2796c"]
}, Open  ]],

Cell["\<\
In order to use such a formula in the model the preprocessor can be used to \
emulate powers in terms of repeated multiplication:\
\>", "Text",
 CellChangeTimes->{{3.729566706414262*^9, 
  3.7295667409742613`*^9}},ExpressionUUID->"838032b4-00cc-4c1e-972c-\
be362e3563fe"],

Cell[BoxData[
 RowBox[{
  RowBox[{"replacePower", "[", "expression_", "]"}], ":=", 
  RowBox[{"With", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"{", "\[IndentingNewLine]", 
     RowBox[{"powers", "=", 
      RowBox[{"Union", "[", 
       RowBox[{"Cases", "[", 
        RowBox[{"expression", ",", 
         RowBox[{
          RowBox[{"Power", "[", 
           RowBox[{"_", ",", "n_"}], "]"}], "\[Rule]", "n"}], ",", "All"}], 
        "]"}], "]"}]}], "\[IndentingNewLine]", "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{"expression", "/.", 
     RowBox[{"Map", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"Function", "[", 
        RowBox[{"p", ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"Power", "[", 
           RowBox[{"v_", ",", "p"}], "]"}], "->", 
          RowBox[{"FMTSymbols`timesT", "@@", 
           RowBox[{"Table", "[", 
            RowBox[{"v", ",", 
             RowBox[{"{", "p", "}"}]}], "]"}]}]}]}], "\[IndentingNewLine]", 
        "]"}], ",", "\[IndentingNewLine]", "powers"}], "\[IndentingNewLine]", 
      "]"}]}]}], "\[IndentingNewLine]", "]"}]}]], "Input",
 CellLabel->"In[21]:=",ExpressionUUID->"2260a2d8-3c52-4943-8fd4-3938db503a37"],

Cell[BoxData[
 RowBox[{"setOption", "[", 
  RowBox[{"sys", ",", "FMTOptions`grammarPreprocessor", ",", "replacePower"}],
   "]"}]], "Input",
 CellChangeTimes->{{3.729506667169649*^9, 3.7295066838666496`*^9}},
 CellLabel->"In[22]:=",ExpressionUUID->"38b55f45-9510-4599-b31b-0d3ec74f2be2"],

Cell[BoxData[
 RowBox[{"declarePackage", "[", 
  RowBox[{"sys", ",", "\"\<Calc`\>\""}], "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.7293174539452643`*^9}, {3.729505232964138*^9, 3.7295052349454393`*^9}, {
   3.729506349264449*^9, 3.7295063501068497`*^9}},
 CellLabel->"In[23]:=",ExpressionUUID->"7ef3af92-96fe-4be7-b46b-dd2ecd7b0488"],

Cell[BoxData[
 RowBox[{"declareOperations", "[", 
  RowBox[{"sys", ",", "\"\<Calc`\>\"", ",", "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{"sumAndPower3", "[", 
     RowBox[{
      RowBox[{"x", "\[Rule]", 
       RowBox[{"in", "[", "integer", "]"}]}], ",", 
      RowBox[{"y", "\[Rule]", 
       RowBox[{"in", "[", "integer", "]"}]}], ",", 
      RowBox[{"z", "\[Rule]", 
       RowBox[{"in", "[", "integer", "]"}]}], ",", 
      RowBox[{"result", "\[Rule]", 
       RowBox[{"out", "[", "integer", "]"}]}]}], "]"}], "\[IndentingNewLine]",
     "}"}]}], "\[IndentingNewLine]", "]"}]], "Input",
 CellChangeTimes->{{3.71912037880402*^9, 3.7191206156588216`*^9}, {
   3.7191206498228216`*^9, 3.719120661335622*^9}, 3.7191207406928225`*^9, {
   3.7293173636888647`*^9, 3.7293173735384645`*^9}, {3.729317452369664*^9, 
   3.729317486602064*^9}, {3.7293175469508643`*^9, 3.729317595083864*^9}, {
   3.7293176275026646`*^9, 3.729317645630864*^9}, {3.7294864992459507`*^9, 
   3.7294865426279507`*^9}, {3.7294870058049507`*^9, 
   3.7294870273857503`*^9}, {3.7294872700283504`*^9, 
   3.7294872887311506`*^9}, {3.7295035285756063`*^9, 
   3.7295035496446066`*^9}, {3.72950518894688*^9, 3.7295052018401413`*^9}, {
   3.7295052408591433`*^9, 3.7295052722237515`*^9}, {3.7295064522660494`*^9, 
   3.729506539027049*^9}, 3.7295067494534492`*^9},
 CellLabel->"In[24]:=",ExpressionUUID->"10c848f6-db37-45ab-94a9-1d45ff3e7be4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"defineOperation", "[", 
  RowBox[{
  "sys", ",", "\"\<Calc`\>\"", ",", "sumAndPower3", ",", 
   "\[IndentingNewLine]", 
   RowBox[{"{", "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{"result", "=", 
    RowBox[{"Synthesize", "[", "\[IndentingNewLine]", 
     RowBox[{"ExpandAll", "[", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", "+", "y", "+", "z"}], ")"}], "^", "3"}], "]"}], 
     "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", "]"}]], "Input",\

 CellLabel->"In[25]:=",ExpressionUUID->"d8ccf03d-0c08-49e2-abc7-e7a9a7b5d449"],

Cell[BoxData[
 StyleBox["\<\"Grammar, names and flows OK.\"\>",
  StripOnInput->False,
  LineColor->RGBColor[0, 1, 0],
  FrontFaceColor->RGBColor[0, 1, 0],
  BackFaceColor->RGBColor[0, 1, 0],
  GraphicsColor->RGBColor[0, 1, 0],
  FontColor->RGBColor[0, 1, 0]]], "Output",
 CellLabel->"Out[25]=",ExpressionUUID->"938f3141-f1fc-4277-9d9b-320b99894e36"]
}, Open  ]],

Cell["\<\
The result in the final generated code (here in C) is now as expected:\
\>", "Text",
 CellChangeTimes->{{3.7295668027536616`*^9, 
  3.7295668209316616`*^9}},ExpressionUUID->"5a06aad6-ab9e-4df7-973c-\
93976d4777f6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"generateCode", "[", 
   RowBox[{"sys", ",", "\"\<C\>\""}], "]"}], "[", "\"\<calc.c\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.7295067160136495`*^9, 3.7295067196494493`*^9}},
 CellLabel->"In[26]:=",ExpressionUUID->"0069fb66-0cf7-47e9-9f2f-af5828b605d9"],

Cell[BoxData["\<\"#include \\\"calc.h\\\"\\n\\n#include <stdint.h>\\n\\n/* \
Operation definitions. */\\n\\nvoid calc_sum_and_power3(int32_t x, int32_t y, \
int32_t z, int32_t * /* out */ result)\\n{\\n    *result = x * x * x + 3 * x \
* x * y + 3 * x * y * y + y * y * y + 3 * x * x\\n        * z + 6 * x * y * z \
+ 3 * y * y * z + 3 * x * z * z + 3 * y * z * z + z\\n        * z * z;\\n}\\n\
\"\>"], "Output",
 CellLabel->"Out[26]=",ExpressionUUID->"517205fd-c57c-436d-87a3-d2dd8ce3acd4"]
}, Open  ]],

Cell["\<\
The default preprocessor emulates the For loop in terms of expanded While \
construct - its implementation is in the FMTOptions` package and since \
setting any of the options overwrites the old option value, typical \
user-provided preprocessing functions will call the default one in order not \
to lose the capability it provides.\
\>", "Text",
 CellChangeTimes->{{3.7295057183050494`*^9, 3.729505795091649*^9}, {
  3.7295058351438494`*^9, 3.7295058687154493`*^9}, {3.7295059186982493`*^9, 
  3.729505995791849*^9}},ExpressionUUID->"1d2d5c81-7f3f-4db0-8393-\
dd305f59883e"]
}, Open  ]]
},
WindowSize->{808, 911},
WindowMargins->{{162, Automatic}, {Automatic, 30}},
FrontEndVersion->"11.3 for Mac OS X x86 (32-bit, 64-bit Kernel) (March 5, \
2018)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 352, 5, 54, "Subsection",ExpressionUUID->"16c46bba-dc55-4b32-bb63-389c38155a22"],
Cell[935, 29, 496, 9, 35, "Text",ExpressionUUID->"61f8085e-1a1e-44ce-8efd-0d2afe16fcb8"],
Cell[1434, 40, 196, 3, 30, "Input",ExpressionUUID->"1c63c929-62a5-414c-b53d-e9020ac47055"],
Cell[1633, 45, 328, 7, 58, "Text",ExpressionUUID->"49e5c6d6-0e23-400d-bf01-ab14ea17c3f7"],
Cell[CellGroupData[{
Cell[1986, 56, 265, 5, 30, "Input",ExpressionUUID->"75744d54-9543-4c08-ac73-9eb97c0a2ad7"],
Cell[2254, 63, 265, 8, 34, "Output",ExpressionUUID->"da7ba3af-e5b7-48a7-9ceb-6bbc0f7ecadc"]
}, Open  ]],
Cell[2534, 74, 323, 7, 58, "Text",ExpressionUUID->"f91e836d-0aa3-4344-8a51-8a08d45cb2be"],
Cell[2860, 83, 575, 13, 30, "Input",ExpressionUUID->"9ad736ca-0960-4a0d-ac75-16e9fd435dd8"],
Cell[3438, 98, 293, 6, 58, "Text",ExpressionUUID->"03feeaff-15c7-438e-ad69-c89ae6b3c51d"],
Cell[3734, 106, 514, 9, 81, "Text",ExpressionUUID->"ce5422d1-6b08-4dfd-9bb0-cff38edc3377"],
Cell[4251, 117, 376, 7, 30, "Input",ExpressionUUID->"2c78ecb3-6467-456f-8333-dfb50ec25ff7"],
Cell[4630, 126, 197, 3, 35, "Text",ExpressionUUID->"d9a95ebc-4068-40f8-9a63-e999232d46b8"],
Cell[4830, 131, 336, 6, 30, "Input",ExpressionUUID->"2011dc5e-d8d8-4ac8-ab36-86a34f7a11e5"],
Cell[5169, 139, 223, 5, 35, "Text",ExpressionUUID->"8f6bdbfa-2c78-4e69-93d9-69105985558e"],
Cell[5395, 146, 432, 7, 30, "Input",ExpressionUUID->"e4067743-3291-4a93-8a1c-cbd6bfaff6f0"],
Cell[5830, 155, 818, 14, 30, "Input",ExpressionUUID->"ca4a08a3-e3bc-49d4-b163-70b06f8c010d"],
Cell[CellGroupData[{
Cell[6673, 173, 1434, 26, 199, "Input",ExpressionUUID->"7244aafc-445e-4833-9236-4a17824f1daa"],
Cell[8110, 201, 349, 8, 34, "Output",ExpressionUUID->"d6d687a5-afb7-48fa-95ed-184b601b07f4"]
}, Open  ]],
Cell[8474, 212, 360, 8, 58, "Text",ExpressionUUID->"f012fa81-7314-4574-a8b3-0899277d16b0"],
Cell[CellGroupData[{
Cell[8859, 224, 328, 6, 30, "Input",ExpressionUUID->"08e2ce84-f5aa-4b86-a724-80608a2d95e1"],
Cell[9190, 232, 22025, 596, 266, "Output",ExpressionUUID->"69c39ad0-7278-4076-9d10-407cbff5b9fa"]
}, Open  ]],
Cell[31230, 831, 251, 4, 35, "Text",ExpressionUUID->"f4596866-d996-4e91-8e1a-7c62c67c37cc"],
Cell[CellGroupData[{
Cell[31506, 839, 395, 8, 30, "Input",ExpressionUUID->"eb4c6ac3-0c47-4003-831f-caac2c9ddc70"],
Cell[31904, 849, 385, 5, 308, "Output",ExpressionUUID->"dab9b77c-b830-4e06-91c7-a87f1beb779a"]
}, Open  ]],
Cell[32304, 857, 400, 6, 81, "Text",ExpressionUUID->"bd5c8098-42d9-486c-b633-09bc2b414254"],
Cell[32707, 865, 237, 6, 35, "Text",ExpressionUUID->"376636f8-8ea6-4160-8157-c8a1d3e36084"],
Cell[32947, 873, 565, 11, 52, "Input",ExpressionUUID->"193d7606-5e5a-47cb-802f-470a64c27b15"],
Cell[33515, 886, 550, 10, 104, "Text",ExpressionUUID->"fa0ab6e3-c0f8-4165-a6ce-62048aa10028"],
Cell[34068, 898, 306, 6, 58, "Text",ExpressionUUID->"c8b69731-b240-4b35-b78c-4c5772ec647d"],
Cell[34377, 906, 377, 7, 30, "Input",ExpressionUUID->"741ff627-8a85-48e6-8221-4b3c53c3d52b"],
Cell[34757, 915, 197, 3, 35, "Text",ExpressionUUID->"92707186-880c-4e89-b58d-fc39d0b09865"],
Cell[34957, 920, 369, 7, 30, "Input",ExpressionUUID->"70812cec-44e2-4306-8c32-cb03f4ffc032"],
Cell[35329, 929, 481, 7, 30, "Input",ExpressionUUID->"36f5f3e3-f615-4a0f-8216-c0ac58ad92d5"],
Cell[35813, 938, 425, 7, 115, "Input",ExpressionUUID->"0566b070-a97a-4199-a8dc-2aa87b5caf4e"],
Cell[36241, 947, 915, 15, 30, "Input",ExpressionUUID->"652011de-6bc4-45b8-bc3e-753464b3e1c7"],
Cell[CellGroupData[{
Cell[37181, 966, 1364, 22, 178, "Input",ExpressionUUID->"04de1c2e-fa4b-47ca-a405-13bfbd3877b5"],
Cell[38548, 990, 350, 8, 34, "Output",ExpressionUUID->"a4003d70-5b2f-45b8-8d96-81f9d320d144"]
}, Open  ]],
Cell[38913, 1001, 385, 7, 58, "Text",ExpressionUUID->"a09bed30-6669-4d7e-86f9-70b32dcf07e4"],
Cell[CellGroupData[{
Cell[39323, 1012, 264, 4, 30, "Input",ExpressionUUID->"abfaa02e-2ef3-4583-bf57-37cbe6ed9df9"],
Cell[39590, 1018, 16104, 415, 292, "Output",ExpressionUUID->"381fa0f7-e8c4-4b28-985a-ddaf63a0b6d1"]
}, Open  ]],
Cell[CellGroupData[{
Cell[55731, 1438, 398, 8, 30, "Input",ExpressionUUID->"67321467-88fe-4aef-b94b-379be442c5f1"],
Cell[56132, 1448, 344, 4, 392, "Output",ExpressionUUID->"f5564976-404e-4f73-a533-1772c3b81dc4"]
}, Open  ]],
Cell[56491, 1455, 343, 5, 81, "Text",ExpressionUUID->"6c98d15c-f2e5-4571-84f1-2aa28274caae"],
Cell[56837, 1462, 257, 5, 30, "Input",ExpressionUUID->"bc714c7c-758f-4bf7-a995-47f8bb0d4d41"],
Cell[57097, 1469, 162, 3, 35, "Text",ExpressionUUID->"e93e15f8-b675-4f30-bc94-8018c27051c5"],
Cell[57262, 1474, 189, 4, 30, "Input",ExpressionUUID->"cd04adc6-65aa-462f-8358-435f8baebcf3"],
Cell[57454, 1480, 558, 10, 81, "Text",ExpressionUUID->"27267746-2a9e-4859-a4a5-bd6ff342adf8"],
Cell[CellGroupData[{
Cell[58037, 1494, 292, 6, 33, "Input",ExpressionUUID->"8c0a32af-b274-42bd-aa33-f627c5fcb12a"],
Cell[58332, 1502, 672, 18, 34, "Output",ExpressionUUID->"5f4cb438-bedc-440a-a3c8-c10001b2796c"]
}, Open  ]],
Cell[59019, 1523, 281, 6, 58, "Text",ExpressionUUID->"838032b4-00cc-4c1e-972c-be362e3563fe"],
Cell[59303, 1531, 1210, 29, 241, "Input",ExpressionUUID->"2260a2d8-3c52-4943-8fd4-3938db503a37"],
Cell[60516, 1562, 287, 5, 30, "Input",ExpressionUUID->"38b55f45-9510-4599-b31b-0d3ec74f2be2"],
Cell[60806, 1569, 532, 8, 30, "Input",ExpressionUUID->"7ef3af92-96fe-4be7-b46b-dd2ecd7b0488"],
Cell[61341, 1579, 1442, 26, 115, "Input",ExpressionUUID->"10c848f6-db37-45ab-94a9-1d45ff3e7be4"],
Cell[CellGroupData[{
Cell[62808, 1609, 625, 15, 178, "Input",ExpressionUUID->"d8ccf03d-0c08-49e2-abc7-e7a9a7b5d449"],
Cell[63436, 1626, 350, 8, 34, "Output",ExpressionUUID->"938f3141-f1fc-4277-9d9b-320b99894e36"]
}, Open  ]],
Cell[63801, 1637, 224, 5, 35, "Text",ExpressionUUID->"5a06aad6-ab9e-4df7-973c-93976d4777f6"],
Cell[CellGroupData[{
Cell[64050, 1646, 292, 6, 30, "Input",ExpressionUUID->"0069fb66-0cf7-47e9-9f2f-af5828b605d9"],
Cell[64345, 1654, 491, 6, 287, "Output",ExpressionUUID->"517205fd-c57c-436d-87a3-d2dd8ce3acd4"]
}, Open  ]],
Cell[64851, 1663, 586, 10, 81, "Text",ExpressionUUID->"1d2d5c81-7f3f-4db0-8393-dd305f59883e"]
}, Open  ]]
}
]
*)

