(* ::Package:: *)

(*
    Formal Methods Toolkit.
    Copyright (C) 2018-2021 Inspirel.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)


Needs["FMTAnalytics`"];
Needs["FMTFormatters`"];
Needs["FMTGeneratorAda`"];
Needs["FMTGeneratorC`"];
Needs["FMTGeneratorCpp`"];
Needs["FMTGrammar`"];
Needs["FMTOptions`"];
Needs["FMTProofs`"];
Needs["FMTProver`"];
Needs["FMTSymbols`"];
Needs["FMTSystem`"];


BeginPackage["FMT`"];


newSystem::usage="newSystem[options] Creates and returns an empty model.\n"<>
"If options are not provided, the default values are used, as defined in the FMTOptions` package.";


getOptions::usage="getOptions[system] Returns the association of current options.";


setOptions::usage="setOptions[system,options] Overrides the whole association of current options.";


setOption::usage="setOption[system,option,value] Overrides a single option.";


declarePackage::usage="declarePackage[system,packageName,comment] Declares a new package within the system.\n"<>
"The packageName string and should follow the Wolfram convention for package naming.\n"<>
"The comment is an option that allows to associate a comment with the new package.";


defineDataTypes::usage="defineDataTypes[system,packageName,typeDefinitions,comment] Defines new types within the given package.\n"<>
"The packageName should be a string and should denote an already declared package.\n"<>
"Types are defined by list of rules (or associations) with mandatory key base and optional key invariant.\n"<>
"The comment is an option that allows to associate a comment with the whole group of data types.";


defineDataObjects::usage="defineDataObjects[system,packageName,objectNames,comment] Defines new data objects within the given package.\n"<>
"The packageName should be a string and should denote an already declared package.\n"<>
"The comment is an option that allows to associate a comment with the whole group of data objects.";


setDataObjectValues::usage="setDataObjectValues[system,settings] Sets values for the given data objects.\n"<>
"Settings should have the form {qualifiedObjectName->newValue, ...}.";


getDataObjectValues::usage="getDataObjectValues[system,objNames] Returns the list of rules with values for the given object names.\n"<>
"Object names should be given as a list of fully qualified symbols.";


declareOperations::usage="declareOperations[system,packageName,operationProfiles,comment] Declares new operation profiles,"<>
" possibly with additional information like pre- and post-conditions, within the given package.\n"<>
"The packageName should be a string and should denote an already declared package.\n"<>
"The comment is an option that allows to associate a comment with the whole group of operations.";


defineOperation::usage="defineOperation[system,packageName,operationName,localObjects,operationBody] Defines the given operation.\n"<>
"Package name should be a string and operation name should be a non-qualified symbol.\n"<>
"The localObject is a (possibly empty) list of rules defining local data objects, {name->type, ...} and body can be an arbitrary (and perhaps compound) expression.";


setGhostDomains::usage="setGhostDomains[system,domains] Sets domains for the given ghost symbol or symbols.\n"<>
"The domains parameter should be Element[symbol,domain] or a list of these.";


reportGrammarChecks::usage="reportGrammarChecks[system,packageNames,operationNames] Performs the grammar correctness check on the selected operations "<>
"and returns report in the form of nested association <|packageName-><|operationName->result, ...|>, ...|>.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are checked.";


showGrammarChecks::usage="showGrammarChecks[system,packageNames,operationNames] Presents the results of grammar checks in a user-friendly way.";


reportGrammarReductions::usage="grammarReductions[system,packageName,operationName] Returns the list of grammar reductions for the given operation.\n"<>
"The packageName should be a string and operationName should be a non-qualified symbol.";


showGrammarReductions::usage="showGrammarReductions[system,packageName,operationName] Presents the grammar reduction steps for the given operation in terms of tree diagrams.";


showLastReductionStep::usage="showLastReductionStep[system,packageName,operationName] Presents the last step of grammar reduction process for the given operation.";


reportGrammar::usage="reportGrammar[system,packageName,operationName] Returns the internal grammar representation for the given operation.";


showGrammar::usage="showGrammar[system,packageName,operationName] Presents the grammar tree in a graphical form.";


reportNameChecks::usage="reportNameChecks[system,packageNames,operationNames] Performs checks of name conflicts in the model\n"<>
"and returns report in the form <|packageName->{packageConflictingNamesList,<|operationName->conflictingNamesList,...|>},...|>.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are checked.";


showNameChecks::usage="showNameChecks[system,packageNames,operationNames] Presents the results of name conflict checks in a user-friendly way.";


reportFlowChecks::usage="reportFlowChecks[system,packageNames,operationNames] Performs checks of data flow conflicts in the model\n"<>
"and returns report in the form <|packageName-><|operationName->{badOperation,...},...|>},...|>,\n"<>
"where badOperation is {index,description}.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are checked.";


showFlowChecks::usage="showFlowChecks[system,packageNames,operationNames] Presents the results of flow checks in a user-friendly way.";


reportCallTable::usage="reportCallTable[system,packageNames,operationNames] Reports the operation call dependencies as an Association of rules caller->callee.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are analyzed.";


showCallTable::usage="showCallList[system,packageNames,operationNames] Presents call information in the form of a table.";


reportCallGraph::usage="reportCallGraph[system,packageNames,operationNames] Reports the operation call dependencies as a flat list of rules caller->callee.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are analyzed.";


showCallGraph::usage="showCallGraph[system,packageNames,operationNames] Presents call information in the form of a graph.";


reportCallCycles::usage="reportCallCycles[system,packageNames,operationNames] Find cycles in the complete call graph.\n"<>
"The result is returned as a list {completeCallGraph,{cycle,...}}.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are checked.";


showCallCycles::usage="showCallCycles[system,packageNames,operationNames] Presents call cycles in a graphical way.";


reportControlFlow::usage="reportControlFlows[system,packageName,operationName,highlighted] Creates the control flow graph for the selected operation,\n"<>
"where graph is a list of directed edges between vertices that relate to individual statements.\n"<>
"The highlighted option is a set of identifiers (numbers) of expression parts that need to be highlighted.";


showControlFlow::usage="showControlFlow[system,packageName,operationName] Presents control flows for the selected operation as graphs.";


reportDataAliasing::usage="reportDataAliasing[system,packageNames,operationNames] Reports the cases of data aliasing by parameter passing.\n"<>
"The report has the form of nested association <|packageName-><|operationName->{{index,{object1,...}},...},...|>,...|>.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are checked.";


showDataAliasing::usage="showDataAliasing[system,packageNames,operationNames] Presents the results of data aliasing checks in a user-friendly way.";


reportProofStatements::usage="reportProofStatements[system,packageName,operationName] Reports the lists of statements and verification conditions for a single operations.\n"<>
"This function is provided as an inspection aid.\n"<>
"The report has the form of nested association <|packageName-><|operationName->{{{index,{statement,...},{{description,condition},...}},{variable->domain,...}}, ...|>, ...|>.";


showProofStatements::usage="showProofStatements[system,packageName,operationName] Presents the list of proof statements in a grid.";


reportProofs::usage="reportProofs[system,packageNames,operationNames] Reports the satisfiability of selected operations with regard to their verification conditions.\n"<>
"The report has the form of nested association {packageName->{operationName->result, ...}, ...}, where result is either True or a list of failure descriptions.\n"<>
"The packageNames and (unqualified) operationNames are options, they can denote single entities or lists of entities, and by default are both set to All, which means that all operations from all packages are checked.";


showProofChecks::usage="showProofChecks[system,packageNames,operationNames] Presents the results of proof checks in a user-friendly way.";


showProofResults::usage="showProofResults[system,packageName,operationName,onlyFailed] Presents the proof results for the single selected operation.\n"<>
"The onlyFailed option, if True (default), filters the results and shows only failing checks.";


runAllChecks::usage="runAllChecks[system,packageNames,operationNames] Runs grammar and proof checks for all or selected operations.";


generateCode::usage="generateCode[system,language,packageNames,translationPolicies] Returns <|fileName->fileContent,...|> for the given packages, according to current system's translation policies.\n"<>
"The packageNames is an option that can denote a single name, list of names or All (which is default).\n"<>
"The translationPolicies is an option (Association) that is merged with system options to override some of the code generation choices.";


saveCode::usage="saveCode[system,language,packageNames,destinationDirectory,translationPolicies] Generates code for selected packages and saves source files in the destination directory.\n"<>
"The packageNames is an option that can denote a single name, list of names or All (which is default).\n"<>
"The destinationDirectory is an option, by default an empty string, which is interpreted as a current directory.\n"<>
"The translationPolicies is an option (Association) that is merged with system options to override some of the code generation choices.";


Begin["`Private`"];


(* Helpers for validating package declarations *)
validatePackageNameComponent[componentName_String]:=Module[
	{
		allowedAsFirstCharacter=CharacterRange["A","Z"],
		allowedAsFurtherCharacter=Join[
			CharacterRange["A","Z"],
			CharacterRange["a","z"],
			CharacterRange["0","9"]
		],
		state
	},
	
	If[componentName=="",
		Message[validatePackageName::invcomponent],
		
		state=allowedAsFirstCharacter;
		Scan[
			Function[c, (* c is subsequent character from name component *)
				If[Not[MemberQ[state,c]],
					Message[validatePackageName::invchar,c]
				];
				state=allowedAsFurtherCharacter
			],
			Characters[componentName]
		]
	]
]

validatePackageDeclaration[packageName_String]:=(
	If[StringTake[packageName,-1]!="`",
		Message[validatePackageName::badend],
		
		Scan[validatePackageNameComponent,StringSplit[packageName,"`"]]
	]
)

validatePackageName::invcomponent="Package name component cannot be empty.";
validatePackageName::invchar="Invalid character in package name: `1`";
validatePackageName::badend="Package name should end with a backtick.";


(* Helper for validating simple numeric types derived from base types *)
validateNumberTypeDefinition[typeDef_FMTSymbols`number]:=(
	Scan[
		Function[d, (* definition component *)
			If[Head[d]=!=Rule,
				Message[validateTypeDefinitions::notypedef,d],
				
				If[(First[d]=!=FMTSymbols`base)&&
					(First[d]=!=FMTSymbols`invariant),
					Message[validateTypeDefinitions::notypeaspect,First[d]],
					
					If[(First[d]===FMTSymbols`base)&&
						(Last[d]=!=FMTSymbols`integerT),
						Message[validateTypeDefinitions::badbasetype,Last[d]]
					]
				]
			]
		],
		typeDef
	];
	
	If[Not[MemberQ[Cases[typeDef,Rule[a_,b_]->a,{1}],FMTSymbols`base]],
		Message[validateTypeDefinitions::nobase,Last[t]]
	]
)

validateEnumTypeDefinition[typeDef_FMTSymbols`enum]:=(
	Scan[
		Function[e, (* enum value *)
			If[Head[e]=!=Symbol,
				Message[validateTypeDefinitions::badenum,e]
			]
		],
		typeDef
	]
)

(* Helper for validating array types *)
validateArrayTypeDefinition[previousTypesInThisGroup_List,typeDef_FMTSymbols`array]:=
	If[Length[typeDef]!=2,
		Message[validateTypeDefinitions::badarray,typeDef],
		
		If[FreeQ[previousTypesInThisGroup,First[typeDef]],
			FMTSystem`validateTypeIsKnown[sys,packageName,First[typeDef]]
		];
		
		If[Head[Last[typeDef]]=!=Integer,
			Message[validateTypeDefinitions::badarray,typeDef]
		]
	]

(* Helper for validating structure types *)
validateStructureTypeDefinition[sys_,packageName_String,previousTypesInThisGroup_List,typeDef_FMTSymbols`structure]:=Module[
	{
		previousFields={}
	},
	
	Scan[
		Function[field, (* field definition *)
			If[Head[field]=!=Rule,
				Message[validateTypeDefinitions::badfield,field],
				
				If[Head[First[field]]=!=Symbol,
					Message[validateTypeDefinitions::badfield,field],
					
					If[MemberQ[previousFields,First[field]],
						Message[validateTypeDefinitions::replacefield,field]
					];
					
					If[FreeQ[previousTypesInThisGroup,Last[field]],
						FMTSystem`validateTypeIsKnown[sys,packageName,Last[field]]
					];
					
					AppendTo[previousFields,First[field]]
				]
			]
		],
		typeDef
	]
]

validateTypeDefinitions[sys_,packageName_String,types_]:=Module[
	{
		(* type definitions are accepted as a whole group,
			but types earlier in the list can be referred to by further definitions,
			\[Rule] this list allows to keep track of already validated type names *)
		previousTypesInThisGroup={}
	},
	
	Scan[
		Function[t,
			If[Head[t]=!=Rule,
				Message[validateTypeDefinitions::notypedef,t],
				If[Head[First[t]]=!=Symbol,
					Message[validateTypeDefinitions::notypename,First[t]],
					Switch[Head[Last[t]],
						FMTSymbols`number,
							(* this should be a definition of a simple numeric type,
								given as a list of properties *)
							validateNumberTypeDefinition[Last[t]],
						
						FMTSymbols`enum,
							(* this should be an enumeration definition *)
							validateEnumTypeDefinition[Last[t]],
						
						FMTSymbols`array,
							(* this should be an array type definition,
								where Last[t] is array[type,size] *)
							validateArrayTypeDefinition[previousTypesInThisGroup,Last[t]],
						
						FMTSymbols`structure,
							(* this should be a structure definition,
								where Last[t] is FMTSymbols`structure[...] *)
							validateStructureTypeDefinition[sys,packageName,previousTypesInThisGroup,Last[t]],
						
						_,
							Message[validateTypeDefinitions::notypedef,Last[t]]
					]
				]
			];
			
			AppendTo[previousTypesInThisGroup,t]
		],
		types
	]
]
	
validateTypeDefinitions::notypedef="`1` is not a proper type defining rule.";
validateTypeDefinitions::notypename="`1` is not a proper type name.";
validateTypeDefinitions::notypeaspect="`1` is not a proper type aspect.";
validateTypeDefinitions::badenum="`1` is not a proper enumeration value.";
validateTypeDefinitions::badbasetype="`1` is not a valid base type.";
validateTypeDefinitions::nobase="`1` does not specify any base type.";
validateTypeDefinitions::badarray="`1` is not a proper array type definition.";
validateTypeDefinitions::badfield="`1` is not a proper structure field definition.";
validateTypeDefinitions::replacefield="`1` replaces an already defined field.";


(* This helper is used to validate both package data and operations' local variables *)
validateObjectDefinitions[sys_,packageName_String,localObjects_]:=
	Scan[
		Function[obj,
			If[Head[obj]=!=Rule,
				Message[validateObjectDefinitions::noobjdef,obj],
				If[Head[First[obj]]=!=Symbol,
					Message[validateObjectDefinitions::noobjname,First[obj]],
					With[
						{
							typeName=Last[obj]
						},
						
						If[Head[typeName]=!=Symbol,
							Message[validateObjectDefinitions::noobjtype,typeName],
							FMTSystem`validateTypeIsKnown[sys,packageName,typeName]
						]
					]
				]
			]
		],
		localObjects
	]

validateObjectDefinitions::noobjdef="`1` is not a proper data object defining rule.";
validateObjectDefinitions::noobjname="`1` is not a proper data object name.";
validateObjectDefinitions::noobjtype="`1` is not a proper type name.";

(* Helper, checks for duplicates in the list of operation's local variables *)
validateLocalVarsNoDuplicates[sys_,qOperationName_Symbol,localObjects_]:=With[
	{
		params=FMTSystem`operationParams[sys,qOperationName]
	},
	
	If[Not[DuplicateFreeQ[Join[params[[All,1]],localObjects[[All,1]]]]],
		Message[validateLocalVarsNoDuplicates::dups]
	]
]

validateLocalVarsNoDuplicates::dups="The operation declares local variables that have duplicates or collide with its parameters.";


validateObjectSettings[sys_,settings_List]:=(
	Scan[
		Function[setting,
			If[Head[setting]=!=Rule,
				Message[validateObjectSettings::norules]
			]
		],
		settings
	];
	FMTSystem`validateObjectsAreKnown[sys,settings[[All,1]]]
)

validateObjectSettings::norules="Object settings should be rules objName->value.";


(* Helper, check if the given element is wrapped in one of parameter mode wrappers *)
parameterModeValidQ[param_]:=
	MemberQ[{FMTSymbols`in,FMTSymbols`inOut,FMTSymbols`out},Head[param]]

validateSingleOperationProfile[sys_,packageName_String,operationProfile_]:=Module[
	{
		opHead=Head[operationProfile],
		paramNames={}
	},
	
	Scan[
		Function[profileElement,
			If[Head[profileElement]===Rule,
				(* parameter must have form:
					name\[Rule]mode[type]
				*)
				If[Head[profileElement[[1]]]===Symbol&&
					parameterModeValidQ[profileElement[[2]]],
					
					FMTSystem`validateTypeIsKnown[sys,packageName,profileElement[[2,1]]];
					AppendTo[paramNames,profileElement[[1]]],
					
					Message[validateNewOperationProfiles::badprof,opHead]
				],
				
				(* parameter must have form:
					mode[name\[Rule]type]
				*)
				If[parameterModeValidQ[profileElement]&&
					Head[profileElement[[1]]]===Rule&&
					Head[profileElement[[1,1]]]===Symbol,
					
					FMTSystem`validateTypeIsKnown[sys,packageName,profileElement[[1,2]]];
					AppendTo[paramNames,profileElement[[1,1]]],
					
					Message[validateNewOperationProfiles::badprof,opHead]
				]
			]
		],
		operationProfile
	];
	
	If[Not[DuplicateFreeQ[paramNames]],
		Message[validateNewOperationProfiles::dups,opHead]
	]
]

validateSingleOperationAdditionalInfo[sys_,operationName_,operationInfo_]:=
	Scan[
		Function[info,
			If[Not[
				Head[info]===Rule&&
				(info[[1]]===FMTSymbols`pre||info[[1]]===FMTSymbols`post)
				],
				Message[validateNewOperationProfiles::badinfo,operationName]
			]
		],
		operationInfo
	]

validateNewOperationProfiles[sys_,packageName_,operationProfiles_]:=
	Scan[
		Function[opProfile,
			With[
				{
					opHead=Head[opProfile]
				},
				
				(* operation profile can be either stand-alone or a rule that defines
					the profile (operation name with parameters) with additional information,
					like pre- and post-conditions *)
				If[opHead===Rule&&
					Head[opProfile[[1]]]=!=Symbol&&
					Head[Head[opProfile[[1]]]]===Symbol,
					
					If[Head[opProfile[[2]]]===List,
						validateSingleOperationProfile[sys,packageName,opProfile[[1]]];
						validateSingleOperationAdditionalInfo[sys,Head[opProfile[[1]]],opProfile[[2]]],
						
						Message[validateNewOperationProfiles::badinfo,Head[opProfile[[1]]]]
					],
					
					If[opHead=!=Symbol&&Head[opHead]===Symbol,
						validateSingleOperationProfile[sys,packageName,opProfile],
						
						Message[validateNewOperationProfiles::notprofs]
					]
				]
			]
		],
		operationProfiles
	]

validateNewOperationProfiles::notprofs="Argument 3 is not a list of proper operation profiles.";
validateNewOperationProfiles::badprof="Operation `1` does not have a proper parameter list.";
validateNewOperationProfiles::badinfo="Operation `1` does not have a proper additional info list.";
validateNewOperationProfiles::dups="Operation `1` has duplicate parameter names.";


validatePackageSelector[packageNames_]:=
	If[packageNames=!=All,
		If[Head[packageNames]=!=String,
			If[Head[packageNames]=!=List,
				Message[validateOperationsSelector::invpkgs],
				Scan[
					Function[pkgName,
						If[Head[pkgName]=!=String,
							Message[validateOperationsSelector::invpkgs]
						]
					],
					packageNames
				]
			]
		]
	]

validateOperationsSelector[packageNames_,operationNames_]:=(
	validatePackageSelector[packageNames];
	
	If[operationNames=!=All,
		If[Head[operationNames]=!=Symbol,
			If[Head[operationNames]=!=List,
				Message[validateOperationsSelector::invops],
				Scan[
					Function[opName,
						If[Head[opName]=!=Symbol,
							Message[validateOperationsSelector::invops]
						]
					],
					operationNames
				]
			]
		]
	];
)

validateOperationsSelector::invpkgs="Invalid selection of packages, should be All, string or list of strings.";
validateOperationsSelector::invops="Invalid selection of operations, should be All, symbol or list of symbols.";


validateLanguage[language_String]:=
	If[Not[MemberQ[{"Ada","C","C++"},language]],
		Message[validateLanguage::invlan,language]
	]

validateLanguage::invlan="`1` is not a recognized language.";


validateSingleGhostDomain[domain_]:=
	If[Not[MatchQ[domain,Element[_Symbol,Booleans|Integers]]],
		Message[validateGhostDomains::invdom,domain]
	]

validateGhostDomains[domains_]:=
	If[Head[domains]===List,
		Scan[validateSingleGhostDomain,domains],
		validateSingleGhostDomain[domains]
	]

validateGhostDomains::invdom="`1` is not a valid domain assignment.";


(*
 * Main operations.
 *)


newSystem[]:=FMTSystem`create[FMTOptions`defaultOptions]
newSystem[options_Association]:=FMTSystem`create[options]


getOptions[system_]:=Module[
	{
		sys
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	FMTSystem`fmtGetCurrentOptions[sys]
]


setOptions[system_,options_Association]:=Module[
	{
		sys
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	FMTSystem`fmtSetCurrentOptions[sys,options];
	
	system[[1]]=sys;
]

SetAttributes[setOptions,HoldFirst]


setOption[system_,option_Symbol,value_]:=Module[
	{
		sys,
		options
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	options=FMTSystem`fmtGetCurrentOptions[sys];
	options[option]=value;
	FMTSystem`fmtSetCurrentOptions[sys,options];
	
	system[[1]]=sys;
]

SetAttributes[setOption,HoldFirst]


declarePackage[system_,packageName_String,OptionsPattern[{comment->Null}]]:=Module[
	{
		sys,
		commentOpt=OptionValue[comment]
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validatePackageDeclaration[packageName],Return[]];
	
	Check[FMTSystem`validateNoPackageDuplicates[sys,packageName],Return[]];
	
	FMTSystem`fmtDeclarePkg[sys,packageName,commentOpt];
	
	system[[1]]=sys;
]

SetAttributes[declarePackage,HoldFirst]


defineDataTypes[system_,packageName_String,types_List,OptionsPattern[{comment->Null}]]:=Module[
	{
		sys,
		commentOpt=OptionValue[comment]
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[FMTSystem`validatePackageIsKnown[sys,packageName],Return[]];
	
	Check[validateTypeDefinitions[sys,packageName,types],Return[]];
	
	FMTSystem`fmtDefineTypes[sys,packageName,types,commentOpt];	
	
	system[[1]]=sys;
]

SetAttributes[defineDataTypes,HoldFirst]


defineDataObjects[system_,packageName_String,objects_List,OptionsPattern[{comment->Null}]]:=Module[
	{
		sys,
		commentOpt=OptionValue[comment]
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[FMTSystem`validatePackageIsKnown[sys,packageName],Return[]];
	
	Check[validateObjectDefinitions[sys,packageName,objects],Return[]];
	
	FMTSystem`fmtDefineData[sys,packageName,objects,commentOpt];	
	
	system[[1]]=sys;
]

SetAttributes[defineDataObjects,HoldFirst]


setDataObjectValues[system_,settings_List]:=Module[
	{
		sys
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateObjectSettings[sys,settings],Return[]];
	
	FMTSystem`fmtSetData[sys,settings];
	
	system[[1]]=sys;
]

SetAttributes[setDataObjectValues,HoldFirst]


getDataObjectValues[system_,objNames_List]:=Module[
	{
		sys
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[FMTSystem`validateObjectsAreKnown[sys,objNames],Return[]];
	
	FMTSystem`fmtGetData[sys,objNames]
]

SetAttributes[getDataObjectValues,HoldFirst]


declareOperations[system_,packageName_String,operationProfiles_List,OptionsPattern[{comment->Null}]]:=Module[
	{
		sys,
		protectedProfiles, (* pre- and post-conditions within Hold *)
		commentOpt=OptionValue[comment]
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[FMTSystem`validatePackageIsKnown[sys,packageName],Return[]];
	
	(* wrap pre- and post-conditions within Hold, there might be code
		that involves Exists|ForAll, which will try to evaluate in the wrong context
		without such protection
	*)
	protectedProfiles=Hold[operationProfiles];
	protectedProfiles=protectedProfiles/.{
		Rule[FMTSymbols`pre,preCond_]->Rule[FMTSymbols`pre,Hold[preCond]],
		Rule[FMTSymbols`post,postCond_]->Rule[FMTSymbols`post,Hold[postCond]]
	};
	protectedProfiles=ReleaseHold[protectedProfiles];
	
	Check[validateNewOperationProfiles[sys,packageName,protectedProfiles],Return[]];
	
	FMTSystem`fmtDeclareOps[sys,packageName,protectedProfiles,commentOpt];
	
	system[[1]]=sys;
]

SetAttributes[declareOperations,HoldAll];


(* Helper, returns True if the check result for name or flow conflicts is empty
	(that is, there are no conflicts or errors) *)
namesCorrectQ[result_]:=With[
	{
		consolidatedResult=Flatten@Map[
			Function[pkgResult, (* pkgResult is {pkgConflictNameList,<|opName\[Rule]opConflictNameList,...|>} *)
				{
					pkgResult[[1]], (* names conflicting at the package level *)
					Values[pkgResult[[2]]] (* names conflicting within package operations *)
				}
			],
			Values[result]
		]
	},
	
	consolidatedResult=={}
]

(* Helper, for generic handling of results that have empty sets when correct *)
allReportsEmptyQ[result_]:=With[
	{
		consolidatedResult=Flatten[Values[Values[result]]]
	},
	
	consolidatedResult=={}
]


defineOperation[system_,packageName_String,operationName_Symbol,localObjects_List,operationBody_]:=Module[
	{
		sys,
		qOpName,
		currentOptions,
		grammarCheck,
		namesCheck,
		flowChecks
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[FMTSystem`validatePackageIsKnown[sys,packageName],Return[]];
	
	qOpName=FMTSystem`fullyQualifiedName[packageName,operationName];
	
	Check[FMTSystem`validateOperationIsKnown[sys,qOpName],Return[]];
	
	Check[validateObjectDefinitions[sys,packageName,localObjects],Return[]];
	Check[validateLocalVarsNoDuplicates[sys,qOpName,localObjects],Return[]];
	
	FMTSystem`fmtDefineOp[sys,packageName,qOpName,localObjects,operationBody];
	
	system[[1]]=sys;
	
	currentOptions=FMTSystem`fmtGetCurrentOptions[sys];
	If[currentOptions[FMTOptions`showInteractiveButtons],
		grammarCheck=FMTGrammar`grammarCorrectQ[
			FMTSystem`typeHierarchy[sys],
			FMTSystem`allEnumValues[sys],
			FMTSystem`resolvedOperationBody[sys,packageName,qOpName,False,{}]
		];
		If[grammarCheck,
			namesCheck=reportNameChecks[system,packageNames->packageName,operationNames->operationName];
			If[namesCorrectQ[namesCheck],
				flowChecks=reportFlowChecks[system,packageNames->packageName,operationNames->operationName];
				If[allReportsEmptyQ[flowChecks],
					Style["Grammar, names and flows OK.",Green],
					
					Row[
						{
							Style["Grammar and names OK, ",Green],
							Style["but there are flow errors.",Red],
							" ",
							Button["  Show details  ",
								CellPrint[
									ExpressionCell[
										FMTFormatters`formatFlowChecks[system,flowChecks],
										"Output"
									]
								]
							]
						}
					]
				],
				
				(* grammar OK, but there are name conflicts *)
				Row[
					{
						Style["Grammar OK, ",Green],
						Style["but there are name conflicts.",Red],
						" ",
						Button["  Show details  ",
							CellPrint[
								ExpressionCell[
									FMTFormatters`formatNameChecks[namesCheck],
									"Output"
								]
							]
						]
					}
				]
			],
			
			Row[
				{
					Style["Grammar failed.  ",Red],
					Button["  Show grammar reductions  ",
						CellPrint[
							ExpressionCell[
								showGrammarReductions[system,packageName,operationName],
								"Output"
							]
						]
					]
				}
			]
		]
	]
]

SetAttributes[defineOperation,HoldAll];


setGhostDomains[system_,domains_]:=Module[
	{
		sys
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateGhostDomains[domains],Return[]];
	
	If[Head[domains]===List,
		FMTSystem`fmtSetGhostDomains[sys,domains],
		FMTSystem`fmtSetGhostDomains[sys,List[domains]]
	];
	
	system[[1]]=sys;
]

SetAttributes[setGhostDomains,HoldFirst]


(* Helper, extract distinct duplicates from the given list *)
findDuplicates[lst_List]:=
	Select[
		Tally[ToString/@lst],
		Function[t,t[[2]]>1]
	][[All,1]]


reportNameChecks[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator->None}]]:=Module[
	{
		sys,
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress=OptionValue[progressIndicator],
		systemPkgNames, (* all package names in the system *)
		pkgLevelNames,
		pkgLevelNamesCache=<||>,
		packageConflicts=<||>,
		operationConflicts,
		singleOpResult,
		totalCount=0,
		count=0
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateOperationsSelector[pkgNames,opNames],Return[]];
	
	If[progress=!=None,
		(* count all items for use with progress indicator: *)
		FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
			Function[dummy,++totalCount]
		]
	];
	
	systemPkgNames=Map[
		Function[pn, (* pn is any package name in the system *)
			StringTrim[pn,"`"]
		],
		FMTSystem`allPackageNames[sys]
	];
	
	operationConflicts=FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
		Function[{pkgName,opName},
			With[
				{
					qOpName=FMTSystem`fullyQualifiedName[pkgName,opName]
				},
				
				(* perform package-level checks only once for each encountered package name *)
				If[KeyExistsQ[pkgLevelNamesCache,pkgName],
					pkgLevelNames=pkgLevelNamesCache[pkgName],
					
					(* the given package is analyzed for the first time *)
					pkgLevelNames=FMTSystem`packageLevelNames[sys,pkgName];
					
					(* pick only duplicate names defined in the given package *)
					packageConflicts[pkgName]=findDuplicates[
						Join[
							FMTSymbols`standardTypes,
							pkgLevelNames
						]
					];
					
					pkgLevelNames=Union[pkgLevelNames];
					pkgLevelNamesCache[pkgName]=pkgLevelNames
				];
				
				(* pick duplicates from all non-qualified names
					visible from within the given operation *)
				singleOpResult=findDuplicates[
					Join[
						FMTSymbols`standardTypes,
						systemPkgNames,
						pkgLevelNames,
						FMTSystem`operationLocals[sys,qOpName][[All,1]]
					]
				];
				
				If[progress=!=None,
					progress[{pkgName,opName},{++count,totalCount}]
				];
				
				singleOpResult
			]
		]
	];
	
	Merge[{packageConflicts,operationConflicts},
		Function[mergeList,
			(* this function performs an "outer-join" by artificially extending
				the result list, as packageConflicts might be shorter than operationConflicts,
				in which case Merge calls with a single argument for those entries
				from operationConflicts that do not have a counterpart in packageConflicts *)
			If[Length[mergeList]>1,
				mergeList,
				{{},mergeList[[1]]}
			]
		]
	]
]


showNameChecks[system_,OptionsPattern[{packageNames->All,operationNames->All}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		count=0,totalCount,
		result
	},
	
	PrintTemporary["checking name conflicts..."];
	Monitor[
		result=reportNameChecks[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Row[{ProgressIndicator[count,{0,totalCount}]," ",count,"/",totalCount}]
	];
	
	If[namesCorrectQ[result],
		Style["No name conflicts.",Green],
		Row[{
			Style["Name conflicts found. ",Red],
			Button["  Show details  ",
				Function[{},
					Print[FMTFormatters`formatNameChecks[result]]
				]
			]
		}]
	]
]


reportGrammarChecks[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator->None}]]:=Module[
	{
		sys,
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress=OptionValue[progressIndicator],
		singleResult,
		totalCount=0,
		count=0
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateOperationsSelector[pkgNames,opNames],Return[]];
	
	If[progress=!=None,
		(* count all items for use with progress indicator: *)
		FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
			Function[dummy,++totalCount]
		]
	];
	
	FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
		Function[{pkgName,opName},
			With[
				{
					qOpName=FMTSystem`fullyQualifiedName[pkgName,opName]
				},
				
				singleResult=If[FMTSystem`operationExists[sys,qOpName],
					FMTGrammar`grammarCorrectQ[
						FMTSystem`typeHierarchy[sys],
						FMTSystem`allEnumValues[sys],
						FMTSystem`resolvedOperationBody[sys,pkgName,qOpName,False,{}]
					],
					
					False
				];
				
				If[progress=!=None,
					progress[{pkgName,opName},{++count,totalCount}]
				];
				
				singleResult
			]
		]
	]
]


showGrammarChecks[system_,OptionsPattern[{packageNames->All,operationNames->All}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		count=0,totalCount,
		result,consolidatedResult
	},
	
	PrintTemporary["checking grammar..."];
	Monitor[
		result=reportGrammarChecks[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Row[{ProgressIndicator[count,{0,totalCount}]," ",count,"/",totalCount}]
	];
	
	consolidatedResult=And@@Flatten[Values[Values[result]]];
	If[consolidatedResult,
		Style["Grammar OK.",Green],
		Row[{
			Style["Grammar failed. ",Red],
			Button["  Show details  ",
				Function[{},
					Print[FMTFormatters`formatGrammarChecks[system,result]]
				]
			]
		}]
	]
]


reportGrammarReductions[system_,packageName_String,operationName_Symbol]:=Module[
	{
		sys,
		qOpName
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[FMTSystem`validatePackageIsKnown[sys,packageName],Return[]];
	
	qOpName=FMTSystem`fullyQualifiedName[packageName,operationName];
	
	Check[FMTSystem`validateOperationIsKnown[sys,qOpName,bodyRequired->True],Return[]];
	
	FMTGrammar`reductionList[
		FMTSystem`typeHierarchy[sys],
		FMTSystem`allEnumValues[sys],
		FMTSystem`resolvedOperationBody[sys,packageName,qOpName,True,{}]
	]
]


showGrammarReductions[system_,packageName_String,operationName_Symbol]:=
	FMTFormatters`formatGrammarReductions[
		reportGrammarReductions[system,packageName,operationName]
	]


showLastReductionStep[system_,packageName_String,operationName_Symbol]:=
	FMTFormatters`formatGrammar[
		Last[reportGrammarReductions[system,packageName,operationName]]
	]


reportGrammar[system_,packageName_,operationName_]:=Module[
	{
		sys,
		qOpName
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[FMTSystem`validatePackageIsKnown[sys,packageName],Return[]];
	
	qOpName=FMTSystem`fullyQualifiedName[packageName,operationName];
	
	Check[FMTSystem`validateOperationIsKnown[sys,qOpName,bodyRequired->True],Return[]];
	
	FMTSystem`resolvedOperationBody[sys,packageName,qOpName,True,{}]
]


showGrammar[system_,packageName_,operationName_]:=
	FMTFormatters`formatGrammar[
		reportGrammar[system,packageName,operationName]
	]


reportFlowChecks[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator->None}]]:=Module[
	{
		sys,
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress=OptionValue[progressIndicator],
		singleResult,
		totalCount=0,
		count=0,
		result
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateOperationsSelector[pkgNames,opNames],Return[]];
	
	If[progress=!=None,
		(* count all items for use with progress indicator: *)
		FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
			Function[dummy,++totalCount]
		]
	];
	
	result=FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
		Function[{pkgName,opName},
			With[
				{
					qOpName=FMTSystem`fullyQualifiedName[pkgName,opName]
				},
				
				singleResult=If[FMTSystem`operationExists[sys,qOpName],
					FMTAnalytics`dataFlow[
						FMTSystem`resolvedOperationBody[sys,pkgName,qOpName,True,{}],
						FMTSystem`operationParamModes[sys,qOpName]
					],
					
					{}
				];
				
				If[progress=!=None,
					progress[{pkgName,opName},{++count,totalCount}]
				];
				
				singleResult
			]
		]
	];
	
	result
]


showFlowChecks[system_,OptionsPattern[{packageNames->All,operationNames->All}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		count=0,totalCount,
		result
	},
	
	PrintTemporary["checking flows..."];
	Monitor[
		result=reportFlowChecks[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Row[{ProgressIndicator[count,{0,totalCount}]," ",count,"/",totalCount}]
	];
	
	If[allReportsEmptyQ[result],
		Style["Flows OK.",Green],
		Row[{
			Style["Flow checks failed. ",Red],
			Button["  Show details  ",
				Function[{},
					Print[FMTFormatters`formatFlowChecks[system,result]]
				]
			]
		}]
	]
]


reportCallTable[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator->None}]]:=Module[
	{
		sys,
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress=OptionValue[progressIndicator],
		opBody,
		singleResult,
		totalCount=0,
		count=0,
		result
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateOperationsSelector[pkgNames,opNames],Return[]];
	
	If[progress=!=None,
		(* count all items for use with progress indicator: *)
		FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
			Function[dummy,++totalCount]
		]
	];
	
	result=FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
		Function[{pkgName,opName},
			With[
				{
					qOpName=FMTSystem`fullyQualifiedName[pkgName,opName]
				},
				
				If[FMTSystem`operationExists[sys,qOpName],
					opBody=FMTSystem`resolvedOperationBody[sys,pkgName,qOpName,False,{}];
					
					singleResult=FMTAnalytics`outgoingCalls[opBody],
					
					singleResult={}
				];
				
				If[progress=!=None,
					progress[{pkgName,opName},{++count,totalCount}]
				];
				
				Thread[DirectedEdge[qOpName,singleResult]]
			]
		]
	];
	
	result
]


showCallTable[system_,OptionsPattern[{packageNames->All,operationNames->All}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		count=0,totalCount,
		result
	},
	
	PrintTemporary["computing call table..."];
	Monitor[
		result=reportCallTable[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Row[{ProgressIndicator[count,{0,totalCount}]," ",count,"/",totalCount}]
	];
	
	FMTFormatters`formatCallTable[result]
]


reportCallGraph[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator->None}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress=OptionValue[progressIndicator],
		result
	},
	
	result=reportCallTable[system,
		packageNames->pkgNames,
		operationNames->opNames,
		progressIndicator->progress
	];
	
	Flatten[Values[Values[result]]]
]


showCallGraph[system_,OptionsPattern[{packageNames->All,operationNames->All,entryPoint->None}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		entry=OptionValue[entryPoint],
		count=0,totalCount,
		result
	},
	
	PrintTemporary["computing call graph..."];
	Monitor[
		result=reportCallGraph[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Row[{ProgressIndicator[count,{0,totalCount}]," ",count,"/",totalCount}]
	];
	
	FMTFormatters`formatCallGraph[result,entry]
]


reportCallCycles[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator->None}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress=OptionValue[progressIndicator],
		callGraph
	},
	
	callGraph=reportCallGraph[system,
		packageNames->pkgNames,
		operationNames->opNames,
		progressIndicator->progress
	];
	
	{callGraph,FindCycle[callGraph,Infinity,All]}
]


showCallCycles[system_,OptionsPattern[{packageNames->All,operationNames->All,entryPoint->None}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		entry=OptionValue[entryPoint],
		count=0,totalCount,
		result
	},
	
	PrintTemporary["computing call graph..."];
	Monitor[
		result=reportCallCycles[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Row[{ProgressIndicator[count,{0,totalCount}]," ",count,"/",totalCount}]
	];
	
	FMTFormatters`formatCallCycles[result,entry]
]


reportControlFlow[system_,packageName_String,operationName_Symbol,OptionsPattern[{highlighted->{}}]]:=Module[
	{
		sys,
		qOpName=FMTSystem`fullyQualifiedName[packageName,operationName],
		highed=OptionValue[highlighted]
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	FMTAnalytics`controlFlow[
		FMTSystem`resolvedOperationBody[sys,packageName,qOpName,True,highed],
		highed
	]
]


showControlFlow[system_,packageName_String,operationName_Symbol]:=
	FMTFormatters`formatControlFlow[
		reportControlFlow[system,packageName,operationName]
	]


reportDataAliasing[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator1->None,progressIndicator2->None}]]:=Module[
	{
		sys,
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress1=OptionValue[progressIndicator1],
		progress2=OptionValue[progressIndicator2],
		globalsUsed=<||>,
		operationsCalled=<||>,
		globalsIndirectlyUsed,
		totalCount=0,
		count
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateOperationsSelector[pkgNames,opNames],Return[]];
	
	If[progress1=!=None&&progress2=!=None,
		(* count all items for use with progress indicator: *)
		FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
			Function[dummy,++totalCount]
		]
	];
	
	(* Phase 1. - gather the list of global objects used by each operation *)
	
	count=0;
	FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
		Function[{pkgName,opName},
			Module[
				{
					qOpName=FMTSystem`fullyQualifiedName[pkgName,opName],
					opBody,
					singleOpGlobals
				},
				
				If[FMTSystem`operationExists[sys,qOpName],
					opBody=FMTSystem`resolvedOperationBody[sys,pkgName,qOpName,False,{}];
					
					singleOpGlobals=FMTSystem`operationGlobals[
						sys,pkgName,qOpName,opBody][[All,1]];
					
					globalsUsed[qOpName]=singleOpGlobals;
					
					operationsCalled[qOpName]=FMTAnalytics`outgoingCalls[opBody]
				];
				
				If[progress1=!=None,
					progress1[{pkgName,opName},{++count,totalCount}]
				]
			]
		]
	];
	
	(* Phase 2. - check calls to other operations and their actual parameter lists:
		- if the same object is being used for two formals and at least one of them is inOut/out,
		or if a global object used by the called operation is used by the caller
		(this includes the use as formal parameter),
		then report data aliasing.
	*)
	
	globalsIndirectlyUsed=FMTAnalytics`indirectObjectUses[globalsUsed,operationsCalled];
	count=0;
	FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
		Function[{pkgName,opName},
			Module[
				{
					qOpName=FMTSystem`fullyQualifiedName[pkgName,opName],
					opBody,
					singleOpAliases={}
				},
				
				If[FMTSystem`operationExists[sys,qOpName],
					opBody=FMTSystem`resolvedOperationBody[sys,pkgName,qOpName,True,{}];
					
					singleOpAliases=FMTAnalytics`dataAliasing[
						qOpName,opBody,globalsUsed,globalsIndirectlyUsed]
				];
				
				If[progress2=!=None,
					progress2[{pkgName,opName},{++count,totalCount}]
				];
				
				singleOpAliases
			]
		]
	]
]


showDataAliasing[system_,OptionsPattern[{packageNames->All,operationNames->All}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		count1=0,count2=0,totalCount,
		result
	},
	
	PrintTemporary["checking flows..."];
	Monitor[
		result=reportDataAliasing[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator1->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count1=counts[[1]];
					totalCount=counts[[2]]
				],
			progressIndicator2->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count2=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Column[{
			Row[{ProgressIndicator[count1,{0,totalCount}]," ",count1,"/",totalCount}],
			Row[{ProgressIndicator[count2,{0,totalCount}]," ",count2,"/",totalCount}]
		}]
	];
	
	If[allReportsEmptyQ[result],
		Style["Data aliasing OK.",Green],
		Row[{
			Style["Data aliasing checks failed. ",Red],
			Button["  Show details  ",
				Function[{},
					Print[FMTFormatters`formatDataAliasingChecks[system,result]]
				]
			]
		}]
	]
]


reportProofStatements[system_,packageName_String,operationName_Symbol]:=Module[
	{
		sys,
		qOpName=FMTSystem`fullyQualifiedName[packageName,operationName],
		opProofStatements,
		variableDomains=<||>
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	opProofStatements=FMTProofs`proofStatements[sys,packageName,qOpName,variableDomains];
	
	{opProofStatements,variableDomains}
]


showProofStatements[system_,packageName_String,operationName_Symbol]:=With[
	{
		proofStatements=reportProofStatements[system,packageName,operationName],
		currentOptions=getOptions[system]
	},
	
	FMTFormatters`formatProofStatements[
		system,packageName,operationName,proofStatements[[1]],
		currentOptions[FMTOptions`formulaFormatter]
	]
]


reportProofs[system_,OptionsPattern[{packageNames->All,operationNames->All,progressIndicator->None}]]:=Module[
	{
		sys,
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		progress=OptionValue[progressIndicator],
		singleResult,
		totalCount=0,
		count=0
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateOperationsSelector[pkgNames,opNames],Return[]];
	
	If[progress=!=None,
		(* count all items for use with progress indicator: *)
		FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
			Function[dummy,++totalCount]
		]
	];
	
	FMTSystem`scanOperationsInPackages[sys,pkgNames,opNames,
		Function[{pkgName,opName},
			Module[
				{
					qOpName=FMTSystem`fullyQualifiedName[pkgName,opName],
					opProofStatements,
					variableDomains=<||>,
					mappedDomains,
					enumTypesUsed,
					enumMappings=<||>
				},
				
				If[FMTSystem`operationExists[sys,qOpName],
					(* 1. generate proof statements for the given operation *)
					opProofStatements=FMTProofs`proofStatements[sys,pkgName,qOpName,variableDomains];
					
					(* 2. obtain numeric mappings for all enum types used in variable domains *)
					enumTypesUsed=Cases[
						Union[Values[variableDomains]],
						Except[Integers|Booleans|_FMTSymbols`array|_FMTSymbols`structure]
					];
					Scan[
						Function[qE, (* fully qualified enum type name *)
							enumMappings[qE]=FMTSystem`enumValues[sys,pkgName,qE];
						],
						
						enumTypesUsed
					];
					
					(* 3. verify proofs for the statements defined above *)
					singleResult=FMTProver`verifyProofs[sys,pkgName,qOpName,
						opProofStatements,variableDomains,enumMappings],
					
					singleResult={}
				];
				
				If[progress=!=None,
					progress[{pkgName,opName},{++count,totalCount}]
				];
				
				singleResult
			]
		]
	]
]


showProofChecks[system_,OptionsPattern[{packageNames->All,operationNames->All}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		currentOptions=getOptions[system],
		count=0,totalCount=1,
		result,consolidatedResult
	},
	
	PrintTemporary["computing proofs..."];
	Monitor[
		result=reportProofs[
			system,
			packageNames->pkgNames,
			operationNames->opNames,
			progressIndicator->
				Function[{names,counts}, (* names is {pkgName,opName},
											counts is {count,totalCount} *)
					count=counts[[1]];
					totalCount=counts[[2]]
				]
		],
		Row[{ProgressIndicator[count,{0,totalCount}]," ",count,"/",totalCount}]
	];

	(* pick only the last column from all results,
		and check whether it contains only provenT entries: *)
	
	consolidatedResult=Head/@Flatten[Values[Values[result]],2][[All,3]];
	consolidatedResult=Union[consolidatedResult];
	If[consolidatedResult==={FMTSymbols`provenT}||consolidatedResult=={},
		Style["Proofs OK.",Green],
		Row[{
			Style["Proofs failed. ",Red],
			Button["  Show details  ",
				Function[{},
					CellPrint[
						ExpressionCell[
							FMTFormatters`formatProofChecks[system,result,
								currentOptions[FMTOptions`formulaFormatter]
							],
							"Output"
						]
					]
				]
			]
		}]
	]
]


showProofResults[system_,packageName_String,operationName_Symbol,OptionsPattern[{onlyFailed->True}]]:=With[
	{
		results=reportProofs[
			system,
			packageNames->packageName,
			operationNames->operationName
		][packageName][operationName],
		
		filtered=OptionValue[onlyFailed],
		currentOptions=getOptions[system]
	},
	
	FMTFormatters`formatSingleOpProofChecks[system,packageName,operationName,
		If[filtered,
			DeleteCases[results,{index_,_List,FMTSymbols`provenT[]},All],
			results
		],
		currentOptions[FMTOptions`formulaFormatter]
	]
]


runAllChecks[system_,OptionsPattern[{packageNames->All,operationNames->All}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		opNames=OptionValue[operationNames],
		grammarChecks,
		nameChecks,
		flowChecks,
		callCycles,
		aliasingChecks,
		proofs
	},
	
	grammarChecks=showGrammarChecks[system,packageNames->pkgNames,operationNames->opNames];
	nameChecks=showNameChecks[system,packageNames->pkgNames,operationNames->opNames];
	flowChecks=showFlowChecks[system,packageNames->pkgNames,operationNames->opNames];
	callCycles=showCallCycles[system,packageNames->pkgNames,operationNames->opNames];
	aliasingChecks=showDataAliasing[system,packageNames->pkgNames,operationNames->opNames];
	proofs=showProofChecks[system,packageNames->pkgNames,operationNames->opNames];
	
	Column[{
		grammarChecks,
		nameChecks,
		flowChecks,
		callCycles,
		aliasingChecks,
		proofs
	}]
]


generateCode[system_,language_String,OptionsPattern[{packageNames->All,translationPolicies-><||>}]]:=Module[
	{
		sys,
		pkgNames=OptionValue[packageNames],
		policies,
		singlePkgResult,
		result=<||>
	},
	
	Check[FMTSystem`validateSystem[system],Return[]];
	
	sys=system[[1]];
	
	Check[validateLanguage[language],Return[]];
	Check[validatePackageSelector[pkgNames],Return[]];
	
	If[pkgNames===All,
		pkgNames=FMTSystem`allPackageNames[sys],
		
		If[Head[pkgNames]===String,
			pkgNames={pkgNames}
		]
	];
	
	policies=Merge[{OptionValue[translationPolicies],FMTSystem`fmtGetCurrentOptions[sys]},First];
	
	(* Merge results for all selected packages *)
	Scan[
		Function[pkgName,
			singlePkgResult=Switch[language,
				"Ada",
					FMTGeneratorAda`fmtGenerateAda[sys,pkgName,policies],
				
				"C",
					FMTGeneratorC`fmtGenerateC[sys,pkgName,policies],
				
				"C++",
					FMTGeneratorCpp`fmtGenerateCpp[sys,pkgName,policies]
			];
			result=Merge[{result,singlePkgResult},First]
		],
		pkgNames
	];
	
	result
]


saveCode[system_,language_String,OptionsPattern[{packageNames->All,destinationDirectory->"",translationPolicies-><||>}]]:=Module[
	{
		pkgNames=OptionValue[packageNames],
		dir=OptionValue[destinationDirectory],
		trPolicies=OptionValue[translationPolicies],
		code
	},
	
	If[dir=="",
		dir=Directory[]
	];
	
	code=generateCode[system,language,packageNames->pkgNames,translationPolicies->trPolicies];
	
	Scan[
		Function[codeForPkg, (* fileName\[Rule]src *)
			With[
				{
					filePath=FileNameJoin[{dir,codeForPkg[[1]]}],
					src=codeForPkg[[2]]
				},
				
				If[src=!=Null,
					Export[filePath,src,"Text"]
				]
			]
		],
		Normal[code]
	];
]


End[];


EndPackage[];
